#ifndef ASSEMBLY_H_INCLUDED
#define ASSEMBLY_H_INCLUDED

#include <string.h>
#include <string>
#include <vector>
#include <math.h>
#include "lex.h"
#include "parse.h"

using namespace std;

class Assembler {
  private:
    FILE *file;
    Parser *parser;

    vector<Lexeme *> symbols;
    stack<int> ifStack;
    stack<int> forStack;

    string errorMessage;
    bool errorFound;

    int  variableCount, ifCount, forCount, lineCount;
	int  tileBank, tileFont, datasetDecoder;
	string datasetName;
	bool datasetEnabled, hasTileFont, hasDecoder;
	bool hasRleDecoder, hasPletterDecoder, hasTileFont1, hasTileFont2, hasTileFont3;

    TagNode *currentTag;
	ActionNode actionData;
	ActionNode actionDataSet;
    bool vdpActive;

    void buildSymbols();
    void buildSymbolsFromTag(TagNode *tag);
    void buildSymbolsFromAction(ActionNode *action);
    void addSymbol(string name);
    void addSymbol(Lexeme *lexeme);
    void addSymbol(ActionNode *action);

    void writeHeader();
    void writeBiosCalls();
    void writeBasicCalls();
    void writeMathCalls();
    bool writeMathKeyword(int i);
    void writeMainCode();
    void writeCodeFromTag(TagNode *tag);
    void writeCodeFromAction(ActionNode *action);
    void writeSymbols();
    void writeSymbolLiteral(Lexeme *lexeme);
    void writeSymbolIdentifier(Lexeme *lexeme);
    void writeSymbolKeyword(Lexeme *lexeme);
    void writeSymbolKeyword(string name);
	void writeDataSymbols();
	void writeSupportCode();
	void writeBasicSupportCode();
    void writeFooter();

    void writeInitializeSymbols();
    void writeInitializeSymbolLiteral(Lexeme *lexeme);
    void writeInitializeSymbolIdentifier(Lexeme *lexeme);

    bool create(char *filename);
    void write(string line);
    void writeNewLine();
    void writeSeparator();
    void close();

    void double2BCD(double value, int *words);
    void float2BCD(float value, int *words);
	void double2FloatLib(double value, int *words);
	void float2FloatLib(float value, int *words);
    double log10(double x);

    static const char * bios_functions;
    static const char * math_pack_functions;
    static const char * basic_functions;

	static const char * support_macros;
	static const char * support_basic_begin;
	static const char * support_basic_end;
	static const char * program_header;
	static const char * program_main_code;
	static const char * program_end_code;
	static const char * program_routines;
	static const char * program_footer;
	static const char * variables_routines;
	static const char * variables_initialize;
	static const char * variables_workarea_begin;
	static const char * variables_workarea_end;
	static const char * math16bits_routines;
	static const char * memory_allocation_routines;
	static const char * compression_routines;
	static const char * font_data_routines;

	static const char * float_wrapper;
	static const char * float_code;
    static const char * gfx_code;
    static const char * vdp_code;

    static const char * symbol_keywords[][2];

  public:
    enum { type_binary, type_rom, type_dos } headerType;
    bool noCtrlBreak, mathBCD, fastGFX, showLineNumber;

    bool build(Parser *parser, char *filename);

    void print();
    void error();

    Assembler();

};


#endif // ASSEMBLY_H_INCLUDED
