#ifndef FLOAT_H_INCLUDED
#define FLOAT_H_INCLUDED

// https://en.wikipedia.org/wiki/IEEE_754
// https://en.wikipedia.org/wiki/Single-precision_floating-point_format
// https://www.omnimaga.org/asm-language/(z80)-floating-point-routines
// https://www.msx.org/wiki/System_variables_and_work_area#Math-Pack
// https://github.com/Konamiman/MSX2-Technical-Handbook/blob/master/md/Appendix2.md
// https://github.com/Zeda/z80float

const char * Assembler::float_wrapper = R"(
;---------------------------------------------------------------------------------------------------------
; MATH PACK WRAPPER
;---------------------------------------------------------------------------------------------------------

CALL_MATH_LIB: exx
			     ld hl, RET_MATH_LIB
			     push hl
                   ld hl, BASIC_DAC
                   ld de, BASIC_ARG
			       ld bc, BASIC_SWPTMP
                   jp (ix)
RET_MATH_LIB:    call COPY_TO.TMP_DAC
               exx
               ret

if defined MATH.ADD

MATH_DECADD:   ld ix, addSingle
               jp CALL_MATH_LIB

endif

if defined MATH.SUB or defined MATH.NEG

MATH_DECSUB:   ld ix, subSingle
			   jp CALL_MATH_LIB

endif

if defined MATH.MULT

MATH_DECMUL:   ld ix, mulSingle
			   jp CALL_MATH_LIB

endif

if defined MATH.DIV

MATH_DECDIV:   ld ix, divSingle
			   jp CALL_MATH_LIB

endif

if defined MATH.POW

MATH_DBLEXP:
MATH_SNGEXP:   ld ix, powSingle
			   jp CALL_MATH_LIB

endif

if defined COS

MATH_COS:      ld ix, cosSingle
			   jp CALL_MATH_LIB

endif

if defined SIN

MATH_SIN:      ld ix, sinSingle
			   jp CALL_MATH_LIB

endif

if defined TAN

MATH_TAN:      ld ix, tanSingle
			   jp CALL_MATH_LIB

endif

if defined ATN

MATH_ATN:      ld ix, atanSingle
			   jp CALL_MATH_LIB

endif

if defined SQR

MATH_SQR:      ld ix, sqrtSingle
			   jp CALL_MATH_LIB

endif

if defined LOG

MATH_LOG:      ld ix, lnSingle
			   jp CALL_MATH_LIB

endif

if defined EXP

MATH_EXP:      ld ix, expSingle
			   jp CALL_MATH_LIB

endif

if defined ABS

MATH_ABSFN:    ld ix, absSingle
			   jp CALL_MATH_LIB

endif

if defined MATH.SEED or defined MATH.NEG

MATH_NEG:      ld ix, negSingle
			   jp CALL_MATH_LIB

endif

if defined SGN

MATH_SGN:      ld ix, sgnSingle
			   jp CALL_MATH_LIB

endif

if defined RND or defined MATH.SEED

MATH_RND:      ld ix, randSingle
               jp CALL_MATH_LIB

endif

MATH_FRCINT:   ld hl, BASIC_DAC
               ld bc, BASIC_DAC+2
			   call single2Int
			   ld ix, BASIC_DAC
			   ld (ix), 0
			   ld (ix+1), 0
			   ;ld (ix+2), l
			   ;ld (ix+3), h
			   ld (ix+4), 0
			   ld (ix+5), 0
			   ld (ix+6), 0
			   ld (ix+7), 0
               ld a, 2
               ld (BASIC_VALTYP), a
               ret

MATH_FRCDBL:                         ; same as MATH_FRCSGL
MATH_FRCSGL:   ld hl, BASIC_DAC+2    ; input address
               ld bc, BASIC_DAC      ; output address
               call int2Single
               ld a, 8
               ld (BASIC_VALTYP), a
               ret

MATH_ICOMP:         ld a, h   ; cp hl, de (alternative to bios DCOMPR)
                    cp d
			        jr nz, MATH_ICOMP.NE.HIGH
			          ld a, l
			          cp e
			          jr nz, MATH_ICOMP.NE.LOW
                        jr MATH_DCOMP.EQ
MATH_ICOMP.NE.HIGH: jr c, MATH_ICOMP.GT.HIGH
                    bit 7, a
                    jr nz, MATH_DCOMP.GT
			          jr MATH_DCOMP.LT
MATH_ICOMP.GT.HIGH: bit 7, d
                    jr z, MATH_DCOMP.GT
			          jr MATH_DCOMP.LT
MATH_ICOMP.NE.LOW:  jr c, MATH_DCOMP.GT
  			          jr MATH_DCOMP.LT

MATH_XDCOMP:                          ; same as MATH_DCOMP
MATH_DCOMP:    ld ix, cmpSingle
			   call CALL_MATH_LIB
			   jr z, MATH_DCOMP.EQ
			   jr c, MATH_DCOMP.LT
MATH_DCOMP.GT: ld a, 0xFF             ; DAC > ARG
               ret
MATH_DCOMP.EQ: ld a, 0                ; DAC = ARG
               ret
MATH_DCOMP.LT: ld a, 1                ; DAC < ARG
               ret

if defined CAST_STR_TO.VAL

MATH_FIN:      ; HL has the source string
               ld a, (BASIC_VALTYP)
               cp 2                   ; test if integer
			   jr nz, MATH_FIN.1
			   ;ld hl, (BASIC_DAC+2)
			   ;ld de, BASIC_STRBUF
			   ex de, hl
			   call StrToInt
			   ld bc, 0
			   ld (BASIC_DAC), bc
			   ld (BASIC_DAC+2), hl
			   ld (BASIC_DAC+4), bc
			   ld (BASIC_DAC+6), bc
			   ;ld hl, BASIC_STRBUF
			   ret
MATH_FIN.1:	   ld BC, BASIC_DAC
			   call str2single
               ret

endif

if defined CAST_INT_TO.STR

MATH_FOUT:     ld a, (BASIC_VALTYP)
               cp 2                   ; test if integer
			   jr nz, MATH_FOUT.1
			   ld hl, (BASIC_DAC+2)
			   ld de, BASIC_STRBUF
			   call IntToStr
			   ld hl, BASIC_STRBUF
			   ret
MATH_FOUT.1:   ld hl, BASIC_DAC
               ld bc, BASIC_STRBUF
               call single2str
			   ld hl, BASIC_STRBUF
               ret

endif

)";

const char * Assembler::float_code = R"(
;---------------------------------------------------------------------------------------------------------
; Z80FLOAT LIBRARY
; Copyright 2018 Zeda A.K. Thomas
;---------------------------------------------------------------------------------------------------------
; References:
; https://github.com/Zeda/z80float
; https://www.omnimaga.org/asm-language/(z80)-floating-point-routines/
; https://en.wikipedia.org/wiki/Single-precision_floating-point_format
;---------------------------------------------------------------------------------------------------------
; Parameters:
; HL points to the first operand
; DE points to the second operand (if needed)
; IX points to the third operand (if needed, rare)
; BC points to where the result should be output
; Floats are stored by a little-endian 24-bit mantissa. However, the highest bit
; is taken as implicitly 1, so we replace it as a sign bit. Next comes an 8-bit
; exponent biased by +128.
;---------------------------------------------------------------------------------------------------------
; Adapted to MSXBas2Asm by Amaury Carvalho, 2019
;---------------------------------------------------------------------------------------------------------

;---------------------------------------------------------------------------------------------------------
; Work area
;---------------------------------------------------------------------------------------------------------

BASIC_HOLD8: equ 0xF806  ; 	48	Work area for decimal multiplications.
BASIC_HOLD2: equ 0xF836  ;	8	Work area in the execution of numerical operators.
BASIC_HOLD:  equ 0xF83E  ;  8	Work area in the execution of numerical operators.
scrap:   equ BASIC_HOLD8
seed0:   equ BASIC_RNDX
seed1:   equ seed0 + 4
var48:   equ scrap + 4
quot:    equ scrap + 1
addend:  equ scrap
addend2: equ scrap+7           ;4 bytes
var_x:   equ BASIC_HOLD8 + 4   ;4 bytes
var_y:   equ var_x + 4         ;4 bytes
var_z:   equ var_y + 4         ;4 bytes
var_a:   equ var_z + 4         ;4 bytes
var_b:   equ var_a + 4         ;4 bytes
var_c:   equ var_b + 4         ;4 bytes
temp:    equ var_c + 4         ;4 bytes
temp1:   equ temp  + 4         ;4 bytes
temp2:   equ temp1 + 4         ;4 bytes
temp3:   equ temp2 + 4         ;4 bytes

pow10exp_single: equ scrap+9
strout_single:   equ 0xF750    ;  PARM2 - BASIC_BUF   ;pow10exp_single+2

;---------------------------------------------------------------------------------------------------------
; addSingle
;---------------------------------------------------------------------------------------------------------

;;Still need to tend to special cases
addSingle:
;;x+y
    push af
    push hl
    push de
    push bc
addInject:
    inc de
    inc de
    inc hl
    inc hl
    ld a,(de)
    xor (hl)
    push af
    inc de
    inc hl
    ex de,hl
    ld a,(de)
    sub (hl)
    ex de,hl
    jr nc,$+5
    ex de,hl
    neg
    cp 24
    jp nc,add_unneeded
    push hl
    ld hl,addend+6
    dec de
    ld bc,0408h
    dec hl
    ld (hl),0
    sub c
    jr nc,$-5
    add a,c
    push af
    push hl
    ex de,hl
    ld a,(hl)
    or 80h
    ld (de),a
    dec de
    dec hl
    ldd
    ldd
    ex de,hl
    dec b
    jr z,$+7
    ld (hl),0
    dec hl
    djnz $-3
    pop hl
    pop af
    ld b,a
    jr z,noshift
    set 7,(hl)
_1:
    push hl
    srl (hl)
    dec hl
    rr (hl)
    dec hl
    rr (hl)
    dec hl
    rr (hl)
    pop hl
    djnz _1
noshift:
    pop hl  ;bigger float
    dec hl
    ld b,(hl)
    dec hl
    dec hl
    ex de,hl
    pop af
    jp m,subtract
    ld hl,addend+2
    ld a,(hl)
    rla
    inc hl
    ld a,(de)
    adc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    adc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    set 7,a
    adc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    ld (hl),a
    jp nc,add_done
    inc (hl)
    jp z,add_overflow
    dec hl
    rr (hl)
    dec hl
    rr (hl)
    dec hl
    rr (hl)
    jp add_done
subtract:
    ld hl,addend
    xor a
    ld c,a
    sub (hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,(de)
    sbc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    sbc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    set 7,a
    sbc a,(hl)
    ld (hl),a
    inc hl
    inc de
    ld a,(de)
    ld (hl),a
    dec de
    ex de,hl
    jr nc,negated
    ld hl,addend
    ld a,80h
    xor b
    ld b,a
    ld a,c
    sub (hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
    inc hl
    ld a,c
    sbc a,(hl)
    ld (hl),a
negated:
    jp m,add_done
    push bc
    ld hl,(addend)
    ld de,(addend+2)
    ld bc,(addend+4)
    ld a,h
    or l
    or d
    or e
    or b
    or c
    jp z,add_underflow
    ld a,(addend+6)
normalize:
    dec a
    jr z,add_underflow
    add hl,hl
    rl e
    rl d
    rl c
    rl b
    jp p,normalize
    ld (addend),hl
    ld (addend+2),de
    ld (addend+4),bc
    ld (addend+6),a
    pop bc
add_done:
;;Need to adjust sign flag
    ld hl,addend+5
    ld a,(hl)
    rla
    rl b
    rra
    ld (hl),a
    dec hl
    dec hl
add_copy:
    pop de
    push de
    ldi
    ldi
    ldi
    ld a,(hl)
    ld (de),a
    pop bc
    pop de
    pop hl
    pop af
    ret
add_underflow:
;;How many push/pops are needed?
;;return ZERO
    ld hl,0
    ld (addend+3),hl
    ld (addend+5),hl
    pop bc
    jr add_done
add_overflow:
;;How many push/pops are needed?
;;return INF
    dec hl
    ld (hl),40h
    jr add_done
add_unneeded:
;;How many push/pops are needed?
;;Return bigger number
    pop af
    dec hl
    dec hl
    dec hl
    jr add_copy

;---------------------------------------------------------------------------------------------------------
; subSingle
;---------------------------------------------------------------------------------------------------------

subSingle:
;;x-y
    push af
    push hl
    push de
    push bc
    push hl
    ex de,hl
    ld de,addend2
    ldi
    ldi
    ld a,(hl)
    xor 80h
    ld (de),a
    inc de
    inc hl
    ld a,(hl)
    ld (de),a
    ex de,hl
    pop hl
    ld de,addend2
    jp addInject    ;jumps in to the addSingle routine

;---------------------------------------------------------------------------------------------------------
; mulSingle
;---------------------------------------------------------------------------------------------------------

mulSingle:
;Inputs: HL points to float1, DE points to float2, BC points to where the result is copied
;Outputs: float1*float2 is stored to (BC)
;573+mul24+{0,35}+{0,30}
;min: 1398cc
;max: 2564cc
;avg: 2055.13839751681cc
    push af
    push hl
    push de
    push bc

    call _2   ;CHLB
    ld a,c
    ex de,hl
    pop hl
    push hl
    ld (hl),b
    inc hl
    ld (hl),e
    inc hl
    ld (hl),d
    inc hl
    ld (hl),a
    pop bc
    pop de
    pop hl
    pop af
    ret


_2:
;;return float in CHLB
    push de
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld c,(hl)
    inc hl
    ld a,(hl)
    or a
    jr z,mulSingle_case0
    ex de,hl
    ex (sp),hl
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld b,(hl)
    inc hl

    ;inc (hl)
    ;dec (hl)
    ;jr z,mulSingle_case1
    push af
    ld a, (hl)
    or a
    jp z,mulSingle_case1
    pop af

    add a,(hl)      ;\
    pop hl          ; |
    rra             ; |Lots of help from Runer112 and
    adc a,a         ; |calc84maniac for optimizing
    jp po,bad       ; |this exponent check.
    xor 80h         ; |
    jr z,underflow  ;/
    push af         ;exponent
    ld a,b
    xor c
    push af         ;sign
    set 7,b
    set 7,c
    call mul24      ;BDE*CHL->HLBCDE, returns sign info
    pop de
    ld a,e
    pop de
    jp m,_3
    rl c
    rl b
    adc hl,hl
    dec d
_3:
    inc d
    jr z,overflow
    rl c
    ld c,d
    ld de,0
    push af
    ld a,b
    adc a,e
    ld b,a
    adc hl,de
    jr nc,_4
    inc c
    jr z,overflow
    rr h
    rr l
    rr b
_4:
    pop af
    cpl
    and $80
    xor h
    ld h,a
    ret
bad:
    jr nc,overflow
underflow:
    ld hl,0
    rl b
    rr h
    ld c,l
    ld b,l
    ret
overflow:
    ld hl,$8000
    jr underflow+3
mulSingle_case1:
;x*0   -> 0
;x*inf -> inf
;x*NaN -> NaN
  pop af
  pop hl
  ld h,b
  ld l,d
  ld b,e
  ld c,0
  ret
mulSingle_case0:
;special*x = special
;NaN*x = NaN
;0*0 = 0
;0*NaN = NaN
;0*Inf = NaN
;Inf*Inf  = Inf
;Inf*-Inf =-Inf
  ;0CDE
  pop hl
  inc hl
  inc hl
  inc hl
  ld a,(hl)
  or a
  jr z,_5
  ld h,c
  ld c,0
  ret
_5:
  dec hl
  ld b,(hl)
;basically, if b|c has bit 5 set, return NaN
  ld a,b
  or c
  ld h,$20
  and h
  jr z,_6
  ld c,0
  ret
_6:
  ld a,c
  xor b
  rl b
  rlca
  rr b
  res 4,b

  rl c
  rrca
  rr c

  ld a,c
  and $E0
  add a,b
  rra
  ld h,a
  ld c,0
  ret
mul24:
;;BDE*CHL -> HLBCDE
;;155 bytes
;;402+3*C_Times_BDE
;;fastest:1201cc
;;slowest:1753cc
;;avg    :1464.9033203125cc (1464+925/1024)
;min: 825cc
;max: 1926cc
;avg: 1449.63839751681cc

    push bc
    ld c,l
    push hl
    call C_Times_BDE
    ld (var48),hl
    ld l,a
    ld h,c
    ld (var48+2),hl

    pop hl
    ld c,h
    call C_Times_BDE
    push bc
    ld bc,(var48+1)
    add hl,bc
    ld (var48+1),hl
    pop bc
    ld b,c
    ld c,a
    ld hl,(var48+3)
    ld h,0
    adc hl,bc
    ld (var48+3),hl

    pop bc
    call C_Times_BDE
    ld de,(var48+2)
    add hl,de
    ld (var48+2),hl
    ld d,c
    ld e,a
    ld b,h
    ld c,l
    ld hl,(var48+4)
    ld h,0
    adc hl,de
    ld de,(var48)
    ret

;---------------------------------------------------------------------------------------------------------
; divSingle
;---------------------------------------------------------------------------------------------------------

divSingle:
;;HL points to numerator
;;DE points to denominator
;;BC points to where the quotient gets written
  call pushpop
divSingle_no_pushpop:
    inc hl
    inc de
    inc hl
    inc de
    ld a,(de)   ;\
    xor (hl)    ; |Get sign of output
    add a,a     ; |
    push af     ;/
    push bc
    inc hl
    inc de
    ld a,(hl)   ;\
    ex de,hl    ; |Get exponent
    sub (hl)    ; |
    ex de,hl    ; |

    ld b,-1
    jr nc,_7
    dec b
_7:
    add a,128
    jr nc,_8
    inc b
_8:
    inc b
    jr z,_9
    jp p,divunderflow
    jp m,divoverflow
_9:
    ld (quot+3),a
    dec hl
    dec de
    ld b,(hl)
    dec hl
    ld a,(hl)
    dec hl
    ld l,(hl)
    ld h,a
    ex de,hl

    ld c,(hl)
    dec hl
    ld a,(hl)
    dec hl
    ld l,(hl)
    ld h,a
    ex de,hl

    set 7,c
    ld a,b
    or 80h
    sbc hl,de
    sbc a,c
    jr nz,_10
    or h
    or l
    jr z,setmantissa0
    xor a
_10:
    jr nc,startdiv
    ld b,a
    ld a,(quot+3)
    dec a
    ld (quot+3),a
    ld a,b
    add hl,hl
    adc a,a
    add hl,de
    adc a,c
startdiv:
    ld b,1
    call divsub0+3
    ld (quot+1),bc
    call divsub0
    ld (quot),bc
    call divsub0
    ld (quot-1),bc
    add hl,hl
    rla
    jr c,_11
    sbc hl,de
    sbc a,c
    ccf
_11:
    ld hl,(quot)
    ld de,(quot+2)
    ld bc,0
    adc hl,bc
    ex de,hl
    adc hl,bc
    ld b,h
    ld c,l
writeback:
    pop hl
    ld (hl),e
    inc hl
    ld (hl),d
    inc hl
    rl c
    pop af
    rr c
    ld (hl),c
    inc hl
    ld (hl),b
    ret
divoverflow:
    ld b,$40
    jr _12
divunderflow:
  ld b,0
  jr _12
setmantissa0:
  ld bc,(quot+2)
_12:
  ld de,0
  ld c,e
  jr writeback
divsub0:
;;882cc max
    call divsub1    ;34 or 66
    call divsub1    ;
    call divsub1
    call divsub1
    call divsub1
    call divsub1
    call divsub1
    call divsub1
    or a
    sbc hl,de
    sbc a,c
    inc b
    ret nc
    dec b
    add hl,de
    adc a,c
    ret
divsub1:
;34cc or 66cc or 93cc
    sla b
    add hl,hl
    rla
    ret nc
    or a
    inc b
    sbc hl,de
    sbc a,c
    ret c
    inc b
    sbc hl,de
    sbc a,c
    ret

;---------------------------------------------------------------------------------------------------------
; powSingle
; https://www.geeksforgeeks.org/write-a-c-program-to-calculate-powxn/
; https://stackoverflow.com/questions/3518973/floating-point-exponentiation-without-power-function
;---------------------------------------------------------------------------------------------------------
;double mypow( double base, double power, double precision )
;{
;   if ( power < 0 ) return 1 / mypow( base, -power, precision );
;   else if ( power >= 1 ) return base * mypow( base, power-1, precision );
;   else if ( precision >= 1 ) {
;	   if( base >= 0 ) return sqrt( base );
;	   else return sqrt( -base );
;   } else return sqrt( mypow( base, power*2, precision*2 ) );
;}

if defined MATH.POW or defined MATH_EXP or defined MATH_LOG or defined MATH_LN

powSingle:
;;Computes y^x
;;HL points to y
;;DE points to x
;;BC points to output
    call pushpop
    push bc
      push de
	    ld bc, var_y     ; power
	    call copySingle
	  pop hl
	  ld bc, var_x       ; base
	  call copySingle
	  ld hl, const_precision
	  ld bc, var_a       ; precision
	  call copySingle
	  ld hl, const_0
	  ld bc, var_z       ; result
	  call copySingle
	  call powSingle.loop
	pop bc
	ld hl, var_z
	call copySingle
	ret

powSingle.loop:
;   if ( power < 0 ) return 1 / mypow( base, -power, precision );
    ld hl, var_y
	ld de, const_0
	call cmpSingle
	jp c, powSingle.1

;   else if ( power >= 1 ) return base * mypow( base, power-1, precision );
    ld hl, var_y
	ld de, const_1
	call cmpSingle
	jp nc, powSingle.2

;   else if ( precision >= 1 ) {
;	   if( base >= 0 ) return sqrt( base );
;	   else return sqrt( -base );
    ld hl, var_a
	ld de, const_1
	call cmpSingle
	jp nc, powSingle.3

;   } else return sqrt( mypow( base, power*2, precision*2 ) );
    ld hl, var_y
	ld de, const_2
	ld bc, var_b
	call mulSingle
	ld hl, var_b
	ld bc, var_y
	call copySingle
    ld hl, var_a
	ld de, const_2
	ld bc, var_b
	call mulSingle
	ld hl, var_b
	ld bc, var_a
	call copySingle
	call powSingle.loop
	ld hl, var_z
	ld bc, var_b
	call sqrtSingle
	ld hl, var_b
	ld bc, var_z
	call copySingle
	ret

powSingle.1:
; return 1 / mypow( base, -power, precision );
    ld hl, const_0
	ld de, var_y
	ld bc, var_b
	call subSingle
	ld hl, var_b
	ld bc, var_y
	call copySingle
	call powSingle.loop
	ld hl, const_1
	ld de, var_z
	ld bc, var_b
	call divSingle
	ld hl, var_b
	ld bc, var_z
	call copySingle
    ret

powSingle.2:
; return base * mypow( base, power-1, precision );
    ld hl, var_y
	ld de, const_1
	ld bc, var_b
	call subSingle
	ld hl, var_b
	ld bc, var_y
	call copySingle
	call powSingle.loop
	ld hl, var_z
	ld de, var_x
	ld bc, var_b
	call mulSingle
	ld hl, var_b
	ld bc, var_z
	call copySingle
    ret

powSingle.3:
;	   if( base >= 0 ) return sqrt( base );
;	   else return sqrt( -base );
    ld hl, var_x
	ld de, const_0
	call cmpSingle
	jp nc, powSingle.1
	;ld hl, var_x
	ld bc, var_b
	call negSingle
	ld hl, var_b
	;ld bc, var_z
	;call sqrtSingle
	;ret

powSingle.3.1:
    ;ld hl, var_x
	ld bc, var_z
	call sqrtSingle
    ret

pow2Single:
;;Computes 2^x
  call pushpop
  push bc

exp_inject:
;if x is on [0,1):
;  2^x = 1.000000001752 + x * (0.693146989552 + x * (0.2402298085906 + x * (5.54833215071e-2 + x * (9.67907584392e-3 + x * (1.243632065103e-3 + x * 2.171671843714e-4)))))
;Please note that usually I like to reduce to [-.5,.5] as the extra overhead is usually worth it.
;In this case, our polynomial is the same degree, with error different by less than 1 bit, so it's just a waste to range-reduce in this way.
;
;int(x) -> out_exp
;x-=int(x)  ;leaves x in [0,1)
;;If x==0    -> out==1
;;if x==inf  -> out==inf
;;if x==-inf -> out==0
;;if x==NAN  -> out==NAN
  ld de,var48+10
  call mov4
  ld hl,(var48+10)
  ld de,(var48+12)
  ld a,e
  add a,a
  push af   ;keep track of sign
  rrca
  ld (var48+12),a
  ld c,a
  ld a,d
    or a
    jp z,exp_spec
    cp 80h-23
    jp c,exp_underflow
    sub 128   ; sub a,128
    jr c,_pow_1 ;int(x)=0
    inc a
    cp 7
    jp nc,exp_overflow
    set 7,c
    ld b,a
    xor a
    add hl,hl
    rl c
    rla
    djnz $-4
    ld b,7Fh
    bit 7,c
    jr nz,exp_normalized
    ld e,a
    ld a,h
    or l
    or c
    ld a,e
    jr z,exp_zeroed
    dec b
    add hl,hl
    rl c
    jp p,$-4
    jr exp_normalized  ;.db $11 ;start of `ld de,**`
exp_zeroed:
    ld b,0
exp_normalized:
    ld (var48+10),hl
    res 7,c
    ld (var48+12),bc
    jr comp_exp   ;.db $06 ;start of 'ld b,*` just to eat the next byte
_pow_1:
    xor a
comp_exp:
  pop hl
  rr l
  jr nc,_pow_2
  cpl
  or a
  jp z,exp_underflow+1
  ;perform 1-(var48+10)--> var48+10
  ld hl,const_1
  ld de,var48+10
  ld b,d
  ld c,e
  call subSingle
_pow_2:
  push af
;our 'x' is at var48+10
;our `temp` is at var48+6 so as not to cause issues with mulSingle)
;uses 14 bytes of RAM
  ld hl,var48+10
  ld de,exp_a6
  ld bc,var48+6
  call mulSingle
  ld d,b
  ld e,c
  ld hl,exp_a5
  call addSingle
  ld hl,var48+10
  call mulSingle
  ld hl,exp_a4
  call addSingle
  ld hl,var48+10
  call mulSingle
  ld hl,exp_a3
  call addSingle
  ld hl,var48+10
  call mulSingle
  ld hl,exp_a2
  call addSingle
  ld hl,var48+10
  call mulSingle
  ld hl,exp_a1
  call addSingle
  ld hl,var48+10
  call mulSingle
  ld hl,const_1
  call addSingle
  ld hl,var48+9
  pop af
  add a,(hl)
  ld (hl),a
  ex de,hl
  pop de
  jp mov4
exp_spec:
;bit 6 means INF
;bit 5 means NAN
;no bits means zero
;NAN -> NAN
;+inf -> +inf
;-inf -> +0  because lim approaches 0 from the right
    ld a,c
    add a,a
    jr z,exp_zero
    jp m,exp_inf
;exp_NAN
    pop af
    ld de,0040h
exp_return_spec:
    pop hl
    rr e
    ld (hl),a
    inc hl
    ld (hl),a
    inc hl
    ld (hl),e
    inc hl
    ld (hl),d
    ret
exp_overflow:
exp_inf:
;+inf -> +inf
;-inf -> +0  because lim approaches 0 from the right
    pop af
    sbc a,a ;FF if should be 0,
    cpl
    and 80h
    ld d,0
    ld e,a
    jr exp_return_spec
exp_underflow:
exp_zero:
    pop af
    or a
    ld de,$8000
    jr exp_return_spec

endif

;---------------------------------------------------------------------------------------------------------
; sqrtSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_SQR or defined MATH_EXP

;Uses 3 bytes at scrap
sqrtSingle:
;552+{0,19}+8{0,3+{0,3}}+pushpop+sqrtHLIX
;min: 1784
;max: 1987
;avg: 1872
  call pushpop
  push bc
  ld c,(hl)
  inc hl
  ld e,(hl)
  inc hl
  ld a,(hl)
  add a,a
  jp c,sqrtSingle_NaN
  scf
  rra
  ld d,a
  inc hl
  ld a,(hl)
  or a
  jp z,sqrtSingle_special
  add a,80h
  rra
  push af   ;new exponent
  jr c,_13
  srl d
  rr e
  rr c
_13:
  ex de,hl
  ld ixh,c
  ld ixl,0
  call sqrtHLIX
;AHL is the new remainder
;Need to divide by 2, then divide by the 16-bit (var_x+4)
  rra
  ld a,h
;HL/DE to 8 bits
;We are just going to approximate it
  res 0,l
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l
  sla l
  rla
  jr c,$+5
  cp d
  jr c,$+4
  sub d
  inc l

  pop bc
  ld a,l
  pop hl
  ;BDEA
  ld (hl),a
  inc hl
  ld (hl),e
  inc hl
  res 7,d
  ld (hl),d
  inc hl
  ld (hl),b
  ret
sqrtSingle_NaN:
  ld hl,const_NaN
  pop de
  jp mov4
sqrtSingle_special:
  dec hl
  dec hl
  pop de
  jp mov4

sqrtHLIX:
;Input: HLIX
;Output: DE is the sqrt, AHL is the remainder
;speed: 754+{0,1}+6{0,6}+{0,3+{0,18}}+{0,38}+sqrtHL
;min: 1130
;max: 1266
;avg: 1190.5


  call sqrtHL
  add a,a
  ld e,a
  jr nc,_14
  inc d
_14:

  ld a,ixh
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_15
  add hl,de
  dec e
  jr _15a      ;.db $FE     ;start of `cp *`
_15:
  inc e
_15a:
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_16
  add hl,de
  dec e
  jr _16a   ;.db $FE     ;start of `cp *`
_16:
  inc e
_16a:
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_17
  add hl,de
  dec e
  jr _17a  ;.db $FE     ;start of `cp *`
_17:
  inc e
_17a:
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_18
  add hl,de
  dec e
  jr _18a  ;.db $FE     ;start of `cp *`
_18:
  inc e
_18a:
;Now we have four more iterations
;The first two are no problem
  ld a,ixl
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_19
  add hl,de
  dec e
  jr _19a  ;.db $FE     ;start of `cp *`
_19:
  inc e
_19a:
  sll e
  rl d
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl
  sbc hl,de
  jr nc,_20
  add hl,de
  dec e
  jr _20a  ;.db $FE     ;start of `cp *`
_20:
  inc e
_20a:
sqrt32_iter15:
;On the next iteration, HL might temporarily overflow by 1 bit
  sll e
  rl d      ;sla e \ rl d \ inc e
  add a,a
  adc hl,hl
  add a,a
  adc hl,hl       ;This might overflow!
  jr c,sqrt32_iter15_br0
;
  sbc hl,de
  jr nc,_21
  add hl,de
  dec e
  jr sqrt32_iter16
sqrt32_iter15_br0:
  or a
  sbc hl,de
_21:
  inc e

;On the next iteration, HL is allowed to overflow, DE could overflow with our current routine, but it needs to be shifted right at the end, anyways
sqrt32_iter16:
  add a,a
  ld b,a        ;either 0x00 or 0x80
  adc hl,hl
  rla
  adc hl,hl
  rla
;AHL - (DE+DE+1)
  sbc hl,de
  sbc a,b
  inc e
  or a
  sbc hl,de
  sbc a,b
  ret p
  add hl,de
  adc a,b
  dec e
  add hl,de
  adc a,b
  ret

sqrtHL:
;returns A as the sqrt, HL as the remainder, D = 0
;min: 376cc
;max: 416cc
;avg: 393cc
  ld de,$5040
  ld a,h
  sub e
  jr nc,_22
  add a,e
  ld d,$10
_22:
  sub d
  jr nc,_23
  add a,d
  jr _23a  ;.db $01   ;start of ld bc,** which is 10cc to skip the next two bytes.
_23:
  set 5,d
_23a:
  res 4,d
  srl d

  set 2,d
  sub d
  jr nc,_24
  add a,d
  jr _24a  ;.db $01   ;start of ld bc,** which is 10cc to skip the next two bytes.
_24:
  set 3,d
_24a:
  res 2,d
  srl d

  inc d
  sub d
  jr nc,_25
  add a,d
  dec d   ;this resets the low bit of D, so `srl d` resets carry.
  jr _25a  ;.db $06   ;start of ld b,* which is 7cc to skip the next byte.
_25:
  inc d
_25a:
  srl d
  ld h,a


  sbc hl,de
  ld a,e
  jr nc,_26
  add hl,de
_26:
  ccf
  rra
  srl d
  rra
  ld e,a

  sbc hl,de
  jr nc,_27
  add hl,de
  jr _27a  ;.db $01   ;start of ld bc,** which is 10cc to skip the next two bytes.
_27:
  or %00100000
_27a:
  xor %00011000
  srl d
  rra
  ld e,a


  sbc hl,de
  jr nc,_28
  add hl,de
  jr _28a  ;.db $01   ;start of ld bc,** which is 10cc to skip the next two bytes.
_28:
  or %00001000
_28a:
  xor %00000110
  srl d
  rra
  ld e,a
  sbc hl,de
  jr nc,_29
  add hl,de
  srl d
  rra
  ret
_29:
  inc a
  srl d
  rra
  ret

endif

;---------------------------------------------------------------------------------------------------------
; lnSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_LOG or defined MATH_LN

lnSingle:
; x / (1 + x/(2-x+4x/(3-2x+9x/(4-3x+16x/(5-4x)))))
; a * x ^ (1/a) - a, where a = 100
  call pushpop
  push bc
    ld de, const_100_inv
	ld bc, temp
	call powSingle         ; temp = x ^ (1/100)
	ld hl, temp
	ld de, const_100
	ld bc, temp1
	call mulSingle         ; temp1 = temp * 100
	ld hl, temp1
  pop bc
  call subSingle           ; bc = temp1 - 100
  ret

endif

;---------------------------------------------------------------------------------------------------------
; logSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_LOG

logSingle:
  call pushpop
  push bc
    ld bc, temp
    call lnSingle
    ld hl, temp
    ld de, const_lg10
  pop bc
  call divSingle
  ret

endif

;---------------------------------------------------------------------------------------------------------
; expSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_EXP

expSingle:
;;Computes e^x
;;HL points to x
;;BC points to the output
  call pushpop
  ld de,const_lg_e
  push bc
pow_inject:
;;DE points to lg(y), HL points to x, BC points to output
  ld bc,var_x
  call mulSingle
  ld h,b
  ld l,c
  jp exp_inject

endif

;---------------------------------------------------------------------------------------------------------
; sinSingle
; https://en.wikipedia.org/wiki/List_of_trigonometric_identities
; https://en.wikipedia.org/wiki/Taylor_series#Trigonometric_functions
; https://cs.stackexchange.com/questions/89245/how-approximate-sine-using-taylor-series
; https://stackoverflow.com/questions/42217069/approximating-sinex-with-a-taylor-series-in-c-and-having-a-lot-of-problems
;---------------------------------------------------------------------------------------------------------

if defined MATH_SIN or defined MATH_TAN or defined MATH_COS

sinSingle:
; taylor: x - x^3/6 + x^5/120 - x^7/5040
;         x(1 - x^2(1/6 - x^2(1/120 - x^2/5040)) )
; reduction:
;         var_b = round( x / (2*PI), 0 )
;         var_c = x - var_b*2*PI
;         temp1 = if( var_c >= 0, var_c, var_c + 2*PI )
;         temp2 = if( temp1 > PI, temp1 - PI, temp1 )
;         var_a = if( temp2 > PI/2, PI - temp2, temp2 ) * if( temp1 > PI, -1, 1 )

  call pushpop
  ld de, const_0
  call cmpSingle
  jr nz, sinSingle.1

  call copySingle      ; return 0
  ret

sinSingle.1:
  call trigRangeReductionSinCos
  push bc
    ld hl, var_a
    ld de, var_a
    ld bc, var_b
    call mulSingle    ; var_b = var_a * var_a
    ld hl, var_b
    ld de, sin_a3
    ld bc, temp
    call mulSingle    ; temp = x^2/5040
    ld hl, sin_a2
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1/120 - temp
    ld hl, var_b
    ld de, temp1
    ld bc, temp
    call mulSingle    ; temp = x^2 * temp1
    ld hl, sin_a1
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1/6 - temp
    ld hl, var_b
    ld de, temp1
    ld bc, temp
    call mulSingle    ; temp = x^2 * temp1
    ld hl, const_1
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1 - temp
    ld hl, var_a
    ld de, temp1
  pop bc
  call mulSingle      ; return x * temp1
  ret

trigRangeReductionSinCos:
  call pushpop
  push hl
; var_b = round( x / (2*PI), 0 )
    ld de, const_2pi
    ld bc, var_c
    call divSingle
    ld hl, var_c
	ld de, 0
	ld bc, var_b
	call roundSingle
; var_c = x - var_b*2*PI
    ld hl, var_b
    ld de, const_2pi
    ld bc, temp
    call mulSingle     ; temp = var_b*2*PI
  pop hl
  ld de, temp
  ld bc, var_c
  call subSingle     ; var_c = x - temp
; temp1 = if( var_c >= 0, var_c, var_c + 2*PI )
  ld hl, var_c
  ld de, const_0
  call cmpSingle
  jr nc, trigRangeReductionSinCos.else.2
    ld hl, var_c
    ld bc, temp1
    call copySingle     ; temp1 = var_c
    jr trigRangeReductionSinCos.endif.2
trigRangeReductionSinCos.else.2:
    ld hl, var_c
    ld de, const_2pi
    ld bc, temp1
    call addSingle     ; temp1 = var_c + 2*PI
trigRangeReductionSinCos.endif.2:
; temp2 = if( temp1 > PI, temp1 - PI, temp1 )
  ld hl, const_pi
  ld de, temp1
  call cmpSingle
  jr c, trigRangeReductionSinCos.else.3
  jr z, trigRangeReductionSinCos.else.3
    ld hl, temp1
    ld de, const_pi
    ld bc, temp2
    call subSingle     ; temp2
    jr trigRangeReductionSinCos.endif.3
trigRangeReductionSinCos.else.3:
    ld hl, temp1
    ld bc, temp2
    call copySingle     ; temp2 = temp1
trigRangeReductionSinCos.endif.3:
; var_a = if( temp2 > PI/2, PI - temp2, temp2 ) * if( temp1 > PI, -1, 1 )
  ld hl, const_half_pi
  ld de, temp2
  call cmpSingle
  jr c, trigRangeReductionSinCos.else.4
  jr z, trigRangeReductionSinCos.else.4
    ld hl, const_pi
    ld de, temp2
    ld bc, var_a
    call subSingle     ; var_a
    jr trigRangeReductionSinCos.endif.4
trigRangeReductionSinCos.else.4:
    ld hl, temp2
    ld bc, var_a
    call copySingle     ; var_a = temp2
trigRangeReductionSinCos.endif.4:
; if( temp > PI, -1, 1 )
  ld hl, temp1
  ld de, const_pi
  call cmpSingle
  jr nc, trigRangeReductionSinCos.endif.5
    ld ix, var_a
    ld a, (ix+2)
    set 7, a
    ld (ix+2), a   ; turn var_a to negative
trigRangeReductionSinCos.endif.5:
; return var_a
  ret

endif

;---------------------------------------------------------------------------------------------------------
; cosSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_COS or defined MATH_TAN

cosSingle:
; taylor: 1 - x^2/2 + x^4/24 - x^6/720
;         1 - x^2(1/2 - x^2(1/24 - x^2/720) )
; reduction: same as sin
;            cos = cos * sign

  call pushpop
  ld de, const_0
  call cmpSingle
  jr nz, cosSingle.1

  ld hl, const_1
  call copySingle      ; return 1
  ret

cosSingle.1:
  ; 1 - x^2(1/2 - x^2(1/24 - x^2/720) )
  call trigRangeReductionSinCos
  push bc
    ld hl, var_a
    ld de, var_a
    ld bc, var_b
    call mulSingle    ; var_b = var_a * var_a
    ld hl, var_b
    ld de, cos_a3
    ld bc, temp
    call mulSingle    ; temp = x^2/720
    ld hl, cos_a2
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1/24 - temp
    ld hl, var_b
    ld de, temp1
    ld bc, temp
    call mulSingle    ; temp = x^2 * temp1
    ld hl, cos_a1
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1/2 - temp
    ld hl, var_b
    ld de, temp1
    ld bc, temp
    call mulSingle    ; temp = x^2 * temp1
    ld hl, const_1
    ld de, temp
    ld bc, temp1
    call subSingle    ; temp1 = 1 - temp

    ; temp3 = abs(var_c)
    ; temp1 = temp1 * if( temp3 >= PI/2, -1, 1 )       ==> cos sign
	ld hl, var_c
	ld bc, temp3
	call copySingle
	ld ix, temp3
	ld a, (ix+2)
    res 7, a
	ld (ix+2), a      ; temp3 = abs(var_c)
	ld hl, temp3
	ld de, const_half_pi
    call cmpSingle    ; if temp3 >= PI/2 then temp1 = -temp1
    jr nc, cosSingle.endif.1
	ld ix, temp1
	ld a, (ix+2)
    set 7, a
	ld (ix+2), a      ; temp1 = -temp1
    cosSingle.endif.1:
  pop bc
  ld hl, temp1
  call copySingle      ; return temp1
  ret

endif

;---------------------------------------------------------------------------------------------------------
; tanSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_TAN

tanSingle:
  call pushpop
  push bc
  ;HL points to input
  ld bc,var_z
  ld d,b
  ld e,c
  call cosSingle
  ld bc,var_x
  call sinSingle
  ld h,b
  ld l,c
  pop bc
  jp divSingle

endif

;---------------------------------------------------------------------------------------------------------
; atanSingle
;---------------------------------------------------------------------------------------------------------

if defined MATH_ATN

atanSingle:
;taylor:    x/(1 + x^2/(3 + (2*x)^2/(5 + (3*x)^2/(7+(4*x)^2/9) ) ) )
;           x < -1: atan - PI/2
;           x >= 1: PI/2 - atan
;reduction: abs(X) > 1 : Y = 1 / X
;           abs(X) <= 1: Y = X
;           X < 0: Y = -Y

  call pushpop
  ld de, const_0
  call cmpSingle
  jr nz, atanSingle.1

  ld hl, const_0
  call copySingle      ; return 0
  ret

atanSingle.1:
  ;x/(1 + x^2/(3 + (2*x)^2/(5 + (3*x)^2/(7+(4*x)^2/9) ) ) )
  call trigRangeReductionAtan
  push bc
  push hl
    ld hl, var_a
    ld de, var_a
    ld bc, var_b
    call mulSingle    ; var_b = var_a * var_a
    ld hl, var_b
    ld de, const_16
    ld bc, temp
    call mulSingle    ; temp = (4*x)^2
    ld hl, temp
    ld de, const_9
    ld bc, temp1
    call divSingle    ; temp1 = temp/9
    ld hl, temp1
    ld de, const_7
    ld bc, temp
    call addSingle    ; temp = 7 + temp1
    ld hl, var_b
    ld de, const_9
    ld bc, temp1
    call mulSingle    ; temp1 = var_b * 9
    ld hl, temp1
    ld de, temp
    ld bc, temp2
    call divSingle    ; temp2 = temp1 / temp
    ld hl, temp2
    ld de, const_5
    ld bc, temp
    call addSingle    ; temp = 5 + temp2
    ld hl, var_b
    ld de, const_4
    ld bc, temp1
    call mulSingle    ; temp1 = var_b * 4
    ld hl, temp1
    ld de, temp
    ld bc, temp2
    call divSingle    ; temp2 = temp1 / temp
    ld hl, temp2
    ld de, const_3
    ld bc, temp
    call addSingle    ; temp = 3 + temp2
    ld hl, var_b
    ld de, temp
    ld bc, temp2
    call divSingle    ; temp2 = var_b / temp
    ld hl, temp2
    ld de, const_1
    ld bc, temp
    call addSingle    ; temp = 1 + temp2
    ld hl, var_a
    ld de, temp
    ld bc, temp2
    call divSingle    ; temp2 = var_a / temp
  pop hl
; x >= 1: PI/2 - atan
  ld de, const_1
  call cmpSingle
  jr nc, atanSingle.2
    ld hl, const_half_pi
    ld de, temp2
    ld bc, temp
    call subSingle
    ld hl, temp
    jr atanSingle.4
atanSingle.2:
; x < -1: atan - PI/2
  push hl
    ld hl, const_0
	ld de, const_1
	ld bc, temp
	call subSingle
  pop hl
  ld de, temp
  call cmpSingle
  jr c, atanSingle.3
    ld hl, temp2
    ld de, const_half_pi
    ld bc, temp
    call subSingle
    ld hl, temp
    jr atanSingle.4
atanSingle.3:
  ld hl, temp2
atanSingle.4:
  pop bc
  call copySingle      ; return temp2
  ret

trigRangeReductionAtan:
;reduction: abs(X) > 1 : Y = 1 / X
;           abs(X) <= 1: Y = X
;           X < 0: Y = -Y
  call pushpop
  push hl
    ld bc, temp
    call copySingle
    ld ix, temp
    ld a, (ix+2)
    res 7, a
    ld (ix+2), a   ; abs(x)
    ld hl, temp
    ld de, const_1
    call cmpSingle
    jr nc, trigRangeReductionAtan.1
      ld hl, const_1
	  pop de
	  push de
	  ld bc, var_a
	  call divSingle
	  jr trigRangeReductionAtan.2
trigRangeReductionAtan.1:
	  pop hl
	  push hl
	  ld bc, var_a
	  call copySingle
trigRangeReductionAtan.2:
  pop hl
  ld de, const_0
  call cmpSingle
  jr c, trigRangeReductionAtan.3
    ld ix, var_a
    ld a, (ix+2)
    set 7, a
    ld (ix+2), a   ; y = -y
trigRangeReductionAtan.3:
  ret

endif

if defined MATH_SIN or defined MATH_TAN or defined MATH_COS

;---------------------------------------------------------------------------------------------------------
; copySingle
;---------------------------------------------------------------------------------------------------------

copySingle:
    call pushpop
	;push bc
	;pop de
	ld d, b
	ld e, c
	ldi
	ldi
	ldi
	ldi
	ret

;---------------------------------------------------------------------------------------------------------
; roundSingle
;---------------------------------------------------------------------------------------------------------

roundSingle:
    call pushpop
	call copySingle
	;push bc
	;pop hl
	ld h, b
	ld l, c
	push de
	  ld a, e
	  ld de, const_10
roundSingle.1:
	  or 0
	  jr z, roundSingle.2
	  ld bc, temp
	  call mulSingle
	  ;push hl
	  ;pop bc
	  ld b, h
	  ld c, l
	  ld hl, temp
	  call copySingle
	  ;push bc
	  ;pop hl
	  ld h, b
	  ld l, c
	  dec a
	  jr roundSingle.1
roundSingle.2:
      ld de, const_half_1
	  ld bc, temp
	  call addSingle
      push hl
	    ld hl, temp
	    ld bc, temp1
	    call single2Int
	    ld hl, temp1
	  pop bc
	  call int2Single
	  ;push bc
	  ;pop hl
	  ld h, b
	  ld l, c
	pop de
    ld a, e
	ld de, const_10
roundSingle.3:
	or 0
	jr z, roundSingle.4
	ld bc, temp
	call divSingle
	;push hl
	;pop bc
	ld b, h
	ld c, l
	ld hl, temp
	call copySingle
	;push bc
	;pop hl
	ld h, b
	ld l, c
	dec a
	jr roundSingle.3
roundSingle.4:
	ret

endif

if defined MATH_ABSFN

;---------------------------------------------------------------------------------------------------------
; absSingle
;---------------------------------------------------------------------------------------------------------

absSingle:
;;HL points to the float
;;BC points to where to output the result
    call pushpop
    ld d,b
    ld e,c
    ldi
    ldi
    ld a,(hl)
    and %01111111
    ld (de),a
    inc hl
    inc de
    ld a,(hl)
    ld (de),a
    ret

endif

if defined MATH_SGN

;---------------------------------------------------------------------------------------------------------
; sgnSingle
;---------------------------------------------------------------------------------------------------------

sgnSingle:
;;HL points to the float
;;BC points to where to output the result
    jp negSingle

endif

if defined powSingle or defined sgnSingle or defined MATH_NEG

;---------------------------------------------------------------------------------------------------------
; negSingle
;---------------------------------------------------------------------------------------------------------

negSingle:
;;HL points to the float
;;BC points to where to output the result
    call pushpop
	push hl
	pop ix
	ld a, (ix+3)
	or 0
	jr nz, negSingle.test.sign
	ld a, (ix+2)
	or 0
	jr nz, negSingle.test.sign
	ld a, (ix+1)
	or 0
	jr nz, negSingle.test.sign
	ld a, (ix)
	or 0
	jr nz, negSingle.test.sign
    ;push bc
    ;pop de
	ld d, b
	ld e, c
    ld hl, const_0
    ldi
    ldi
    ldi
    ldi
    ret
negSingle.test.sign:
	ld a, (ix+2)
	bit 7, a
	jr z, negSingle.positive
negSingle.negative:
    push bc
	pop ix
	call negSingle.positive
	ld a, (ix+2)
	set 7, a
	ld (ix+2), a
    ret
negSingle.positive:
    ;push bc
    ;pop de
	ld d, b
	ld e, c
    ld hl, const_1
    ldi
    ldi
    ldi
    ldi
    ret

endif

if defined MATH_DCOMP or defined MATH.POW or defined MATH_EXP or defined MATH_LOG or defined MATH_LN or defined MATH_SIN or defined MATH_TAN or defined MATH_COS or defined MATH_ATN

;---------------------------------------------------------------------------------------------------------
; cmpSingle
;---------------------------------------------------------------------------------------------------------

cmpSingle:
;Input: HL points to float1, DE points to float2
;Output:
;      float1 >= float2 : nc
;      float1 <  float2 : c,nz
;      float1 == float2 : z
;  There is a margin of error allowed in the lower 2 bits of the mantissa.
;
;Currently fails when both numbers have magnitude less than about 2^-106
  push hl
  push de
  push bc
  ld c, a
  push bc
    ex de, hl
    call _30
  pop bc
  ld a, c
  pop bc
  pop de
  pop hl
  ret
_30:
  inc de
  inc de
  inc de
  ld a,(de)
  inc hl
  inc hl
  inc hl
  cp (hl)
  jr nc,_31
  ld a,(hl)
_31:
  dec hl
  dec hl
  dec hl
  dec de
  dec de
  dec de
  push af
  ld bc,scrap
  call subSingle
  ld a,(scrap+3)    ;new power
  pop bc            ;B is old power
  or a
  jr z,cmp_close
  sub b
  jr nc,cmp_is_sign
  dec a
  add a,22
  jr nc,cmp_close
cmp_is_sign:
  ld a,(scrap+2)
  or 1    ;not equal, so reset z flag
  rla     ;if negative, float1<float2, setting c flag as wanted, else nc.
  ret
cmp_close:
  xor a
  ret

endif

if defined MATH_RND

;---------------------------------------------------------------------------------------------------------
; randSingle
;---------------------------------------------------------------------------------------------------------

randSingle:
;Stores a pseudo-random number on [0,1)
;it won't produce values on (0,2^-23)
  call pushpop
  push bc
  call rand
  push hl
  call rand
  pop de
  ex de,hl
  ld bc,$207F
;DEHL is the mantissa, B is the exponent
  ld a,d
  or a
  jp m,rand_normed
_32:
  dec c
  add hl,hl
  rl e
  rl d
  jp m,rand_normed
  djnz _32
rand_zero:
  ld c,l
  ld b,l
  jr rand_done
rand_normed:
;If we needed to shift more than 8 bits, we'll load in more random data
  ld a,b
  cp 8
  jr c,rand_zero
  sub 24
  jp nc,rand_no_more_rand_data
  push bc
  push de
  call rand
  pop de
  ld e,h
  ld h,l
  pop bc
rand_no_more_rand_data:
  ld b,e
  ld e,d
  ld d,c
  ld c,h
  res 7,e
rand_done:
  pop hl
  ;DEBC
  ld (hl),b
  inc hl
  ld (hl),c
  inc hl
  ld (hl),e
  inc hl
  ld (hl),d
  ret

rand:
;;Tested and passes all CAcert tests
;;Uses a very simple 32-bit LCG and 32-bit LFSR
;;it has a period of 18,446,744,069,414,584,320
;;roughly 18.4 quintillion.
;;LFSR taps: 0,2,6,7  = 11000101
;;323cc
;;Thanks to Runer112 for his help on optimizing the LCG and suggesting to try the much simpler LCG. On their own, the two are terrible, but together they are great.
;Uses 64 bits of state
  ld hl,(seed0)
  ld de,(seed0+2)
  ld b,h
  ld c,l
  add hl,hl
  rl e
  rl d
  add hl,hl
  rl e
  rl d
  inc l
  add hl,bc
  ld (seed0),hl
  ld hl,(seed0+2)
  adc hl,de
  ld (seed0+2),hl
  ex de,hl
;;lfsr
  ld hl,(seed1)
  ld bc,(seed1+2)
  add hl,hl
  rl c
  rl b
  ld (seed1+2),bc
  sbc a,a
  and %11000101
  xor l
  ld l,a
  ld (seed1),hl
  ex de,hl
  add hl,bc
  ret

endif

if defined MATH_FOUT

;---------------------------------------------------------------------------------------------------------
; single2Str
; in  HL = Single address
;     BC = String address
; out A = String size
; http://0x80.pl/notesen/2015-12-29-float-to-string.html
; http://0x80.pl/articles/convert-float-to-integer.html
;---------------------------------------------------------------------------------------------------------

single2str:
  call pushpop
  push bc
  call _33
  pop de
  xor a
  cp (hl)
  ldi
  jr nz,$-3

  ret
_33:
; Move the float to scrap
  ld de,scrap
  call mov4

; Make the float negative, write a '-' if already negative
  ld de,strout_single
  ld hl,scrap+2
  ld a,(hl)
  ;rlca
  ;scf
  ;rra
  bit 7, a
  jr z, _34
  ld a,'-'      ; write '-' simbol
  ld (de),a
  inc de
  ld a,(hl)
_34:
  set 7, a
  ld (hl),a

; Check if the exponent field is 0 (a special value)
  inc hl
  ld a,(hl)
  or a
  jp z,strcase_single


; We should write '0' next. When rounding 9.999999... for example, not padding with a 0 will return '.' instead of '1.'
  ex de,hl
  ld (hl),'0'
  inc hl

; Save the pointer
  push hl

; Now we need to perform signed (A-128)*77 (approximation of exponent*log10(2))
  ld de,77
  ld h,a
  ld l,d
  call mul8_preset
  ld de,-77*128
  add hl,de
  ld a,h
  ld (pow10exp_single),a    ;The base-10 exponent
  ld de,pown10LUT
  jr c,_35
  neg
  ld de,pow10LUT   ;get the table of 10^-(2^k)
_35:
  ld hl, pow10exp_single
  ld bc,scrap
  call singletostr_mul
  call singletostr_mul
  call singletostr_mul
  call singletostr_mul
  call singletostr_mul
  call singletostr_mul
;now the number is pretty close to a nice value

; If it is less than 1, multiply by 10
  ld a,(scrap+3)
  sub 128
  jr nc,_36
  ld de,const_10
  ;ld hl,scrap    ;Since singletostr_mul returns BC = scrap, can do this cheaper
  ;ld b,h
  ;ld c,l
  ld h,b
  ld l,c
  call mulSingle
  ld hl,pow10exp_single
  dec (hl)
  ld a,(scrap+3)
  sub 128
_36:

; Convert to a fixed-point number !
  inc a
  ld b,a
  xor a
_37:
  ld hl,scrap
  sla (hl)
  inc hl
  rl (hl)
  inc hl
  rl (hl)
  rla
  djnz _37

;We need to get 7 digits
  ld b,6
  pop hl    ;Points to the string

;The first digit can be as large as 20, so it'll actually be two digits
  cp 10
  jr c,_38
  dec b
;Increment the exponent :)
  ld de,(pow10exp_single-1)
  inc d
  ld (pow10exp_single-1),de
;
  ld (hl),'0'-1
  inc (hl)
  sub 10
  jr nc,$-3
  add a,10
  inc hl
_38:
; Get the remaining digits.
_39:
  add a,'0'
  ld (hl),a
  inc hl
  push hl
  push bc
  call singletostrmul10
  pop bc
  pop hl
  djnz _39

;Save the pointer to the end of the string
  ld d,h
  ld e,l
  ;ld (hl), 0

;Now let's round!
  cp 5
  jr c,rounding_done_single
  jr _40a  ;.db $DA ;start of `jp c,*` in order to skip the next instruction
_40:
  ld (hl),'0'
_40a:
  dec hl
  inc (hl)
  ld a,(hl)
  cp $3A
  jr z,_40
rounding_done_single:


;Strip the leading zero if it exists (rounding may have bumped this to `1`)
  ld hl,strout_single
  ld a,(hl)
  cp '-'
  jr nz,_41
  inc hl
  ld a,(hl)
_41:
  cp '0'
  jr nz,_42
  dec de
  ex de,hl
  ;Now lets move HL-DE bytes at DE+1 to DE
  sbc hl,de
  ld b,h
  ld c,l
  ld h,d
  ld l,e
  inc hl
  ldir
  cp a
_42:

  push de
;If z flag is reset, this means that the exponent should be bumped up 1
  ld a,(pow10exp_single)
  jr z,_43
  inc a
  ld (pow10exp_single),a
_43:

  ;if -4<=A<=6, then need to insert the decimal place somewhere.
  add a,4
  cp 10
  jp c,movdec_single
_44:
  ;for this, we need to insert the decimal after the first digit
  ;Then, we need to append the exponent string
  ld hl,strout_single
  ld de,strout_single-1
  ld a,(hl)
  cp '-'    ;negative sign
  jr nz,_45
  ldi
_45:
  ldi
  ld a,'.'
  ld (de),a

;remove any stray zeroes at the end before appending the exponent
  pop hl
  call strip_zeroes

; Write the exponent
  ld (hl),'e'
  inc hl
  ld a,(pow10exp_single)
  or a
  jp p,_46
  ld (hl),'-'    ;negative sign
  inc hl
  neg
_46:
  cp 10
  jr c,_47
  ld (hl),'0'-1
  inc (hl)
  sub 10
  jr nc,$-3
  add a,10
  inc hl
_47:
  add a,'0'
  ld (hl),a
  inc hl
  ld (hl),0

  ld de, strout_single
  xor a
  sbc hl, de
  ld a, l         ; string size

  ld hl,strout_single-1
  ret

movdec_single:
  ld a,(pow10exp_single)
  or a
  jp p,posdec_single
  ld l,a
;need to put zeroes before everything
  ld de,strout_single
  ld a,(de)
  cp '-'    ;negative sign
  push af
  ld a,'0'
  jr z,$+3
_48:
  dec de
  ld (de),a
  inc l
  jr nz,_48
_49:
  ex de,hl
  ld (hl),'.'
  pop af
  jr nz,_50
  dec hl
  ld (hl),a
_50:
  ex de,hl
  pop hl
  call strip_zeroes
  ld (hl),0
  ex de,hl
  ret

posdec_single:
  ld hl,strout_single
  ld de,strout_single-1
  ld c,a
  ld a,(hl)
  ld b,0
  cp '-'    ;negative sign
  jr nz,_51
  inc c
_51:
  inc c
  ldir
  ld a,'.'
  ld (de),a
  pop hl
  call strip_zeroes
  ld (hl),0
  ld hl,strout_single-1
  ret

strcase_single:
  ld hl,str_Zero
  ld a,(scrap+2)
  add a,a
  and $C0
  jr z,_52
  ld hl,str_Inf
  jp pe,_52
  ld hl,str_NaN
_52:
  call mov4
  ld hl,strout_single
  ret

singletostrmul10:
;multiply the 0.24 fixed point number at scrap by 10
;overflow in A register
  ld a,(scrap+2)
  ld e,a
  ld hl,(scrap)
  xor a
  ld d,e
  ld b,h
  ld c,l
  add hl,hl
  rl d
  rla
  add hl,hl
  rl d
  rla
  add hl,bc
  ld b,a
  ld a,d
  adc a,e
  ld d,a
  ld a,b
  adc a,0
  add hl,hl
  rl d
  rla
  ld (scrap+1),de
  ld (scrap),hl
  ret

strip_zeroes:
  ld a,'0'
_53:
  dec hl
  cp (hl)
  jr z,_53

;Check that the last  digit isn't a decimal!
  ld a,'.'
  cp (hl)
  ret z
  inc hl
  ret

singletostr_mul:
  rra
  call c,_54
  ld hl,4
  add hl,de
  ex de,hl
  ret
_54:
  ld h,b
  ld l,c
  jp mulSingle
mul8:
;H*E => HL
  ld l,0
  ld d,l
mul8_preset:
  sla h
  jr nc,$+3
  ld l,e
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  jr nc,$+3
  add hl,de
  add hl,hl
  ret nc
  add hl,de
  ret

endif


if defined MATH_FIN

;---------------------------------------------------------------------------------------------------------
; str2Single
; https://www.ticalc.org/pub/86/asm/source/routines/atof.asm
;---------------------------------------------------------------------------------------------------------

char_NEG: equ  '-'
char_ENG: equ  ','
char_DEC: equ  '.'
ptr_sto: equ scrap+9

;;#Routines/Single Precision
;;Inputs:
;;  HL points to the string
;;  BC points to where the float is output
;;Output:
;;  scrap+9 is the pointer to the end of the string
;;Destroys:
;;  11 bytes at scrap ?

str2single:
  call pushpop
  push bc
;Check if there is a negative sign.
;   Save for later
;   Advance ptr
  ld a,(hl)
  sub char_NEG
  sub 1
  push af
  jr nc,$+3
  inc hl
;Skip all leading zeroes
  ld a,(hl)
  cp '0'
  jr z,$-4      ;jumps back to the `inc hl`
;Set exponent to 0
  ld b,0
;Check if the next char is char_DEC
  sub char_DEC
  or a      ;to reset the carry flag
  jr nz,_55
  jr _54a   ;.db $FE   ;start of cp *
;Get rid of zeroes
  dec b
_54a:
  inc hl
  ld a,(hl)
  cp '0'
  jr z,$-5      ;jumps back to the `dec b`
  scf
_55:
;Now we read in the next 8 digits
  ld de,scrap+3
  call ascii_to_uint8
  call ascii_to_uint8
  call ascii_to_uint8
  call ascii_to_uint8
;Now `scrap` holds the 4-digit base-100 number.
;b is the exponent
;if carry flag is set, just need to get rid of remaining digits
;Otherwise, need to get rid of remaining digits, while incrementing the exponent
  sbc a,a
  inc a
  ld c,a
_56:
  ld a,(hl)
  cp 30h
  jr nz,_57
  inc hl
  ld a,b
  add a,c
  jp z,strToSingle_inf
  ld b,a
  jr _56
;Now check for engineering `E` to modify the exponent
_57:
  cp char_NEG
  call z,str_eng_exp
;Gotta multiply the number at (scrap) by 2^24
  ld (ptr_sto),hl
  ld d,100
  call scrap_times_256
  ld a,c
  ld (scrap+6),a
  call scrap_times_256
  ld a,c
  ld (scrap+5),a
  call scrap_times_256
  ld a,c
  ld (scrap+4),a
  call scrap_times_256
  ld a,c
  ld (scrap+3),a
;Now scrap+3 is a 4-byte mantissa that needs to be normalized
;
  ld hl,(scrap+3)
  ld a,h
  or l
  ld hl,(scrap+5)
  or l
  or h
  jp z,strToSingle_zero-1
  ld c,$7F
  ld a,h
  or a
  jp m,strToSingle_normed
  ;Will need to iterate at most three times
_58:
  dec c
  ld hl,scrap+3
  sla (hl)
  inc hl
  rl (hl)
  inc hl
  rl (hl)
  inc hl
  adc a,a
  jp p,_58
strToSingle_normed:
;Move the number to scrap
  ld hl,(scrap+4)
  ld (scrap),hl
  ld l,a
  ld h,c
  sla l
  pop af
  rr l
  ld (scrap+2),hl
;now (scrap) is our number, need to multiply by power of 10!
;Power of 10 is stored in B, need to put in A first
  xor a
  sub b
  ld de,pown10LUT
  jp p,_59
  ld a,b
  ld de,pow10LUT
  cp 40
  jp nc,strToSingle_inf+1
_59:
  cp 40
  jp nc,strToSingle_zero
  ld hl,scrap
  ld b,h
  ld c,l
  call _60
  call _60
  call _60
  call _60
  call _60
  call _60
  pop de
  jp mov4
_60:
  rra
  call c,mulSingle
  inc de
  inc de
  inc de
  inc de
  ret
str_eng_exp:
  ld de,0
  inc hl
  ld a,(hl)
  cp char_NEG    ;negative exponent?
  push af
  jr nz,$+3
  inc hl
_61:
  ld a,(hl)
  sub 3Ah
  add a,10
  jr nc,_62
  inc hl
  push hl
  ld h,d
  ld l,e
  add hl,hl
  add hl,hl
  add hl,de
  add hl,hl
  add a,l
  ld l,a
  ex de,hl
  pop hl
  jp c,eng_overflow
  inc d
  dec d
  jp z,_61
  jp nz,eng_overflow
_62:
  ld a,e
  cp 40
  jr nc,eng_overflow
  pop af
  ld a,b
  jr nz,_63
  sub e
  ld b,a
  ret
_63:
  add a,e
  ld b,a
  ret
scrap_times_256:
  ld e,8
_64:
  or a
  ld hl,scrap
  call _65
  call _65
  rl c
  dec e
  jr nz,_64
  ret
_65:
  call scrap_times_sub
scrap_times_sub:
  ld a,(hl)
  rla
  cp d
  jr c,$+3
  sub d
  ld (hl),a
  inc hl
  ccf
  ret
eng_overflow:
  pop af
  jr nz,strToSingle_inf
  pop af
strToSingle_zero:
  ld hl,const_0
  pop de
  jp mov4
strToSingle_inf:
;return inf
  pop af
  ld hl,const_inf
  jr nc,_66
  ld hl,const_NegInf
_66:
  pop de
  jp mov4

endif

if defined roundSingle or defined MATH_FRCSGL

;---------------------------------------------------------------------------------------------------------
; int2Single
; http://wikiti.brandonw.net/index.php?title=Z80_Routines:Math:Division#24.2F8_division
;---------------------------------------------------------------------------------------------------------

int2Single:
    call pushpop
	push bc
	    push hl
		pop ix
	    ld l, (ix)            ; convert integer parameter to single float
		ld h, (ix+1)
		ld bc, 0x1000         ; bynary digits count + sign

int2Single.test.zero:
        xor a
		or h                  ; test if hl is not zero
		jr nz, int2Single.test.negative
		or l
		jr nz, int2Single.test.negative
		ld hl, 0
		ld de, 0
		jp int2Single.save

int2Single.test.negative:
        bit 7, h              ; test if hl is negative
		jr z, int2Single.normalize
		ld c, 0x80            ; sign negative
		ld a, h               ;\
		cpl                   ; |
		ld h, a               ; | abs(hl)
		ld a, l               ; |
		cpl                   ; |
		ld l, a               ;/
		inc hl

int2Single.normalize:
        dec b
        bit 7, h
		jr nz, int2Single.mount
		sla l
		rl h
		jr int2Single.normalize

int2Single.mount:
		res 7, h              ; turn off upper bit

        ld a, c               ; restore sign
        or h                  ; put sign...
        ld h, a               ; ...into upper mantissa

		ld e, h               ; sign+mantissa
		ld h, l               ; high mantissa
		ld l, 0               ; low mantissa

        ld a, b               ; binary digits count
        or 0x80               ; exponent bias
        ld d, a               ; exponent

int2Single.save:
    pop ix
	ld (ix),   l          ; low mantissa
	ld (ix+1), h          ; high mantissa
	ld (ix+2), e          ; sign + mantissa
	ld (ix+3), d          ; expoent
	ld (ix+4), 0
	ld (ix+5), 0
	ld (ix+6), 0
	ld (ix+7), 0
	ret

endif

if defined roundSingle or defined MATH_FRCINT

;---------------------------------------------------------------------------------------------------------
; single2Int
; http://0x80.pl/articles/convert-float-to-integer.html
;---------------------------------------------------------------------------------------------------------
single2Int:
;Input:
; HL points to the single-precision float
;Output:
; HL is the 16-bit signed integer part of the float
; BC points to 16-bit signed integer
  call pushpop
  push bc
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld a,(hl)
    add a,a
    push af
    scf
    rra
    ld c,a
    inc hl
    ld a,(hl)
    ld hl,0
    sub 80h
    jr c,no_shift_single_to_int16
    cp 39
    jr nc,no_shift_single_to_int16
    sub 8
    jr c,_67
    ld l,c
    ld c,d
    ld d,e
    ld e,h
    sub 8
    jr c,_67
    ld h,l
    ld l,c
    ld c,d
    ld d,e
    sub 8
    jr c,_67
    ld h,l
    ld l,c
    ld c,d
    sub 8
    jr c,_67
    ld h,l
    ld l,c
    jr _67a ;.db $11 ;start of ld de,*
_67:
    add a,9
_67a:
    ld b,a
    ld a,e
_68:
    add a,a
    rl d
    rl c
    adc hl,hl
    djnz _68
no_shift_single_to_int16:
    pop af
    jr nc,_69
    ;need to negate
    xor a
    sub e
    ld e,0
    ld a,e
    sbc a,d
    ld a,e
    sbc a,c
    ld d,e
    ex de,hl
    sbc hl,de
_69:
  pop ix
  ld (ix), l
  ld (ix+1), h
  ret

endif

;---------------------------------------------------------------------------------------------------------
; Auxiliary routines
;---------------------------------------------------------------------------------------------------------

str_Zero: db "0",0
str_Inf:  db "inf",0
str_NaN:  db "NaN",0

start_const:
const_pi:      db $DB,$0F,$49,$81
const_e:       db $54,$f8,$2d,$81
const_lg_e:    db $3b,$AA,$38,$80
const_ln_2:    db $18,$72,$31,$7f
const_log2:    db $9b,$20,$1a,$7e
const_lg10:    db $78,$9a,$54,$81
const_0:       db $00,$00,$00,$00
const_1:       db $00,$00,$00,$80
const_2:       dw 0, 33024
const_3:       dw 0, 33088
const_4:       dw 0, 33280
const_5:       dw 0, 33312
const_7:       dw 0, 33376
const_9:       dw 0, 33552
const_16:      dw 0, 33792
const_100:     db $00,$00,$48,$86
const_100_inv: dw 55050, 31011
const_precision: db $77,$CC,$2B,$65  ;10^-8
const_half_1:  dw 0, 32512
const_inf:     db $00,$00,$40,$00
const_NegInf:  db $00,$00,$C0,$00
const_NaN:     db $00,$00,$20,$00
const_log10_e: db $D9,$5B,$5E,$7E
const_2pi:     db $DB,$0F,$49,$82
const_2pi_inv: db $83,$F9,$22,$7D
const_half_pi: dw 4059, 32841
const_p25:     db $00,$00,$00,$7E
const_p5:      db $00,$00,$00,$7F
;     db $,$,$,$
end_const:
sin_a1: dw 43691, 32042
sin_a2: dw 34952, 30984
sin_a3: dw 3329, 29520
cos_a1: equ const_half_1
cos_a2: dw 43691, 31530
cos_a3: dw 2914, 30262
exp_a1: db $15,$72,$31,$7F  ;.693146989552
exp_a2: db $CE,$FE,$75,$7D  ;.2402298085906
exp_a3: db $7B,$42,$63,$7B  ;.0554833215071
exp_a4: db $FD,$94,$1E,$79  ;.00967907584392
exp_a5: db $5E,$01,$23,$76  ;.001243632065103
exp_a6: db $5F,$B7,$63,$73  ;.0002171671843714
const_1p40625: db $00,$00,$34,$80  ;1.40625

if defined MATH_CONSTSINGLE

iconstSingle:
    ex (sp),hl
    ld a,(hl)
    inc hl
    ex (sp),hl
constSingle:
;A is the constant ID#
;returns nc if failed, c otherwise
;HL points to the constant
    cp (end_const-start_const)>>2
    ret nc
    ld hl,start_const
    add a,a
    add a,a
    add a,l
    ld l,a
;#if ((end_const-4)>>8)!=(start_const>>8)
;    ccf
;    ret c
;    inc h
;#endif
    scf
    ret

endif

;;LUTs used
lut:
pown10LUT:
db $CD,$CC,$4C,$7C  ;.1
db $0A,$D7,$23,$79  ;.01
db $17,$B7,$51,$72  ;.0001
db $77,$CC,$2B,$65  ;10^-8
db $95,$95,$66,$4A  ;10^-16
db $1F,$B1,$4F,$15  ;10^-32
pow10LUT:
const_10:
db $00,$00,$20,$83 ;10
db $00,$00,$48,$86 ;100
db $00,$40,$1C,$8D ;10000
db $20,$BC,$3E,$9A ;10^8
db $CA,$1B,$0E,$B5 ;10^16
db $AE,$C5,$1D,$EA ;10^32

C_Times_BDE:
;;C*BDE => CAHL
;C = 0     157
;C = 1     141
;141+
;C>=128    135+6{0,33+{0,1}}+{0,20+{0,8}}
;C>=64     115+5{0,33+{0,1}}+{0,20+{0,8}}
;C>=32     95+4{0,33+{0,1}}+{0,20+{0,8}}
;C>=16     75+3{0,33+{0,1}}+{0,20+{0,8}}
;C>=8      55+2{0,33+{0,1}}+{0,20+{0,8}}
;C>=4      35+{0,33+{0,1}}+{0,20+{0,8}}
;C>=2      15+{0,20+{0,8}}
;min: 141cc
;max: 508cc
;avg: 349.21279907227cc

  ld a,b
  ld h,d
  ld l,e
  sla c
  jr c,mul8_24_1
  sla c
  jr c,mul8_24_2
  sla c
  jr c,mul8_24_3
  sla c
  jr c,mul8_24_4
  sla c
  jr c,mul8_24_5
  sla c
  jr c,mul8_24_6
  sla c
  jr c,mul8_24_7
  sla c
  ret c
  ld a,c
  ld h,c
  ld l,c
  ret
mul8_24_1:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_2:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_3:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_4:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_5:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_6:
    add hl,hl
    rla
    rl c
    jr nc,$+7
    add hl,de
    adc a,b
    jr nc,$+3
    inc c
mul8_24_7:
    add hl,hl
    rla
    rl c
    ret nc
    add hl,de
    adc a,b
    ret nc
    inc c
    ret

pushpop:
;26 bytes, adds 118cc to the traditional routine
  ex (sp),hl
  push de
  push bc
  push af
  push hl
  ld hl,pushpopret
  ex (sp),hl
  push hl
  push af
  ld hl,12
  add hl,sp
  ld a,(hl)
  inc hl
  ld h,(hl)
  ld l,a
  pop af
  ret
pushpopret:
  pop af
  pop bc
  pop de
  pop hl
  ret

mov4:
  ldi
  ldi
  ldi
  ldi
  ret

if defined MATH_FIN

ascii_to_uint8:
;c flag means don't increment the exponent
  ld c,0
  ld a,(hl)
  jr c,ascii_to_uint8_noexp
  cp char_DEC
  jr z,ascii_to_uint8_noexp-2
_70:
  sub 3Ah
  add a,10
  jr nc,ascii_to_uint8_noexp_end
  inc b
  ld c,a
  add a,a
  add a,a
  add a,c
  add a,a
  ld c,a
  inc hl
_71:
  ld a,(hl)
  cp char_DEC
  jr z,ascii_to_uint8_noexp_2nd
_72:
  sub 3Ah
  add a,10
  jr nc,ascii_to_uint8_noexp_end
  inc b
  add a,c
  inc hl
  ld (de),a
  dec de
  or a
  ret

  inc hl
  ld a,(hl)
ascii_to_uint8_noexp:
  sub 3Ah
  add a,10
  jr nc,ascii_to_uint8_noexp_end
  ld c,a
  add a,a
  add a,a
  add a,c
  add a,a
  ld c,a
ascii_to_uint8_noexp_2nd:
  inc hl
  ld a,(hl)
  sub 3Ah
  add a,10
  jr nc,ascii_to_uint8_noexp_end
  add a,c
  inc hl
  jr ascii_2  ;.db $FE   ;start of `cp **`, saves 1cc
ascii_to_uint8_noexp_end:
  ld a,c
ascii_2:
  ld (de),a
  dec de
  scf
  ret

endif

if defined MATH_RSUBSINGLE

rsubSingle:
;;-x+y
    push af
    push hl
    push de
    push bc
    push de
    ld de,addend2
    ldi
    ldi
    ld a,(hl)
    xor 80h
    ld (de),a
    inc de
    inc hl
    ld a,(hl)
    ld (de),a
    pop de
    ld hl,addend2
    jp addInject    ;jumps in to the addSingle routine

endif

if defined MATH_MOD1SINGLE

;This routine performs `x mod 1`, returning a non-negative value.
;+inf -> NaN
;-inf -> NaN
;NaN  -> NaN
mod1Single:
  call pushpop
  push bc
  ld e,(hl)
  inc hl
  ld d,(hl)
  inc hl
  ld c,(hl)
  ld a,c
  xor 80h
  push af
  jp p,mod1Single.1
  ld c,a
mod1Single.1:

  inc hl
  ld a,(hl)
  ld b,a
  or a
  jr z,mod1Single_special
  sub $80
  jr c,mod1_end
  inc a
  ld b,a
  ld a,c
  ex de,hl
mod1Single.2:
  add hl,hl
  rla
  djnz mod1Single.2
  ld c,a

;If it is zero, need to set exponent to zero and return
  or h
  or l
  ex de,hl
  jr z,mod1_end

;Need to normalize
  ld b,$7F
  ld a,c
  or a
  jp m,mod1_end
  ex de,hl
mod1Single.3:
  dec b
  add hl,hl
  adc a,a
  jp p,mod1Single.3
  ld c,a
  ex de,hl
mod1_end:
  pop af
  pop hl
  jp m,mod1Single.4
  ;make sure it isn't zero else we need to add 1
  ld a,b
  or a
  jr z,mod1Single.4
  ld (scrap),de
  ld (scrap+2),bc
  ld b,h
  ld c,l
  ld hl,scrap
  ld de,const_1
  jp addSingle
mod1Single_special:
;If INF, need to return NaN instead
;For 0 and NaN, just return itself :)
  pop af
  pop hl
  ld a,c
  add a,a
  jp p,mod1Single.4
  ld c,$40
mod1Single.4:
  res 7,c
  ld (hl),e
  inc hl
  ld (hl),d
  inc hl
  ld (hl),c
  inc hl
  ld (hl),b
  ret

endif

if defined MATH_FOUT

; --------------------------------------------------------------
; Converts a signed integer value to a zero-terminated ASCII
; string representative of that value (using radix 10).
; References:
; Brandon Wilson WikiTI
; http://wikiti.brandonw.net/index.php?title=Z80_Routines:Other:DispA#Decimal_Signed_Version
; --------------------------------------------------------------
; INPUTS:
;     HL     Value to convert (two's complement integer).
;     DE     Base address of string destination. (pointer).
; --------------------------------------------------------------
; OUTPUTS:
;     A      Size of string
; --------------------------------------------------------------
; REGISTERS/MEMORY DESTROYED
; AF HL
; --------------------------------------------------------------

IntToStr:
   push    de
   push    bc

; Detect sign of HL.
    bit    7, h
    jr     z, _DoConvert

; HL is negative. Output '-' to string and negate HL.
    ld     a, '-'
    ld     (de), a
    inc    de

; Negate HL (using two's complement)
    xor    a
    sub    l
    ld     l, a
    ld     a, 0     ; Note that XOR A or SUB A would disturb CF
    sbc    a, h
    ld     h, a

; Convert HL to digit characters
_DoConvert:
    ld     b, 0     ; B will count character length of number
_DoConvert.1:
    ld     c, 10
    call div_hl_c; HL = HL / A, A = remainder
    push   af
    inc    b
    ld     a, h
    or     l
    jr     nz, _DoConvert.1

; Retrieve digits from stack
_DoConvert.2:
    pop    af
    or     $30
    ld     (de), a
    inc    de
    djnz   _DoConvert.2

; Terminate string with NULL
    xor    a
    ld     (de), a

    ld h, d
    ld l, e

    pop    bc
    pop    de

    sbc hl, de
    ld a, l           ; string size

    ret

endif

if defined MATH_FIN

;===============================================================
; Convert a string of base-10 digits to a 16-bit value.
; http://z80-heaven.wikidot.com/math#toc32
;Input:
;     DE points to the base 10 number string in RAM.
;Outputs:
;     HL is the 16-bit value of the number
;     DE points to the byte after the number
;     BC is HL/10
;     z flag reset (nz)
;     c flag reset (nc)
;Destroys:
;     A (actually, add 30h and you get the ending token)
;Size:  23 bytes
;Speed: 104n+42+11c
;       n is the number of digits
;       c is at most n-2
;       at most 595 cycles for any 16-bit decimal value
;===============================================================

StrToInt:
     ld hl,0          ;  10 : 210000
     dec de
StrToInt.Skip_spaces:
     inc de           ;   6 : 13
     ld a,(de)        ;   7 : 1A
     or 0
     ret z
     cp 32
     jr z, StrToInt.Skip_spaces
StrToInt.Init:
     push af
     cp '-'
     jr nz, StrToInt.ConvLoop
     inc de
StrToInt.ConvLoop:             ;
     ld a,(de)        ;   7 : 1A
     or 0
     jr z, StrToInt.End
     sub 30h          ;   7 : D630
     cp 10            ;   7 : FE0A
     jr nc, StrToInt.End
                      ;
     inc de           ;   6 : 13
     ld b,h           ;   4 : 44
     ld c,l           ;   4 : 4D
     add hl,hl        ;  11 : 29
     add hl,hl        ;  11 : 29
     add hl,bc        ;  11 : 09
     add hl,hl        ;  11 : 29
                      ;
     add a,l          ;   4 : 85
     ld l,a           ;   4 : 6F
     jr nc, StrToInt.ConvLoop   ;12|23: 30EE
     inc h            ; --- : 24
     jr StrToInt.ConvLoop      ; --- : 18EB

StrToInt.End:
     pop af
     cp '-'
     ret nz
     ld de, 0xffff
     ex de, hl
     dec de
     xor a
     sbc hl, de
     ret

endif

if defined IntToStr

; divides hl by c
; return remainder in a
; http://wikiti.brandonw.net/index.php?title=Z80_Routines:Math:Division
div_hl_c:
   push bc
   xor	a
   ld	b, 16
div_hl_c.loop:
   add	hl, hl
   rla
   jr	c, $+5
   cp	c
   jr	c, $+4
   sub	c
   inc	l
   djnz	div_hl_c.loop
   pop bc
   ret

endif

if defined DIV_EHL

; http://wikiti.brandonw.net/index.php?title=Z80_Routines:Math:Division#24.2F8_division
div_ehl_d:
   xor	a
   ld	b, 24
div_ehl_d.loop:
   add	hl, hl
   rl	e
   rla
   jr	c, $+5
   cp	d
   jr	c, $+4
   sub	d
   inc	l
   djnz	div_ehl_d.loop
   ret

div_dehl_c:
   push bc
   xor	a
   ld	b, 32
div_dehl_c.loop:
   add	hl, hl
   rl	e
   rl	d
   rla
   jr	c, $+5
   cp	c
   jr	c, $+4
   sub	c
   inc	l
   djnz div_dehl_c.loop
   pop bc
   ret

endif
)";

#endif // FLOAT_H_INCLUDED
