#ifndef GFX_H_INCLUDED
#define GFX_H_INCLUDED

const char * Assembler::gfx_code = R"(
;---------------------------------------------------------------------------------------------------------
; GRAPHICS LIBRARY
; By: Amaury Carvalho, 2019
;---------------------------------------------------------------------------------------------------------
; References:
; https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm#Algorithm_for_integer_arithmetic
; https://en.wikipedia.org/wiki/Midpoint_circle_algorithm
; https://rosettacode.org/wiki/Bitmap/Midpoint_circle_algorithm#C
; https://www.msx.org/wiki/MSX-BASIC_Instructions
;---------------------------------------------------------------------------------------------------------

;---------------------------------------------------------------------------------------------------------
; Bios functions
;---------------------------------------------------------------------------------------------------------

BIOS_WRTVDP:  EQU 0x0047
BIOS_RDVRM:   EQU 0x004A
BIOS_WRTVRM:  EQU 0x004D
BIOS_LDIRVM:  EQU 0x005C
BIOS_LDIRMV:  EQU 0x0059
BIOS_PNTINI:  EQU 0x18CF
BIOS_RIGHTC:  EQU 0x16C5 ; Move current pixel physical address right
BIOS_TRIGHTC: EQU 0x16AC ; Test then RIGHTC if legal
BIOS_LEFTC:   EQU 0x16EE ; Move current pixel physical address left
BIOS_TLEFTC:  EQU 0x16D8 ; Test then LEFTC if legal
BIOS_UPC:     EQU 0x175D ; Move current pixel physical address up
BIOS_TUPC:    EQU 0x173C ; Test then UPC if legal
BIOS_DOWNC:   EQU 0x172A ; Move current pixel physical address down
BIOS_TDOWNC:  EQU 0x170A ; Test then DOWNC if legal
BIOS_DCOMPR:  EQU 0x146A ; compare HL and DE (Flag NC if HL>DE, Flag Z if HL=DE, Flag C if HL<DE)
BIOS_FILVRM:  EQU 0x0056 ; fill VRAM with value

BIOS_BIGFIL:  EQU 0x016B ; msx 2
BIOS_NRDVRM:  EQU 0x0174 ; msx 2
BIOS_NWRVRM:  EQU 0x0177 ; msx 2
BIOS_NRDVDP:  EQU 0x013E ; msx 2
BIOS_VDPSTA:  EQU 0x0131 ; msx 2
BIOS_NWRVDP:  EQU 0x012D ; msx 2 (0x0647)

BASIC_SUB_LINE:           equ 0x58fc
BASIC_SUB_LINEBOX:        equ 0x5912
BASIC_SUB_LINEBOXFILLED:  equ 0x58C1
BASIC_SUB_CIRCLE:         equ 0x5B19
BASIC_SUB_PAINT1:         equ 0x59DA   ;0x59C8
BASIC_SUB_PAINT2:         equ 0x0069   ;0x2664   ;0x2651+3

;---------------------------------------------------------------------------------------------------------
; Work areas
;---------------------------------------------------------------------------------------------------------

BIOS_RG0SAV: EQU 0xF3DF
BIOS_RG1SAV: EQU 0xF3E0
BIOS_RG8SAV: EQU 0xFFE7
BIOS_BDRATR: EQU 0xFCB2
BIOS_STATFL: EQU 0xF3E7  ; VDP status register

BIOS_CXOFF:  EQU 0xF945
BIOS_CYOFF:  EQU 0xF947
BIOS_GXPOS:  EQU 0xFCB3
BIOS_GYPOS:  EQU 0xFCB5

BIOS_GRPNAM: EQU 0xF3C7  ; pattern name table
BIOS_GRPCOL: EQU 0xF3C9  ; colour table
BIOS_GRPCGP: EQU 0xF3CB  ; pattern generator table
BIOS_GRPATR: EQU 0xF3CD  ; sprite attribute table
BIOS_GRPPAT: EQU 0xF3CF  ; sprite generator table
BIOS_CGPNT:  EQU 0xF920  ; 2 - current MSX Font location (0x1BBF)
BIOS_ATRBAS: EQU 0xF928  ; sprite attribute table

BIOS_MLTNAM: EQU 0xF3D1  ; pattern name table (screen 3, multicolor)
BIOS_MLTCOL: EQU 0xF3D3  ; colour table (screen 3, multicolor)
BIOS_MLTCGP: EQU 0xF3D5  ; pattern generator table (screen 3, multicolor)
BIOS_MLTATR: EQU 0xF3D7  ; sprite attribute table (screen 3, multicolor)
BIOS_MLTPAT: EQU 0xF3D9  ; sprite generator table (screen 3, multicolor)

BIOS_ASPECT:       equ 0xF931   ;2	Aspect ratio of the circle; set by <ratio> of CIRCLE.
BIOS_CENCNT:       equ 0xF933   ;2  Counter used by CIRCLE.
BIOS_CLINEF:       equ 0xF935	;1	Flag to draw line to centre, Used set by CIRCLE
BIOS_CNPNTS:       equ 0xF936	;2	Point to be plottted in a 45° segment, Used set by CIRCLE
BIOS_CPLOTF:       equ 0xF938	;1	Plot polarity flag, Used set by CIRCLE
BIOS_CPCNT:        equ 0xF939   ;2	Number of points in 1/8 of circle, Used set by CIRCLE.
BIOS_CPCNT8:       equ 0xF93B   ;2	Number of points in the circle. Used by CIRCLE.
BIOS_CRCSUM:       equ 0xF93D   ;2	Cyclic redundancy check sum of the circle. Used by CIRCLE.
BIOS_CSTCNT:       equ 0xF93F	;2	Variable to maintain the number of points of the starting angle. Used by the instruction CIRCLE
BIOS_CSCLXY:       equ 0xF941	;1	Scale of X & Y. Used by the instruction CIRCLE
BIOS_ASPCT1:       equ 0xF40B	;2	256/aspect ratio for Basic instruction CIRCLE.
BIOS_ASPCT2:       equ 0xF40D   ;2	256*aspect ratio for Basic instruction CIRCLE.
BIOS_MAXUPD:       equ 0xF3EC   ;3	Work area used by the instruction CIRCLE, contains JP 0000h at start.
BIOS_MINUPD:       equ 0xF3EF   ;3	Work area used by the instruction CIRCLE, contains JP 0000h at start.

BIOS_PARM1:  EQU 0xF6E8 ; 100
BIOS_PARM2:  EQU 0xF750 ; 100

GFX_TEMP:   EQU BIOS_PARM1     ; 2
GFX_TEMP1:  EQU GFX_TEMP  + 2  ; 2
GFX_TEMP2:  EQU GFX_TEMP1 + 2  ; 2
GFX_TEMP3:  EQU GFX_TEMP2 + 2  ; 2
GFX_TEMP4:  EQU GFX_TEMP3 + 2  ; 2
GFX_TEMP5:  EQU GFX_TEMP4 + 2  ; 2
GFX_TEMP6:  EQU GFX_TEMP5 + 2  ; 2
GFX_TEMP7:  EQU GFX_TEMP6 + 2  ; 2
GFX_TEMP8:  EQU GFX_TEMP7 + 2  ; 2
GFX_TEMP9:  EQU GFX_TEMP8 + 2  ; 2

GFX_MAX_X:           EQU 0xFCA4  ; 1 (CASSETE LOWLIM)
GFX_MAX_Y:           EQU 0xFCA5  ; 1 (CASSETE WINWID)

GFX_SPRITE_FLAGS:        EQU 0xF40A  ; 1 (CASSETE HEADER) - bits 0=check screen limits, 1=check walls, 2=check hotspots,
                                     ;                           3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
GFX_SPRITE_SIZE_DAT:     EQU 0xF3FC  ; 1 (CASSETE CS1200)
GFX_SPRITE_SIZE_SCR:     EQU 0xF3FD  ; 1
GFX_SPRITE_WALLS:        EQU 0xF406  ; 2 (CASSETE LOW)
GFX_SPRITE_HOTSPOTS:     EQU 0xF408  ; 2 (CASSETE HIGH)
GFX_SPRITE_HOTSPOT_TILE: EQU 0xF405  ; 1 (CASSETE CS2400)
GFX_SPRITE_COLLIDER:     EQU 0xF400  ; 1 (CASSETE CS1200)
GFX_SPRITE_COLLISION:    EQU 0xF7B5  ; 2 (ARYTA2)

GFX_MUSIC_START:         EQU 0xF401  ; 2 (CASSETE CS2400)
GFX_MUSIC_NEXT:          EQU 0xF403  ; 2
GFX_MUSIC_PREV:          EQU 0xF74C  ; 2 (PRMPRV)

GFX_TEMP10:              EQU 0xF3FE  ; 2 (CASSETE CS1200)

;BIOS_SCR_SIZE_X: dw 240, 256, 256, 64, 256, 256, 512, 512, 256, 512, 256, 256, 256
;BIOS_SCR_SIZE_Y: dw 192, 192, 192, 48, 192, 212, 212, 212, 212, 384, 212, 212, 212
BIOS_SCR_SIZE_X: db 239, 255, 255, 63, 255, 255, 255, 255, 255, 255, 255, 255, 255
BIOS_SCR_SIZE_Y: db 191, 191, 191, 47, 191, 211, 211, 211, 211, 255, 211, 211, 211


;---------------------------------------------------------------------------------------------------------
; gfxIsScreenModeMSX2
; return if screen mode is from MSX 2
; out C is set, if MSX2 and screen mode above 3
;---------------------------------------------------------------------------------------------------------

gfxIsScreenModeMSX2:
  ld a, (BIOS_VERSION)
  or 0
  jp nz, BIOS_CHKNEW  ; if not MSX1, jump to CHKNEW
  scf
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetScreenMode
; set current screen mode
; in A = screen number
;---------------------------------------------------------------------------------------------------------

gfxSetScreenMode:
  push af
    call gfxInitScreenWorkspace
    xor a
    ld a, (BIOS_VERSION)
    or 0
    jr nz, gfxSetScreenMode.1  ; if not MSX1, jump
  pop af
  cp 4
  call nc, gfxSetScreenMode.0  ; if screen mode >= 4, change to screen 2
  __call_bios BIOS_CHGMOD      ; change the screen mode (msx1)
  jr gfxSetScreenMode.2

gfxSetScreenMode.0:
  ld a, 2
  ret

gfxSetScreenMode.1:
  pop af
  ld ix, BIOS_CHGMOD2          ; change the screen mode (msx2)
  call BIOS_EXTROM

gfxSetScreenMode.2:
  call gfxGetScreenHeight
  call gfxGetScreenWidth
  call gfxGetSpriteSize
  jp gfxFillSpriteCollisionTable

gfxInitScreenWorkspace:
  ld a, 1
  ld (GFX_SPRITE_FLAGS), a     ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  xor a
  ld (GFX_SPRITE_WALLS), a
  ld (GFX_SPRITE_WALLS+1), a
  ld (GFX_SPRITE_HOTSPOTS), a
  ld (GFX_SPRITE_HOTSPOTS+1), a
  ld (GFX_MUSIC_START), a
  ld (GFX_MUSIC_START+1), a
  ld (GFX_MUSIC_NEXT), a
  ld (GFX_MUSIC_NEXT+1), a
  ld (GFX_MUSIC_PREV), a
  ld (GFX_MUSIC_PREV+1), a
  ld (GFX_SPRITE_HOTSPOT_TILE), a
  ld (GFX_SPRITE_COLLIDER), a
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetScreenMode
; return current screen mode
; out A = screen number (0=40x24 Text Mode, 1=32x24 Text Mode, 2=Graphics Mode, 3=Multicolour Mode)
;---------------------------------------------------------------------------------------------------------

gfxGetScreenMode:
  ld a, (BIOS_SCRMOD)
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetXY
; set current screen location
; in BC = x
;    DE = y
;---------------------------------------------------------------------------------------------------------

gfxSetXY:
  ld (BIOS_GRPACX), bc    ; x
  ;ld (BIOS_GXPOS), bc
  ld (BIOS_GRPACY), de    ; y
  ;ld (BIOS_GYPOS), de
  jr gfxRefreshXY.1

;---------------------------------------------------------------------------------------------------------
; gfxRefreshXY
; refresh current screen location
;---------------------------------------------------------------------------------------------------------

gfxRefreshXY:
  ld bc, (BIOS_GRPACX)    ; x
  ld de, (BIOS_GRPACY)    ; y
gfxRefreshXY.1:
  call gfxIsScreenModeMSX2
  jr nc, gfxRefreshXY.2  ; if MSX2 and screen mode above 3
    __call_bios BIOS_SCALXY ; BC = X, DE = Y
    __call_bios BIOS_MAPXYC ; in BC = X, DE = Y
	ret
gfxRefreshXY.2:
  ld ix, BIOS_SCALXY2 ; BC = X, DE = Y
  call BIOS_EXTROM
  ld ix, BIOS_MAPXYC2 ; in BC = X, DE = Y
  jp BIOS_EXTROM

;---------------------------------------------------------------------------------------------------------
; gfxGetXY
; get current screen location
; out BC = x
;     DE = y
;---------------------------------------------------------------------------------------------------------

gfxGetXY:
  ld bc, (BIOS_GRPACX)    ; x
  ld de, (BIOS_GRPACY)    ; y
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetScreenHeight
; get screen height
; out a = screen height
;---------------------------------------------------------------------------------------------------------

gfxGetScreenHeight:
  push hl
  push de
    ld hl, BIOS_SCR_SIZE_Y
    ld a, (BIOS_SCRMOD)
    ld d, 0
    ld e, a
    add hl, de
    ld a, (hl)
	ld (GFX_MAX_Y), a
  pop de
  pop hl
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetScreenWidth
; get screen width
; out a = screen height
;---------------------------------------------------------------------------------------------------------

gfxGetScreenWidth:
  push hl
  push de
    ld hl, BIOS_SCR_SIZE_X
    ld a, (BIOS_SCRMOD)
    ld d, 0
    ld e, a
    add hl, de
    ld a, (hl)
	ld (GFX_MAX_X), a
  pop de
  pop hl
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetSpriteSize
; get sprite data size
; out a = sprite data size
;---------------------------------------------------------------------------------------------------------

gfxGetSpriteSize:
  push bc
    ld bc, 0x0808
    ld a, (BIOS_RG1SAV)  		; bit 0 = double size, bit 1 = sprite size (0=8 pixels, 1=16 pixels)
    bit 1, a
    jr z, gfxGetSpriteSize.1
      ld bc, 0x1010

gfxGetSpriteSize.1:
    bit 0, a
    jr z, gfxGetSpriteSize.2
      sll b

gfxGetSpriteSize.2:
    ld (GFX_SPRITE_SIZE_DAT), bc
    ld a, c
  pop bc
  ret


if defined GFX_FAST and defined PAINT

;---------------------------------------------------------------------------------------------------------
; gfxUp
; move screen current location up
; out: carry if off screen
;---------------------------------------------------------------------------------------------------------

gfxUp:
  push bc
    ld hl, 0
	ld de, (BIOS_GRPACY)
	or a
	sbc hl, de
	jr z, gfxUp.1
    dec de
	ld (BIOS_GRPACY), de
    call gfxRefreshXY
	scf
	ccf
	jr gfxUp.2
gfxUp.1:
	scf
gfxUp.2:
  pop bc
  ret

;---------------------------------------------------------------------------------------------------------
; gfxDown
; move screen current location down
; out: carry if off screen
;---------------------------------------------------------------------------------------------------------

gfxDown:
  push bc
    ld a, (GFX_MAX_Y)
    ld l, a
    ld h, 0
	ld de, (BIOS_GRPACY)
	or a
    sbc hl, de
	jr z, gfxDown.1
    inc de
	ld (BIOS_GRPACY), de
    call gfxRefreshXY
	scf
	ccf
	jr gfxDown.2
gfxDown.1:
	scf
gfxDown.2:
  pop bc
  ret

;---------------------------------------------------------------------------------------------------------
; gfxLeft
; move screen current location left
; out: carry if off screen
;---------------------------------------------------------------------------------------------------------

gfxLeft:
  push bc
    ld hl, 0
	ld de, (BIOS_GRPACX)
	or a
	sbc hl, de
	jr z, gfxLeft.1
    dec de
	ld (BIOS_GRPACX), de
    call gfxRefreshXY
	scf
	ccf
	jr gfxLeft.2
gfxLeft.1:
	scf
gfxLeft.2:
  pop bc
  ret

;---------------------------------------------------------------------------------------------------------
; gfxRight
; move screen current location right
; out: carry if off screen
;---------------------------------------------------------------------------------------------------------

gfxRight:
  push bc
    ld a, (GFX_MAX_X)
    ld l, a
    ld h, 0
	ld de, (BIOS_GRPACX)
	or a
    sbc hl, de
	jr z, gfxRight.1
    inc de
	ld (BIOS_GRPACX), de
    call gfxRefreshXY
	scf
	ccf
	jr gfxRight.2
gfxRight.1:
	scf
gfxRight.2:
  pop bc
  ret

endif

;---------------------------------------------------------------------------------------------------------
; gfxPushXY
; push current screen location
;---------------------------------------------------------------------------------------------------------

gfxPushXY:
  di
    pop  ix
    ld iy, (BIOS_GRPACX)    ; x
    push iy
    ld iy, (BIOS_GRPACY)    ; y
    push iy
    push ix
  ei
  ret

;---------------------------------------------------------------------------------------------------------
; gfxPopXY
; pop current screen location
; out BC = x
;     DE = y
;---------------------------------------------------------------------------------------------------------

gfxPopXY:
  pop  ix
  pop  de     ; y
  pop  bc     ; x
  push ix
  jp gfxSetXY

;---------------------------------------------------------------------------------------------------------
; gfxSetForeColor
; set current foreground color
; A = color
;---------------------------------------------------------------------------------------------------------

gfxSetForeColor:
  ld (BIOS_FORCLR), a     ; foreground color
  ld (BIOS_ATRBYT), a
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetForeColor
; get current foreground color
; out A = color
;---------------------------------------------------------------------------------------------------------

gfxGetForeColor:
  ld a, (BIOS_FORCLR)     ; foreground color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetBackColor
; set current background color
; A = color
;---------------------------------------------------------------------------------------------------------

gfxSetBackColor:
  ld (BIOS_BAKCLR), a     ; foreground color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetBackColor
; get current background color
; out A = color
;---------------------------------------------------------------------------------------------------------

gfxGetBackColor:
  ld a, (BIOS_BAKCLR)     ; foreground color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetBorderColor
; set current border color
; A = color
;---------------------------------------------------------------------------------------------------------

gfxSetBorderColor:
  ld (BIOS_BDRCLR), a     ; border color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetBorderColor
; get current border color
; out A = color
;---------------------------------------------------------------------------------------------------------

gfxGetBorderColor:
  ld a, (BIOS_BDRCLR)     ; border color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetBorderFill
; set fill border color
; A = color
;---------------------------------------------------------------------------------------------------------

gfxSetBorderFill:
  ld (BIOS_BDRATR), a     ; border color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxGetBorderFill
; get fill border color
; out A = color
;---------------------------------------------------------------------------------------------------------

gfxGetBorderFill:
  ld a, (BIOS_BDRATR)     ; border color
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetColor
; set current color (foreground, background and border)
;---------------------------------------------------------------------------------------------------------

gfxSetColor:
  ld a, (BIOS_SCRMOD)
  bit 3, a
  jp nz, BIOS_CHGCLR2 ; change VDP colors - msx2
  bit 2, a
  jp nz, BIOS_CHGCLR2 ; change VDP colors - msx2
  jp BIOS_CHGCLR      ; change VDP colors
  ; __call_bios BIOS_SETATR ; change the pixel color
  ;ret

;---------------------------------------------------------------------------------------------------------
; gfxSetPixel
; set pixel in current position to current foreground color
;---------------------------------------------------------------------------------------------------------

gfxSetPixel:
  call gfxIsScreenModeMSX2
  jr nc, gfxSetPixel.1  ; if MSX2 and screen mode above 3
    __call_bios BIOS_SETC
	ret
gfxSetPixel.1:
  ld ix, BIOS_SETC2
  jp BIOS_EXTROM

;---------------------------------------------------------------------------------------------------------
; gfxGetPixel
; get pixel color in current position
; out A = pixel color
;---------------------------------------------------------------------------------------------------------

gfxGetPixel:
  call gfxIsScreenModeMSX2
  jr nc, gfxGetPixel.1  ; if MSX2 and screen mode above 3
    __call_bios BIOS_READC
	ret
gfxGetPixel.1:
  ld ix, BIOS_READC2
  jp BIOS_EXTROM

if defined LINE

;---------------------------------------------------------------------------------------------------------
; gfxDrawLine
; plot a line from current position to informed destination
; in BC = destination x
;    DE = destination y
; https://en.wikipedia.org/wiki/Bresenham%27s_line_algorithm#Algorithm_for_integer_arithmetic
; https://rosettacode.org/wiki/Bitmap/Bresenham%27s_line_algorithm#C
;---------------------------------------------------------------------------------------------------------
;void line(int x0, int y0, int x1, int y1) {
;  int dx = abs(x1-x0), sx = x0<x1 ? 1 : -1;
;  int dy = abs(y1-y0), sy = y0<y1 ? 1 : -1;
;  int err = (dx>dy ? dx : -dy)/2, e2;
;  for(;;){
;    setPixel(x0,y0);
;    if (x0==x1 && y0==y1) break;
;    e2 = err;
;    if (e2 >-dx) { err -= dy; x0 += sx; }
;    if (e2 < dy) { err += dx; y0 += sy; }
;  }
;}

gfxDrawLine:
if not defined GFX_FAST
  ld hl, (BIOS_GRPACX)
  ld (BIOS_GXPOS), hl
  ld hl, (BIOS_GRPACY)
  ld (BIOS_GYPOS), hl
  __call_basic BASIC_SUB_LINE
  ret

else

  ld (GFX_TEMP2), bc        ; x
  ld (GFX_TEMP3), de        ; y

  ld hl, (BIOS_GRPACX)      ; x0
  ld de, (GFX_TEMP2)        ; x
  ;or a
  ;sbc hl, de
  ;jp po, gfxLine.1            ; x0 >= x? else jump to endif
  call MATH.COMP.S16
  jp c, gfxLine.1             ; x0 < x? jump
    or a
    sbc hl, de
    ld (GFX_TEMP4), hl        ; dx = x0 - x
    ld hl, 0xffff
    ld (GFX_TEMP5), hl        ; sx = -1
    jr gfxLine.2

gfxLine.1:
  ld de, (BIOS_GRPACX)      ; x0
  ld hl, (GFX_TEMP2)        ; x
  or a
  sbc hl, de
  ld (GFX_TEMP4), hl        ; dx = x - x0
  ld hl, 1
  ld (GFX_TEMP5), hl        ; sx = 1

gfxLine.2:
  ld hl, (BIOS_GRPACY)      ; y0
  ld de, (GFX_TEMP3)        ; y
  ;or a
  ;sbc hl, de
  ;jp po, gfxLine.3           ; y0 >= y? else, jump to endif
  call MATH.COMP.S16
  jp c, gfxLine.3             ; y0 < y? jump
    or a
    sbc hl, de
    ld (GFX_TEMP6), hl        ; dy =  y0 - y
    ld hl, 0xffff
    ld (GFX_TEMP7), hl        ; sy = -1
    jr gfxLine.4

gfxLine.3:
  ld de, (BIOS_GRPACY)      ; y0
  ld hl, (GFX_TEMP3)        ; y
  or a
  sbc hl, de
  ld (GFX_TEMP6), hl        ; dy = y - y0
  ld hl, 1
  ld (GFX_TEMP7), hl        ; sy = 1

gfxLine.4:
  ld hl, (GFX_TEMP6)       ; dy
  ld de, 0
  call MATH.COMP.S16
  jp z, gfxLine.h          ; dy = 0?

  ld hl, (GFX_TEMP4)       ; dx
  ld de, 0
  call MATH.COMP.S16
  jp z, gfxLine.v          ; dx = 0?

  ld de, (GFX_TEMP4)       ; dx
  ld hl, (GFX_TEMP6)       ; dy
  or a
  adc hl, de
  dec hl
  dec hl
  ld (GFX_TEMP9), hl       ; dxy = dx + dy

  ld de, (GFX_TEMP4)       ; dx
  ld hl, (GFX_TEMP6)       ; dy
  ;or a
  ;sbc hl, de
  ;jp pe, gfxLine.5         ; dy < dx? else, jump to endif
  call MATH.COMP.S16
  jp z, gfxLine.5          ; dy = dx? jump
  jp nc, gfxLine.5         ; dy > dx? jump
    ;ld hl, 0
    ld de, (GFX_TEMP6)     ; dy
	or a
    srl d
    rr e                   ; dy / 2
	;or a
	;sbc hl, de             ; -dy
    ld (GFX_TEMP8), de     ; err = -dy / 2
    jr gfxLine.loop

gfxLine.5:
  ld hl, (GFX_TEMP4)       ; dx
  or a
  srl h
  rr l
  ld (GFX_TEMP8), hl       ; err = dx/2

gfxLine.loop:
  call gfxSetPixel

  ld hl, (GFX_TEMP9)         ; dxy
  ld de, 0
  call MATH.COMP.S16
  jp nc, gfxLine.loop.0      ; dxy > 0? jump
  ;jp z, gfxLine.loop.0      ; dxy = 0? jump

  ret

gfxLine.loop.0:
  ld hl, 0
  ld de, (GFX_TEMP4)     ; dx
  or a
  sbc hl, de             ; -dx
  ld de, (GFX_TEMP8)     ; e2 = err
  push de
    ;or a
    ;sbc hl, de
    ;jp pe, gfxLine.loop.1  ; if -dx < e2, else jump to endif
	call MATH.COMP.S16
	jp z, gfxLine.loop.1   ; -dx = e2? jump
	jp nc, gfxLine.loop.1  ; -dx > e2? jump
      ld hl, (GFX_TEMP8)   ; err
      ld de, (GFX_TEMP6)   ; dy
      or a
      sbc hl, de
      ld (GFX_TEMP8), hl   ; err -= dy

      ld hl, (GFX_TEMP9)   ; dxy
      dec hl
      ld (GFX_TEMP9), hl   ; dxy -= 1

      ld hl, (BIOS_GRPACX)
      ld de, (GFX_TEMP5)
      or a
      adc hl, de
      ld (BIOS_GRPACX), hl  ; x0 += sx

gfxLine.loop.1:
  pop hl                    ; e2
  ld de, (GFX_TEMP6)        ; dy
  ;or a
  ;sbc hl, de
  ;jp pe, gfxLine.loop.2     ; if e2 < dy, else jump to endif
  call MATH.COMP.S16
  jp z, gfxLine.loop.2      ; e2 = dy? jump
  jp nc, gfxLine.loop.2     ; e2 > dy? jump
    ld hl, (GFX_TEMP8)      ; err
    ld de, (GFX_TEMP4)      ; dx
    or a
    adc hl, de
    ld (GFX_TEMP8), hl      ; err += dx

    ld hl, (GFX_TEMP9)      ; dxy
    dec hl
    ld (GFX_TEMP9), hl      ; dxy -= 1

    ld hl, (BIOS_GRPACY)
    ld de, (GFX_TEMP7)
    or a
    adc hl, de
    ld (BIOS_GRPACY), hl    ; y0 += sy

gfxLine.loop.2:
  call gfxRefreshXY
  jp gfxLine.loop

gfxLine.h:
  ld a, (GFX_TEMP5)        ; sx
  bit 7, a
  jr z, gfxLine.h.1        ; if a is positive
    ld hl, (BIOS_GRPACX)
    ld de, (GFX_TEMP4)
    or a
    sbc hl, de
    ld (BIOS_GRPACX), hl
    call gfxRefreshXY

gfxLine.h.1:
  __call_bios BIOS_FETCHC
  ld hl, (GFX_TEMP4)       ; dx
  inc hl
  call gfxDrawHorLine      ; HL = pixel count
  ret

gfxLine.v:
  ld a, (GFX_TEMP7)        ; sy
  bit 7, a
  jr z, gfxLine.v.1        ; if a is positive
    ld hl, (BIOS_GRPACY)
    ld de, (GFX_TEMP6)
    or a
    sbc hl, de
    ld (BIOS_GRPACY), hl
    call gfxRefreshXY

gfxLine.v.1:
  call gfxSetPixel
  ld hl, (GFX_TEMP6)       ; dy
  inc hl
  ld (GFX_TEMP3), hl
  jp gfxBox.drawVerLine
endif

endif

if defined BOX or defined FBOX or defined BOX_STEP or defined FBOX_STEP

;---------------------------------------------------------------------------------------------------------
; gfxDrawBox
; plot a box from current position to informed destination
; in BC = destination x
;    DE = destination y
;    A  = filled flag (0 = not filled, <>0 = filled)
;---------------------------------------------------------------------------------------------------------

gfxDrawBox:

if not defined GFX_FAST
  ld hl, (BIOS_GRPACX)
  ld (BIOS_GXPOS), hl
  ld hl, (BIOS_GRPACY)
  ld (BIOS_GYPOS), hl
  ld hl, BASIC_SUB_LINEBOX
  or 0
  jr z, gfxDrawBox.1
    call gfxIsScreenModeMSX2
    jr nc, gfxDrawBox.2
    ld hl, BASIC_SUB_LINEBOXFILLED

gfxDrawBox.1:
  push hl
  pop ix
  jp BIOS_CALBAS   ; BIOS_CALSLT


gfxDrawBox.2:
  xor a
  ld hl, BASIC_BUF
  ld (hl), a
  ld ix, BIOS_DOBOXF
  jp BIOS_EXTROM

else
  call gfxAdjustDestXY
  or 0
  jr nz, gfxBox.filled

gfxBox.notFilled:
  call gfxPushXY
    call gfxBox.drawHorLine
    ld hl, (GFX_TEMP2)
    ld de, (BIOS_GRPACX)
    or a
    adc hl, de
    dec hl
    ld (BIOS_GRPACX), hl
    call gfxRefreshXY
    call gfxBox.drawVerLine
  call gfxPopXY
  call gfxBox.drawVerLine
  call gfxBox.drawHorLine
  ret

gfxBox.filled:
  ld hl, (GFX_TEMP3)
  ;inc hl
gfxBox.filled.loop:
  push hl
    ld bc, (BIOS_GRPACX)
	push bc
      call gfxBox.drawHorLine
	pop bc
	ld (BIOS_GRPACX), bc
	ld bc, (BIOS_GRPACY)
	inc bc
	ld (BIOS_GRPACY), bc
    call gfxRefreshXY
  pop hl
  ld de, 1
  or a
  sbc hl, de
  jr nz, gfxBox.filled.loop
  ret

gfxBox.drawHorLine:
  ld hl, (GFX_TEMP2)
  ;inc hl
  call gfxDrawHorLine
  ret

gfxBox.drawVerLine:
  ld hl, (GFX_TEMP3)
  dec hl
gfxBox.drawVerLine.loop:
  push hl
    call gfxDown
	call gfxSetPixel
  pop hl
  ld de, 1
  or a
  sbc hl, de
  jr nz, gfxBox.drawVerLine.loop
  ret

;---------------------------------------------------------------------------------------------------------
; gfxDrawHorLine
; draw a horizontal line
; HL = pixel count
;---------------------------------------------------------------------------------------------------------

gfxDrawHorLine:
  ld a, (BIOS_SCRMOD)
  cp 5
  jr c, gfxDrawHorLine.2 ; if screen mode < 5 then jump
    bit 7, h
    ret nz               ; return if negative
    xor a
    cp h
    jr nz, gfxDrawHorLine.1
    cp l
    ret z                ; return if hl = 0
    call gfxPushXY
gfxDrawHorLine.1:
    push hl
	  call gfxSetPixel
      call gfxRight
    pop hl
    ld de, 1
    sbc hl, de
    jr nz, gfxDrawHorLine.1
    call gfxPopXY
    ret
gfxDrawHorLine.2:
  __call_bios BIOS_NSETCX    ; HL = fill count
  ret

;---------------------------------------------------------------------------------------------------------
; gfxAdjustDestXY
; invert if dest XY is less than current XY position
; BC = dest x
; DE = dest y
;---------------------------------------------------------------------------------------------------------

gfxAdjustDestXY:
  push af
    ld (GFX_TEMP2), bc        ; x
	ld (GFX_TEMP3), de        ; y

    ; verify x againt current position
    ld hl, (BIOS_GRPACX)
	ld de, (GFX_TEMP2)
	and a
	sbc hl, de                ; dx = x1 - x0
	bit 7, h                  ; result is negative?
	jr z, gfxAdjustDestXY.1
	add hl, de
	ld (BIOS_GRPACX), hl
    ex de, hl
    or a
	sbc hl, de
gfxAdjustDestXY.1:
    inc hl
	ld (GFX_TEMP2), hl

    ; verify y againt current position
    ld hl, (BIOS_GRPACY)
	ld de, (GFX_TEMP3)
	and a
	sbc hl, de                ; dy = y1 - y0
	bit 7, h                  ; result is negative?
	jr z, gfxAdjustDestXY.2
	add hl, de
	ld (BIOS_GRPACY), hl
    ex de, hl
    or a
	sbc hl, de
gfxAdjustDestXY.2:
    inc hl
	ld (GFX_TEMP3), hl

	; refresh new position
	call gfxRefreshXY
  pop af
  ret
endif

endif

if defined CIRCLE

;---------------------------------------------------------------------------------------------------------
; gfxDrawCircle
; plot a circle centered in current position
; BC = tracing end x
; DE = tracing end y
; HL = radius
; A  = filled flag (0 = not filled, <>0 = filled)
; https://en.wikipedia.org/wiki/Midpoint_circle_algorithm
; https://rosettacode.org/wiki/Bitmap/Midpoint_circle_algorithm#C
;---------------------------------------------------------------------------------------------------------

gfxDrawCircle:

if not defined GFX_FAST
  bit 7,h
  ret nz                ; return if negative radius
  ld (BIOS_GXPOS), hl   ; circle ray
  ld de, (BIOS_GXPOS)
  ld bc, (BIOS_GRPACY)
  ld (BIOS_GYPOS), bc
  xor a
  ;cp h
  ;jr nz, gfxDrawCircle.1
  ;cp l
  ;ret z
gfxDrawCircle.1:
  ld hl, BASIC_BUF
  ld (hl), a
  ld ix, 0xFFFF
  __call_basic BASIC_SUB_CIRCLE
  ret
else
  ld (GFX_TEMP),  hl   ; radius
  ld de, 0
  sbc hl, de
  ret z                ; return if zero radius
  bit 7,h
  ret nz               ; return if negative radius

  call gfxGetXY
  ld (GFX_TEMP1), bc   ; x0
  ld (GFX_TEMP2), de   ; y0

  or 0
  jp nz, gfxCircle.filled

gfxCircle.notFilled:
  ld hl, 1
  ld de, (GFX_TEMP)    ; radius
  or a
  sbc hl, de
  ld (GFX_TEMP3), hl   ; f = 1 - radius

  ld hl, 0
  ld (GFX_TEMP4), hl   ; ddF_x = 0

  ld hl, (GFX_TEMP)
  or a
  adc hl, hl
  ex de, hl
  ld hl, 0
  or a
  sbc hl, de
  ld (GFX_TEMP5), hl   ; ddF_y = -2 * radius

  ld hl, 0
  ld (GFX_TEMP6), hl   ; x = 0

  ld hl, (GFX_TEMP)
  ld (GFX_TEMP7), hl   ; y = radius

  ; plot(x0, y0 + radius)
  ld de, (GFX_TEMP)    ; radius
  ld hl, (GFX_TEMP2)   ; y0
  or a
  adc hl, de
  ex de, hl
  ld bc, (GFX_TEMP1)   ; x0
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0, y0 - radius)
  ld de, (GFX_TEMP)    ; radius
  ld hl, (GFX_TEMP2)   ; y0
  or a
  sbc hl, de
  ex de, hl
  ld bc, (GFX_TEMP1)   ; x0
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 + radius, y0)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP)    ; radius
  or a
  adc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld de, (GFX_TEMP2)   ; y0
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 - radius, y0)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP)    ; radius
  or a
  sbc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld de, (GFX_TEMP2)   ; y0
  call gfxSetXY
  call gfxSetPixel
  jp gfxCircle.notFilled.3

gfxCircle.notFilled.1:
  ld hl, (GFX_TEMP3)   ; f
  bit 7, h
  jr nz, gfxCircle.notFilled.2  ; if( f < 0 ), jump

    ld hl, (GFX_TEMP7)        ; y -= 1
    dec hl
    ld (GFX_TEMP7), hl

    ld hl, (GFX_TEMP5)        ; ddF_y += 2
    inc hl
    inc hl
    ld (GFX_TEMP5), hl

    ld hl, (GFX_TEMP3)        ; f
    ld de, (GFX_TEMP5)        ; ddF_y
    or a
    adc hl, de
    ld (GFX_TEMP3), hl        ; f += ddF_y

gfxCircle.notFilled.2:
  ld hl, (GFX_TEMP6)   ; x
  inc hl
  ld (GFX_TEMP6), hl   ; x++

  ld hl, (GFX_TEMP4)   ; ddF_x += 2
  inc hl
  inc hl
  ld (GFX_TEMP4), hl

  ld hl, (GFX_TEMP3)   ; f
  ld de, (GFX_TEMP4)   ; ddF_x
  or a
  adc hl, de
  inc hl
  ld (GFX_TEMP3), hl   ; f += ddF_x + 1

  ; plot(x0 + x, y0 + y)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP6)   ; x
  or a
  adc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP7)   ; y
  or a
  adc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 - x, y0 + y)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP6)   ; x
  or a
  sbc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP7)   ; y
  or a
  adc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 + x, y0 - y)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP6)   ; x
  or a
  adc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP7)   ; y
  or a
  sbc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 - x, y0 - y)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP6)   ; x
  or a
  sbc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP7)   ; y
  or a
  sbc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 + y, y0 + x)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP7)   ; y
  or a
  adc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP6)   ; x
  or a
  adc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 - y, y0 + x)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP7)   ; y
  or a
  sbc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP6)   ; x
  or a
  adc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 + y, y0 - x)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP7)   ; y
  or a
  adc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP6)   ; x
  or a
  sbc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

  ; plot(x0 - y, y0 - x)
  ld hl, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP7)   ; y
  or a
  sbc hl, de
  ;push hl
  ;pop bc
  ld b, h
  ld c, l
  ld hl, (GFX_TEMP2)   ; y0
  ld de, (GFX_TEMP6)   ; x
  or a
  sbc hl, de
  ex de, hl
  call gfxSetXY
  call gfxSetPixel

gfxCircle.notFilled.3:
  ld hl, (GFX_TEMP6)           ; x
  ld de, (GFX_TEMP7)           ; y
  or a
  sbc hl, de
  jp c, gfxCircle.notFilled.1  ; while( x < y )
  ld bc, (GFX_TEMP1)   ; x0
  ld de, (GFX_TEMP2)   ; y0
  call gfxSetXY
  ret

gfxCircle.filled
  call gfxCircle.notFilled
  ld hl, (BIOS_BDRATR)
  push hl
	ld hl, (BIOS_FORCLR)
	ld (BIOS_BDRATR), hl
	ld a, 1
	call gfxBorderFill
  pop hl
  ld (BIOS_BDRATR), hl
  ret
endif

endif

if defined PAINT or (defined CIRCLE and defined GFX_FAST)

;---------------------------------------------------------------------------------------------------------
; gfxBorderFill
; Fill current region delimited by border attribute color changing pixels to foreground color
; in: a = fill type (0 = not symmetric, 1 = symmetric)
;---------------------------------------------------------------------------------------------------------

gfxBorderFill:

if not defined GFX_FAST

  jp PAINT_HOOK

else

  ld (GFX_TEMP1), a
  ld hl, (BIOS_GRPACY)
  push hl
    call gfxBorderFill.down
  pop hl
  push hl
    ld (BIOS_GRPACY), hl
    call gfxRefreshXY
    call gfxBorderFill.up
  pop hl
  ld (BIOS_GRPACY), hl
  call gfxRefreshXY
  ret

gfxBorderFill.down:
  call gfxBorderFill.line
  call gfxDown
  ret c
  call gfxGetPixel
  ld hl, BIOS_BDRATR
  cp (hl)
  jr nz, gfxBorderFill.down
  ret

gfxBorderFill.up:
  call gfxBorderFill.line
  call gfxUp
  ret c
  call gfxGetPixel
  ld hl, BIOS_BDRATR
  cp (hl)
  jr nz, gfxBorderFill.up
  ret

gfxBorderFill.line:
  ld de, (BIOS_GRPACX)
  push de
    ld b, 1    ; fill flag
    ld de, 1   ; skip count
    __call_bios BIOS_SCANR
	;inc hl
	;ld (GFX_TEMP2), hl        ; pixel count transversed
  pop de
  push de
    ld (BIOS_GRPACX), de
    call gfxRefreshXY
    ;ld a, (GFX_TEMP1)
    ;cp 0                       ; 0 = not symmetric, 1 = symmetric
	;jr z, gfxBorderFill.line.1
	;ld hl, (GFX_TEMP2)
	;__call_bios BIOS_NSETCX    ; HL = fill count
	;jr gfxBorderFill.line.2
gfxBorderFill.line.1:
    ld b, 1    ; fill flag
    ld de, 0   ; skip count
    __call_bios BIOS_SCANL
gfxBorderFill.line.2:
  pop de
  ld (BIOS_GRPACX), de
  call gfxRefreshXY
  ret

;---------------------------------------------------------------------------------------------------------
; gfxFloadFill
; Fload fill current region changing current pixel color to foreground color
; https://en.wikipedia.org/wiki/Flood_fill
;---------------------------------------------------------------------------------------------------------

gfxFloadFill:
  call gfxGetPixel
  ld (GFX_TEMP), a    ; replacement-color
  call gfxFloadFill.recursive
  ret

gfxFloadFill.recursive:
  ; 1. If target-color is equal to replacement-color, return.
  ld a, (GFX_TEMP)
  ld hl, BIOS_FORCLR
  cp (hl)
  ret z

  ; 2. ElseIf the color of node is not equal to target-color, return.
  call gfxGetPixel
  ld hl, GFX_TEMP
  cp (hl)
  ret nz

  ; 3. Else Set the color of node to replacement-color.
  call gfxSetPixel

  ; 4. Perform Flood-fill (one step to the left of node, target-color, replacement-color).
gfxFloadFill.recursive.left:
  call gfxLeft
  jr c, gfxFloadFill.recursive.right
    call gfxFloadFill.recursive
    call gfxRight

  ;    Perform Flood-fill (one step to the right of node, target-color, replacement-color).
gfxFloadFill.recursive.right:
  call gfxRight
  jr c, gfxFloadFill.recursive.up
    call gfxFloadFill.recursive
    call gfxLeft

  ;    Perform Flood-fill (one step to the up of node, target-color, replacement-color).
gfxFloadFill.recursive.up:
  call gfxUp
  jr c, gfxFloadFill.recursive.down
    call gfxFloadFill.recursive
    call gfxDown

  ;    Perform Flood-fill (one step to the down of node, target-color, replacement-color).
gfxFloadFill.recursive.down:
  call gfxDown
  ret c
    call gfxFloadFill.recursive
    call gfxUp
  ret

endif

endif

if defined SPRITEMODE

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteMode
; set current sprite mode
; A = sprite mode
;     0: Spritesize is 8 by 8 pixels - default value
;     1: Spritesize is 8 by 8 pixels, magnified to 16 by 16 pixels
;     2: Spritesize is 16 by 16 pixels
;     3: Spritesize is 16 by 16 pixels, magnified to 32 by 32 pixels
; RG1SAV bit 0 = magnify sprite (double size)
; RG1SAV bit 1 = sprite size (0=8 pixels, 1=16 pixels)
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteMode:
  and 3                        ; keeps only bits 0 and 1 from A
  ld b, a
  di
    ld a, (BIOS_RG1SAV)         ; get copy from register #1 of VDP
    and 0xFC                    ; clear bits 0 and 1 from A
    or b                        ; put parameter to A (bits 0 and 1)
    ld (BIOS_RG1SAV), a         ; restore to register #1 of VDP
    ld b, a                     ; value to write
    ld c, 1                     ; register number to write
    call gfxWRTVDP     ; write register to VDP
    call gfxCLRSPR     ; clear sprites
  ei
  call gfxGetSpriteSize
  jp gfxFillSpriteCollisionTable

endif

if defined SPRITE or defined COLOR_SPRITE or defined PUT_SPRITE_COLOR or defined PUT_SPRITE_STEP_COLOR or defined PUT_SPRITE_COLOR_PATNUM or defined PUT_SPRITE_STEP_COLOR_PATNUM or defined LOAD_SET

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteColorInt
; A = sprite number
; BC = color number
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteColorInt:
  ld b, a         ; save sprite number
  call gfxCALATR    ; get sprite attribute table address
  inc hl
  inc hl
  inc hl
  ld a, c         ; color
  ;call gfxSetSpriteColor.Adjust
  call gfxWRTVRM

  ld a, (BIOS_SCRMOD)
  cp 3
  ret c           ; if screen mode < 3, do not adjust sprite multicolor

  ld a, b         ; recover sprite number
  call gfxGetSpriteColorTable

  ld a, c         ; color
  ;call gfxSetSpriteColor.Adjust
  ld b, 16        ; array of 16 bytes

gfxSetSpriteColorInt.1:
  push bc
    call gfxWRTVRM
  pop bc
  inc hl
  djnz gfxSetSpriteColorInt.1
  ret

;gfxSetSpriteColor.Adjust:
;  cp 0x0f
;  ret z
;  ret c
;  srl a
;  srl a
;  srl a
;  srl a
;  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteColorStr
; A = sprite number
; DE = address to color byte array
; BC = color byte array size
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteColorStr:
  ld b, a         ; save sprite number
  push bc
  push de
    call gfxCALATR    ; get sprite attribute table
  pop de
  pop bc

  ld a, c         ; byte array zero length?
  or a
  ret z

  inc hl
  inc hl
  inc hl
  ld a, (de)      ; color
  ;call gfxSetSpriteColor.Adjust
  call gfxWRTVRM

  ld a, (BIOS_SCRMOD)
  cp 3
  ret c           ; if screen mode < 3, do not adjust sprite mode 2

  ld a, b         ; recover sprite number
  call gfxGetSpriteColorTable

  ld b, 16        ; array of 16 bytes (color table)
  ld a, c         ; size of color array
  push de
  pop iy          ; save array start

gfxSetSpriteColorStr.1:
  push af
    ld a, (de)
    ;call gfxSetSpriteColor.Adjust
    call gfxWRTVRM
    inc hl
    inc de
  pop af
  dec a
  jr nz, gfxSetSpriteColorStr.2
    ld a, c      ; recover array size
	push iy
	pop de       ; recover array start

gfxSetSpriteColorStr.2:
  djnz gfxSetSpriteColorStr.1
  ret



;---------------------------------------------------------------------------------------------------------
; gfxGetSpriteColorTable
; A = sprite number
; HL = address to color table
;---------------------------------------------------------------------------------------------------------

gfxGetSpriteColorTable:
  push af
  push de
    ld h, 0
    ld l, a         ; recover sprite number
    add hl, hl
    add hl, hl
    add hl, hl
    add hl, hl      ; multiply by 16 (shift left 4)
    push hl
      xor a
      call gfxCALATR    ; get sprite attribute table address
    pop de
    add hl, de
    xor a
    ld de, 512
    sbc hl, de      ; address of color table from sprite multicolor
  pop de
  pop af
  ret

endif

if defined PUT_SPRITE or defined PUT_SPRITE_STEP or defined PUT_SPRITE_COLOR or defined PUT_SPRITE_STEP_COLOR or defined PUT_SPRITE_COLOR_PATNUM or defined PUT_SPRITE_STEP_COLOR_PATNUM or defined PUT_SPRITE_PATNUM or defined PUT_SPRITE_STEP_PATNUM

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteXY
; A = sprite number
; BC = XY
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteXY:
  push bc
    call gfxCALATR  ; get sprite attribute table address
  pop bc
  ld a, c                  ; y
  call gfxWRTVRM
  inc hl
  ld a, b                  ; x
  call gfxWRTVRM
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSpriteStepCheck
; BC = XY
;---------------------------------------------------------------------------------------------------------

gfxSpriteStepCheck:
  push hl
  push de
  push bc

    ld de, (GFX_SPRITE_SIZE_DAT)
    ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
    and 7                      ; clear touched and collided flags
    ld (GFX_SPRITE_FLAGS), a

    bit 0, a
    jr z, gfxSpriteStepCheck.corners

gfxSpriteStepCheck.limits:
      ld a, (GFX_MAX_X)
      sub e                    ; sprite size
      cp b                     ; jump if x > max_x?
      jr c, gfxSpriteStepCheck.limits.touched

      ld a, (GFX_MAX_Y)
      sub e                    ; sprite size
      cp c                     ; jump if y > max_y?
      jr c, gfxSpriteStepCheck.limits.touched

      jr gfxSpriteStepCheck.corners

gfxSpriteStepCheck.limits.touched:
      ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
      or 8                       ; set limit touched flag
      ld (GFX_SPRITE_FLAGS), a

gfxSpriteStepCheck.corners:
    ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
    and 2+4                    ; check walls or hotspots
    jr z, gfxSpriteStepCheck.end

      ld (GFX_TEMP10), bc
      call gfxSpriteStepCheck.corner

      ld a, (GFX_SPRITE_SIZE_DAT)
      add a, b
      ld b, a
      call gfxSpriteStepCheck.corner

      ld a, (GFX_SPRITE_SIZE_DAT)
      add a, c
      ld c, a
      call gfxSpriteStepCheck.corner

      ld bc, (GFX_TEMP10)
      ld a, (GFX_SPRITE_SIZE_DAT)
      add a, c
      ld c, a
      call gfxSpriteStepCheck.corner

gfxSpriteStepCheck.end:
  pop bc
  pop de
  pop hl
  ret

gfxSpriteStepCheck.corner:
  call gfxGetTileFromXY
  ld d, a                ; tile to be searched

  ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  bit 1, a
  call nz, gfxSpriteStepCheck.walls

  ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  bit 2, a
  call nz, gfxSpriteStepCheck.hotspots
  ret

gfxSpriteStepCheck.walls:
  ld hl, (GFX_SPRITE_WALLS)
  ld e, 16   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  jp gfxSpriteStepCheck.search

gfxSpriteStepCheck.hotspots:
  ld hl, (GFX_SPRITE_HOTSPOTS)
  ld e, 32   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  call gfxSpriteStepCheck.search

  ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  bit 5, a
  ret z

  ld a, (GFX_TEMP1)
  ld (GFX_SPRITE_HOTSPOT_TILE), a
  ret

; d = tile, hl = search table, e = flag
gfxSpriteStepCheck.search:
  ld a, h
  or l
  ret z

  ld a, d   ; search for this tile
  push bc
     ld c, (hl)
     inc hl
     ld b, (hl)
     inc hl
     cpir             ; inc HL searching for A until BC=0 (Z flag settled if found)
  pop bc
  ret nz

gfxSpriteStepCheck.search.found:
  ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
  or e                       ; set touched flag
  ld (GFX_SPRITE_FLAGS), a
  ret

; b = x, c = y
gfxGetTileFromXY:
  push bc
    ld a, c   ; y
    srl b     ; divide by 2
    srl b     ; divide by 4
    srl b     ; divide by 8
    ld c, b
    srl a
    srl a
    srl a
    ld b, a
    call gfxGetScreenTile   ; b = y, c = x, a = tile
  pop bc
  ret

endif

if defined SPRITE or defined PUT_SPRITE or defined PUT_SPRITE_PATNUM or defined PUT_SPRITE_STEP_PATNUM or defined PUT_SPRITE_COLOR_PATNUM or defined PUT_SPRITE_STEP_COLOR_PATNUM

;---------------------------------------------------------------------------------------------------------
; gfxSetSpritePattern
; A = sprite number
; BC = pattern number
;---------------------------------------------------------------------------------------------------------

gfxSetSpritePattern:
  push bc
    call gfxCALATR   ; get sprite attribute table address
  pop bc
  inc hl
  inc hl
  ld a, (BIOS_RG1SAV)  		; bit 0 = double size, bit 1 = sprite size (0=8 pixels, 1=16 pixels)
  bit 1, a
  jr z, gfxSetSpritePattern.1
    sll c
    sll c
gfxSetSpritePattern.1:
  ld a, c                   ; pattern number
  call gfxWRTVRM
  ret

endif

if defined SPRITE

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteData
; HL = point to sprite data as a string of 8 or 32 characters according the sprites size (8x8 or 16x16)
; A = sprite number
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteData:
  push AF
    push HL
      call gfxCALPAT    ; get sprite pattern data address
      ex DE, HL
      call gfxGSPSIZ    ; return in 'a' sprite default size
    pop HL
    ld b, 0
	cp c
	jr z,  gfxSetSpriteData.1
	jr nc, gfxSetSpriteData.1
      ld c, a
gfxSetSpriteData.1:
	push bc
	  ld c, a
      xor a
	  call gfxFILVRM
	pop bc
  pop AF
  call gfxLDIRVM
  ret

endif

if defined GFX_SPRITES

;---------------------------------------------------------------------------------------------------------
; gfxInitSprites
; initialises all sprites
;---------------------------------------------------------------------------------------------------------

gfxInitSprites:
  call gfxCLRSPR
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetSpriteAttrs
; set sprite default x, y, pattern and color
; A = sprite number
;---------------------------------------------------------------------------------------------------------

gfxSetSpriteAttrs:
  push af
    call gfxCALATR
    ld a, (BIOS_GRPACY)   ; y
	call gfxWRTVRM
	inc hl
    ld a, (BIOS_GRPACX)   ; x
	call gfxWRTVRM
	inc hl
  pop af                  ; pattern
  call gfxWRTVRM
  inc hl
  ld a, (BIOS_FORCLR)     ; color
  call gfxWRTVRM
  ret

endif

if defined EXIST_DATA_SET

;---------------------------------------------------------------------------------------------------------
; gfxIsTileMode
;---------------------------------------------------------------------------------------------------------

gfxIsTileMode:
    ld a, (BIOS_SCRMOD)
    cp 2
    ret z
    cp 4
    ret

;---------------------------------------------------------------------------------------------------------
; gfxClearTileScreen
;---------------------------------------------------------------------------------------------------------

gfxClearTileScreen:
    ld a, 0x20           ; space
    ld bc, 768
    ld hl, (BIOS_GRPNAM)
    call gfxFILVRM

    ld de, 0x0020
    call gfxSetTileDefaultColor
    ld de, 0x0120
    call gfxSetTileDefaultColor
    ld de, 0x0220
    jp gfxSetTileDefaultColor

gfxSetTileDefaultColor
    call gfxGetTileColorAddr
    ex de, hl
    call gfxGetTileDefaultColor
    ld bc, 8
    jp gfxFILVRM

gfxGetTileDefaultColor:
    ld a, (BIOS_FORCLR)
    sla a
    sla a
    sla a
    sla a
    push hl
      ld hl, BIOS_BAKCLR
      or (hl)
    pop hl
    ret

; set tile color
; HL = tile color pointer
; BC = tile color size
; DE = tile number

;---------------------------------------------------------------------------------------------------------
; gfxSetTileData
; set tile data
; HL = tile data pointer
; BC = tile data size
; DE = tile number
; A = flip (0=no, 1=yes)
;---------------------------------------------------------------------------------------------------------

gfxSetTileData:
  or a
  jr nz, gfxSetTileDataFlip

gfxSetTileDataNoFlip:
  call gfxGetTileDataAddr
  jp gfxLDIRVM

gfxSetTileDataFlip:
  call gfxGetTileDataAddr
  ex de, hl
gfxSetTileDataFlip.Loop:
  ld a, (de)
  call gfxReverseA
  call gfxWRTVRM
  inc de
  inc hl
  dec bc
  ld a, b
  cp c
  jr nz, gfxSetTileDataFlip.Loop
  ret

; in de = tile number
; out de = tile number address
gfxGetTileDataAddr:
  push hl
    ex de, hl
    add hl, hl
    add hl, hl
    add hl, hl   ; tile number * 8
    ld de, (BIOS_GRPCGP)
    add hl, de
    ex de, hl
  pop hl
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetTileColor
; set tile color
; HL = tile color pointer
; BC = tile color size
; DE = tile number
;---------------------------------------------------------------------------------------------------------

gfxSetTileColor:
  call gfxGetTileColorAddr
  jp gfxLDIRVM

; in de = tile number
; out de = tile number address
gfxGetTileColorAddr:
  push hl
    ex de, hl
    add hl, hl
    add hl, hl
    add hl, hl   ; tile number * 8
    ld de, (BIOS_GRPCOL)
    add hl, de
    ex de, hl
  pop hl
  ret

;---------------------------------------------------------------------------------------------------------
; gfxSetScreenTile
; set screen tile at x,y
; C = x
; B = y
; a = tile number
;---------------------------------------------------------------------------------------------------------

gfxSetScreenTile:
  ex af, af'
  ld a, 23
  cp b
  ret c
  ld a, 31
  cp c
  ret c
  ;push bc
  ;  ld h, 0
  ;  ld l, b      ; slow y * 32
  ;  ld bc, 32
  ;  call MATH.MULT.16
  ;pop bc
  ld h, b
  sra h
  sra h
  sra h
  ld l, b
  sla l
  sla l
  sla l
  sla l
  sla l            ; fast y * 32
  ld d, 0
  ld e, c        ; x
  add hl, de
  ld de, (BIOS_GRPNAM)
  add hl, de
  ex af, af'
  jp gfxWRTVRM

endif

if defined gfxGetTileFromXY or defined EXIST_DATA_SET

;---------------------------------------------------------------------------------------------------------
; gfxGetScreenTile
; get screen tile at x,y
; C = x
; B = y
; a = tile number
;---------------------------------------------------------------------------------------------------------

gfxGetScreenTile:
  ld a, 23
  cp b
  ret c
  ld a, 31
  cp c
  ret c
  ld h, b
  sra h
  sra h
  sra h
  ld l, b
  sla l
  sla l
  sla l
  sla l
  sla l            ; fast y * 32
  ld d, 0
  ld e, c        ; x
  add hl, de
  ld de, (BIOS_GRPNAM)
  add hl, de
  jp gfxRDVRM

endif

;---------------------------------------------------------------------------------------------------------
; Sprite collision table routines
;---------------------------------------------------------------------------------------------------------

gfxInitSpriteCollisionTable:
  ld bc, 128
  call memory.alloc
  ld (GFX_SPRITE_COLLISION), ix
  ret

; copy sprite attribute table to ram
gfxFillSpriteCollisionTable:
  ld hl, (BIOS_ATRBAS)          ; source: attribute table
  ld de, (GFX_SPRITE_COLLISION) ; dest: ram
  ld bc, 128                    ; 32*4 = size of attribute table
  call gfxLDIRMV

; pre-calculate each sprite width
gfxCalculateSpriteCollisionTable:
  ld ix, (GFX_SPRITE_COLLISION)   ; start of sprites attributes
  ld b, 32                        ; sprite count

gfxCalculateSpriteCollisionTable.start:
  ld a, (GFX_SPRITE_SIZE_DAT)     ; sprite size
  ld c, a
  ld d, 208                       ; Y no-display flag

  ; test screen mode
  ld a, (BIOS_SCRMOD)
  cp 3           ; above screen 3?
  jr c, gfxCalculateSpriteCollisionTable.next
	ld d, 216    ; Y no-display flag

gfxCalculateSpriteCollisionTable.next:
  ; test IC flag (no collision) in color table
  ; test EC flag (early clock, shift 32 dots to the left) in color table
  ; set x1 = x + size, y1 = y + size
  ld a, (ix)   ; y
  cp d         ; test if sprite will not be displayed
  ret z

  add a, c
  ld (ix+2), a ; y1
  ld a, (ix+1) ; x
  add a, c
  ld (ix+3), a ; x1

  inc ix
  inc ix
  inc ix
  inc ix
  djnz gfxCalculateSpriteCollisionTable.next
  ret

; BC = sprite number to check
gfxUpdateSpriteCollisionTable:
  ld (BIOS_TEMP), bc
  ld h, b
  ld l, c
  add hl, hl
  add hl, hl
  ld b, h
  ld c, l
  ld hl, (GFX_SPRITE_COLLISION) ; dest: ram
  add hl, bc
  ex de, hl
  ld hl, (BIOS_ATRBAS)          ; source: attribute table
  add hl, bc
  ld bc, 4                      ; 4 = size of attribute table to 1 sprite]
  push de
    call gfxLDIRMV
  pop ix
  ld b, 1                       ; sprite count
  push ix
    call gfxCalculateSpriteCollisionTable.start
  pop iy

  ld a, (GFX_SPRITE_FLAGS)
  and 0xBF                      ; clear collision flag
  ld (GFX_SPRITE_FLAGS), a

  ld a, (BIOS_STATFL)           ; verify if collision occurred
  bit 5, a
  ret z
  jr gfxCheckSpriteCollisionTable.start

; BC = sprite number to check
gfxCheckSpriteCollisionTable:
  ; get target sprite address (iy)
  ld (BIOS_TEMP), bc
  ld h, b
  ld l, c
  ld de, (GFX_SPRITE_COLLISION)
  add hl, hl            ; x4
  add hl, hl
  add hl, de
  push hl
  pop iy

gfxCheckSpriteCollisionTable.start:
  ; start test against others sprites
  ld ix, (GFX_SPRITE_COLLISION)
  ld bc, 0
  ld d, 208      ; Y no-display flag

  ; test screen mode
  ld a, (BIOS_SCRMOD)
  cp 3           ; above screen 3?
  jr c, gfxCheckSpriteCollisionTable.test
    ld d, 216    ; Y no-display flag

gfxCheckSpriteCollisionTable.test:
  ; skip target sprite
  ld a, (BIOS_TEMP)
  cp c
  jr z, gfxCheckSpriteCollisionTable.next

  ; test if x1 > nx and x < nx1 and y1 > ny and y < ny1
  ld a, (ix)   ; ny
  cp d         ; test if sprite will not be displayed
  ret z        ; return false

  cp (iy+2)    ; y1
  jr nc, gfxCheckSpriteCollisionTable.next

  ld a, (ix+1) ; nx
  cp (iy+3)    ; x1
  jr nc, gfxCheckSpriteCollisionTable.next

  ld a, (iy+1) ; x
  cp (ix+3)    ; nx1
  jr nc, gfxCheckSpriteCollisionTable.next

  ld a, (iy)   ; y
  cp (ix+2)    ; ny1
  jr nc, gfxCheckSpriteCollisionTable.next

  ; if so, save collider sprite
  ld a, c
  ld (GFX_SPRITE_COLLIDER), a

  ld a, (GFX_SPRITE_FLAGS)
  or 0x40                      ; set collision flag
  ld (GFX_SPRITE_FLAGS), a
  ret                          ; return true

  ; else, next
gfxCheckSpriteCollisionTable.next:
  inc c
  inc ix
  inc ix
  inc ix
  inc ix
  ld a, 32
  cp c
  ret z                           ; return false
  jr gfxCheckSpriteCollisionTable.test



;---------------------------------------------------------------------------------------------------------
; VDP / VRAM support routines
;---------------------------------------------------------------------------------------------------------

; WRITE TO VDP
; in b = data
;    c = register number
;    a = register number
gfxWRTVDP:
  bit 7, a
  ret nz                ; is negative? read only
  cp 8
  ret z                 ; is register 8? then status register 0 (read only)
  jr nc, gfxWRTVDP.1    ; is > 8? then control registers numbers added 1
  jr gfxWRTVDP.3
gfxWRTVDP.1:
  ld ix, BIOS_SCRMOD
  bit 3, (ix)
  jr nz, gfxWRTVDP.2
  bit 2, (ix)
  jr nz, gfxWRTVDP.2
  ret
gfxWRTVDP.2:
  dec a
  ld c, a
gfxWRTVDP.3:
  ;ld ix, BIOS_SCRMOD
  ;bit 3, (ix)
  ;jp nz, BIOS_NWRVDP    ; msx 2
  ;bit 2, (ix)
  ;jp nz, BIOS_NWRVDP    ; msx 2
  jp BIOS_WRTVDP        ; msx 1

; READ FROM VDP
; in  a = register number
; out a = data
gfxRDVDP:
  bit 7, a
  jr nz, gfxRDVDP.1     ; is negative? then status register 1 to 9
  cp 8
  jr z,  gfxRDVDP.2     ; is register 8? then status register 0
  cp 9
  jr nc, gfxRDVDP.3     ; is >= 9? then control registers numbers added 1
    ld hl, BIOS_RG0SAV  ; else is correct control registers numbers
    jr gfxRDVDP.4
gfxRDVDP.1:
  ld ix, BIOS_SCRMOD
  bit 3, (ix)
  jr nz, gfxRDVDP.1.a
  bit 2, (ix)
  jr nz, gfxRDVDP.1.a
  xor a
  ret
gfxRDVDP.1.a:
  neg
  jp BIOS_NRDVDP   ;BIOS_VDPSTA
gfxRDVDP.2:
  ld a, (BIOS_STATFL)
  ret
  ;xor a
  ;jp BIOS_VDPSTA
gfxRDVDP.3:
  ld hl, BIOS_RG8SAV-9
gfxRDVDP.4:
  ld d, 0
  ld e, a
  add hl,de
  ld a, (hl)
  ret

; in: A=Data byte, BC=Length, HL=VRAM address
gfxFILVRM:
  ld ix, BIOS_SCRMOD
  bit 3, (ix)
  jp nz, BIOS_BIGFIL
  bit 2, (ix)
  jp nz, BIOS_BIGFIL
  jp BIOS_FILVRM

; in: A=Sprite pattern number
; out: HL=Sprite pattern address
gfxCALPAT:
  ld iy, BIOS_SCRMOD
  ld ix, BIOS_CALPAT2
  bit 3, (iy)
  jp nz, BIOS_EXTROM
  bit 2, (iy)
  jp nz, BIOS_EXTROM
  jp BIOS_CALPAT

; in: A=Sprite number
; out: HL=Sprite attribute address
gfxCALATR:
  ld iy, BIOS_SCRMOD
  ld ix, BIOS_CALATR2
  bit 3, (iy)
  jp nz, BIOS_EXTROM
  bit 2, (iy)
  jp nz, BIOS_EXTROM
  jp BIOS_CALATR

; out: A=Bytes in sprite pattern (8 or 32)
gfxGSPSIZ:
  ld iy, BIOS_SCRMOD
  ld ix, BIOS_GSPSIZ2
  bit 3, (iy)
  jp nz, BIOS_EXTROM
  bit 2, (iy)
  jp nz, BIOS_EXTROM
  jp BIOS_GSPSIZ

gfxCLRSPR:
  ld iy, BIOS_SCRMOD
  ld ix, BIOS_CLRSPR2
  bit 3, (iy)
  jp nz, BIOS_EXTROM
  bit 2, (iy)
  jp nz, BIOS_EXTROM
  jp BIOS_CLRSPR

; RAM to VRAM
; in: BC=Length, dest DE=VRAM address, source HL=RAM address
gfxLDIRVM:
  jp BIOS_LDIRVM

; VRAM to RAM
; in: BC=Length, dest DE=RAM address, source HL=VRAM address
gfxLDIRMV:
  jp BIOS_LDIRMV

; WRITE TO VRAM
; in hl = address
;     a = data
gfxWRTVRM:
  ld ix, BIOS_SCRMOD
  bit 3, (ix)
  jp nz, BIOS_NWRVRM
  bit 2, (ix)
  jp nz, BIOS_NWRVRM
  jp BIOS_WRTVRM

; READ FROM VRAM
; in hl = address
; out a = data
gfxRDVRM:
  ld ix, BIOS_SCRMOD
  bit 3, (ix)
  jp nz, BIOS_NRDVRM
  bit 2, (ix)
  jp nz, BIOS_NRDVRM
  jp BIOS_RDVRM

; reverse bits in A
; 8 bytes / 206 cycles
; http://www.retroprogramming.com/2014/01/fast-z80-bit-reversal.html
gfxReverseA:
  ld b,8
  ld l,a
gfxReverseA.loop:
  rl l
  rra
  djnz gfxReverseA.loop
  ret
)";

const char * Assembler::vdp_code = R"(
;---------------------------------------------------------------------------------------------------------
; VDP LIBRARY
;---------------------------------------------------------------------------------------------------------
; References:
; https://github.com/Konamiman/MSX2-Technical-Handbook/blob/master/md/Chapter4b.md
;---------------------------------------------------------------------------------------------------------
;Logical operations
; 0 = IMP: Destination Color = Source Color
; 1 = AND: DC = SC * DC
; 2 = OR:  DC = SC + DC
; 3 = EOR: DC = !SC * DC + SC * !DC
; 4 = NOT: DC = !SC
; 8 = TIMP: if SC=0 then DC=DC else DC=SC
; 9 = TAND: if SC=0 then DC=DC else DC=SCxDC
; 10 = TOR: if SC=0 then DC=DC...
; 11 = TEOR: if...
; 12 = TNOT: if...
;---------------------------------------------------------------------------------------------------------
;Declarations and VDP support routines
;---------------------------------------------------------------------------------------------------------
VDP.READ:   EQU     0006H
VDP.WRITE:  EQU     0007H

VDP.GET.STATUS:
        PUSH    BC
        LD      BC,(VDP.WRITE)
        INC     C
        OUT     (C),A
        LD      A,8FH
        OUT     (C),A
        LD      BC,(VDP.READ)
        INC     C
        IN      A,(C)
        POP     BC
        RET
VDP.WAIT:
        LD      A,2
        CALL    VDP.GET.STATUS
        AND     1
        JP      NZ,VDP.WAIT
        XOR     A
        CALL    VDP.GET.STATUS
        RET

;---------------------------------------------------------------------------------------------------------
;pset (x:H, y:L), color:E, logic-OP:A
;---------------------------------------------------------------------------------------------------------
VDP.PSET:
        DI
        PUSH    AF
        CALL    VDP.WAIT
        LD      BC,(VDP.WRITE)
        INC     C
        LD      A,36
        OUT     (C),A
        LD      A,80H+17
        OUT     (C),A
        PUSH    BC
        INC     C
        INC     C
        XOR     A
        OUT     (C),H
        OUT     (C),A
        OUT     (C),L
        OUT     (C),A
        POP     BC
        LD      A,44
        OUT     (C),A
        LD      A,80H+17
        OUT     (C),A
        INC     C
        INC     C
        OUT     (C),E
        XOR     A
        OUT     (C),A
        LD      E,01010000B
        POP     AF
        OR      E
        OUT     (C),A
        EI
        RET

;---------------------------------------------------------------------------------------------------------
;color_code:A = POINT( x:H, y:L )
;---------------------------------------------------------------------------------------------------------
VDP.POINT:
        DI
        CALL    VDP.WAIT
        LD      A,(VDP.WRITE)
        LD      C,A
        INC     C
        LD      A,32
        OUT     (C),A
        LD      A,80H+17
        OUT     (C),A
        INC     C
        INC     C
        XOR     A
        OUT     (C),H
        OUT     (C),A
        OUT     (C),L
        OUT     (C),A
        DEC     C
        DEC     C
        OUT     (C),A
        LD      A,80H+45
        OUT     (C),A
        LD      A,01000000B
        OUT     (C),A
        LD      A,80H+46
        OUT     (C),A
        CALL    VDP.WAIT
        LD      A,7
        CALL    VDP.GET.STATUS
        PUSH    AF
        XOR     A
        CALL    VDP.GET.STATUS
        POP     AF
        EI
        RET

;---------------------------------------------------------------------------------------------------------
;line (H,L)-(D,E) with color B, log-op A
;---------------------------------------------------------------------------------------------------------
VDP.LINE:
        DI                              ;disable interrupt
        PUSH    AF                      ;save LOGICAL OPERATION
        PUSH    BC                      ;save COLOR
        CALL    VDP.WAIT                ;wait end of command
        LD      A,(VDP.WRITE)
        LD      C,A
        INC     C                       ;C := PORT#1's address
        LD      A,36
        OUT     (C),A
        LD      A,80H+17
        OUT     (C),A                   ;R#17 := 36
        INC     C
        INC     C                       ;C := PORT#3's address
        XOR     A
        OUT     (C),H                   ;DX
        OUT     (C),A
        OUT     (C),L                   ;DY
        OUT     (C),A
        LD      A,H                     ;make DX and DIX
        SUB     D
        LD      D,00000100B
        JP      NC, VDP.LINE.1
        LD      D,00000000B
        NEG
VDP.LINE.1:
        LD      H,A                     ;H := DX , D := DIX
        LD      A,L                     ;make DY and DIY
        SUB     E
        LD      E,00001000B
        JP      NC, VDP.LINE.2
        LD      E,00000000B
        NEG
VDP.LINE.2:
        LD      L,A                     ;L := DY , E := DIY
        CP      H                       ;make Maj and Min
        JP      C, VDP.LINE.3
        XOR     A
        OUT     (C),L                   ;long side
        OUT     (C),A
        OUT     (C),H                   ;short side
        OUT     (C),A
        LD      A,00000001B             ;MAJ := 1
        JP      VDP.LINE.4
VDP.LINE.3:
        XOR     A
        OUT     (C),H                   ;NX
        OUT     (C),A
        OUT     (C),L                   ;NY
        OUT     (C),A
        LD      A,00000000B             ;MAJ := 0
VDP.LINE.4:
        OR      D
        OR      E                       ;A := DIX , DIY , MAJ
        POP     HL                      ;H := COLOR
        OUT     (C),H
        OUT     (C),A
        POP     AF                      ;A := LOGICAL OPERATION
        OR      01110000B
        OUT     (C),A
        LD      A,8FH
        OUT     (C),A
        EI
        RET

;---------------------------------------------------------------------------------------------------------
; srch (x:H, y:L, color:E, arg(reg#45) : A)
;    returns: Z  (not found)
;             NZ (A := X)
;---------------------------------------------------------------------------------------------------------
VDP.SRCH:
        DI                              ;disable interrupt
        PUSH    AF                      ;save arg
        CALL    VDP.WAIT
        LD      A,(VDP.WRITE)
        LD      C,A
        INC     C                       ;C := PORT#1's address
        LD      D,0
        LD      A,32+80H
        OUT     (C),H
        OUT     (C),A                   ;R#32 := H
        INC     A
        OUT     (C),D
        OUT     (C),A                   ;R#33 := 0
        INC     A
        OUT     (C),L
        OUT     (C),A                   ;R#34 := L
        INC     A
        OUT     (C),D
        OUT     (C),A                   ;R#35 := 0
        LD      A,44+80H
        OUT     (C),E
        OUT     (C),A                   ;R#44 := E
        INC     A
        LD      E,A
        POP     AF                      ;A := ARG
        OUT     (C),A
        OUT     (C),E                   ;R#45 := A
        LD      A,01100000B
        OUT     (C),A
        INC     E
        OUT     (C),E                   ;R#46 := SRCH command
VDP.SRCH.LOOP:
        LD      A,2
        CALL    VDP.GET.STATUS
        BIT     0,A
        JP      NZ, VDP.SRCH.LOOP
        LD      E,A
        LD      A,8
        CALL    VDP.GET.STATUS
        LD      D,A
        LD      A,9
        CALL    VDP.GET.STATUS
        LD      A,D
        BIT     4,E
        EI
        RET

;---------------------------------------------------------------------------------------------------------
;  List 4.22  paint routine using PSET and POINT
;       ENTRY:  X:H, Y:L, BORDER COLOR:D, PAINT COLOR:E
;---------------------------------------------------------------------------------------------------------

Q.LENGTH        EQU     256*2*2
MAX.Y           EQU     211

;----- paint main routine -----

VDP.PAINT:
        CALL    POINT
        CP      D
        RET     Z
        CALL    INIT.Q
        LD      (COL),DE
        CALL    PUT.Q
        LD      A,(COL)
        LD      E,A
        XOR     A                       ;logi-OP : PSET
        CALL    VDP.PSET
PAINT0: CALL    GET.Q
        RET     C
        INC     H
        CALL    NZ,PAINT.SUB
        DEC     H
        JP      Z,PAINT1
        DEC     H
        CALL    PAINT.SUB
        INC     H
PAINT1: DEC     L
        LD      A,-1
        CP      L
        CALL    NZ,PAINT.SUB
        INC     L
        INC     L
        LD      A,MAX.Y
        CP      L
        CALL    NC,PAINT.SUB
        JP      PAINT0

;----- check point and pset -----

PAINT.SUB:
        CALL    VDP.POINT
        LD      D,A
        LD      A,(BORD)
        CP      D
        RET     Z
        LD      A,(COL)
        CP      D
        RET     Z
        LD      E,A
        XOR     A
        CALL    VDP.PSET
        CALL    PUT.Q
        RET

;----- init Q.BUFFER pointer -----

INIT.Q:
        PUSH    HL
        LD      HL,Q.BUF
        LD      (Q.TOP),HL
        LD      (Q.BTM),HL
        POP     HL
        RET

;----- put point to Q.BUF (X:H , Y:L) -----

PUT.Q:
        EX      DE,HL
        LD      HL,(Q.TOP)
        LD      BC,Q.BUF+Q.LENGTH+1
        OR      A                       ;clear CARRY
        PUSH    HL
        SBC     HL,BC
        POP     HL
        JP      C,PUT.Q1
        LD      HL,Q.BUF
PUT.Q1:
        LD      (HL),D
        INC     HL
        LD      (HL),E
        INC     HL
        LD      (Q.TOP),HL
        EX      DE,HL
        RET

;----- take point data to D, E -----
;      returns:   NC   H:x, L:y
;                 C    buffer empty

GET.Q:  LD      HL,(Q.BTM)
        LD      BC,(Q.TOP)
        OR      A
        SBC     HL,BC
        JP      NZ,GET.Q0
        SCF
        RET

GET.Q0: LD      HL,(Q.BTM)
        LD      BC,Q.BUF+Q.LENGTH+1
        OR      A
        PUSH    HL
        SBC     HL,BC
        POP     HL
        JP      C,GET.Q1

        LD      HL,Q.BUF
GET.Q1: LD      D,(HL)
        INC     HL
        LD      E,(HL)
        INC     HL
        LD      (Q.BTM),HL
        OR      A
        EX      DE,HL
        RET

;----- work area -----

COL     DS      1
BORD    DS      1
Q.TOP   DS      2
Q.BTM   DS      2
Q.BUF   DS      Q.LENGTH

;---------------------------------------------------------------------------------------------------------
; draw (x:sp+2, y:sp+3, search_color:sp+4, arg:sp+5, pcol:sp+6)
;---------------------------------------------------------------------------------------------------------
)";

#endif // GFX_H_INCLUDED
