#ifndef KERNEL_H_INCLUDED
#define KERNEL_H_INCLUDED

#define isSymbolKeyword(a)    a == symbol_keywords[i][0]
#define existSymbolKeyword(a) symbol_keywords[a][0]
#define getSymbolKeyword(a,b) symbol_keywords[a][b]
#define SYMBOL_NOP 0

#define getVariableName1   ( 65 + (variableCount / 26) )
#define getVariableName2   ( 65 + (variableCount % 26) )

// https://github.com/Konamiman/MSX2-Technical-Handbook
// https://github.com/gseidler/The-MSX-Red-Book

/*
*/

const char * Assembler::bios_functions = R"(
;---------------------------------------------------------------------------------------------------------
; BIOS FUNCTIONS
;---------------------------------------------------------------------------------------------------------

BIOS_CALBAS:    equ 0x0159
BIOS_OUTDO:     equ 0x0018   ; output to current device (i.e. screen)
BIOS_CHPUT:     equ 0x00A2
BIOS_CLS:       equ 0x00C3
BIOS_POSIT:     equ 0x00C6
BIOS_BEEP:      equ 0x00C0
BIOS_CHGET:     equ 0x009F
BIOS_CHSNS:     equ 0x009C
BIOS_INLIN:     equ 0x00B1
BIOS_PINLIN:    equ 0x00AE
BIOS_QINLIN:    equ 0x00B4
BIOS_GTSTCK:    equ 0x00D5
BIOS_GTTRIG:    equ 0x00D8
BIOS_GTPAD:     equ 0x00DB
BIOS_GTPDL:     equ 0x00DE
BIOS_DISSCR:    equ 0x0041
BIOS_ENASCR:    equ 0x0044
BIOS_CHGMOD:    equ 0x005F
BIOS_CHGCLR:    equ 0x0062
BIOS_CLRSPR:    equ 0x0069
BIOS_INITXT:    equ 0x006C    ; init text mode 40 columns
BIOS_INIT32:    equ 0x006F    ; init text mode 32 columns
BIOS_INIGRP:    equ 0x0072
BIOS_INIMLT:    equ 0x0075
BIOS_SETTXT:    equ 0x0078    ; set text mode 40 columns
BIOS_SETT32:    equ 0x007B    ; set text mode 32 columns
BIOS_SETGRP:    equ 0x007E
BIOS_SETMLT:    equ 0x0081
BIOS_CALPAT:    equ 0x0084
BIOS_CALATR:    equ 0x0087
BIOS_GSPSIZ:    equ 0x008A
BIOS_GRPPRT:    equ 0x008D
BIOS_ERAFNK:    equ 0x00CC
BIOS_DSPFNK:    equ 0x00CF
BIOS_TOTEXT:    equ 0x00D2
BIOS_BREAKX:    equ 0x00B7
BIOS_ISCNTC:    equ 0x03FB
BIOS_CHKRAM:    equ 0x0000
BIOS_GICINI:    equ 0x0090
BIOS_WRTPSG:    equ 0x0093
BIOS_REDPSG:    equ 0x0096
BIOS_STRTMS:    equ 0x0099
BIOS_KEYINT:    equ 0x0038
BIOS_CALSLT:    equ 0x001C
BIOS_ENASLT:    equ 0x0024
BIOS_RSLREG:    equ 0x0138
BIOS_SCALXY:    equ 0x010E
BIOS_MAPXYC:    equ 0x0111      ; in BC = X, DE = Y
BIOS_READC:     equ 0x011D      ; out A = color of current pixel
BIOS_SETATR:    equ 0x011A      ; in A = color code
BIOS_SETC:      equ 0x0120      ; set current pixel to color from SETATR
BIOS_NSETCX:    equ 0x0123      ; in HL = pixel fill count
BIOS_SCANR:     equ 0x012C      ; in B=Fill switch, DE=Skip count, out DE=Skip remainder, HL=Pixel count
BIOS_SCANL:     equ 0x012F      ; out HL=Pixel count
BIOS_FETCHC:    equ 0x0114      ; out A = cursor mask, HL = VRAM address of cursor
BIOS_STOREC:    equ 0x0117      ; in A = cursor mask, HL = VRAM address of cursor
BIOS_RESET:     equ 0x7D17      ; restart BASIC
BIOS_IOALLOC:   equ 0X7e6b      ; memory setup

BIOS_GETVCP:    equ 0x0150      ; get PSG voice buffer address (in A = voice number, out HL = address of byte 2)
BIOS_GETVC2:    equ 0x0153      ; get PSG voice buffer address (VOICEN = voice number, in L = byte number 0-36, out HL = address)

BIOS_CHPUT_LF:  equ 0x0908
BIOS_CHPUT_CR:  equ 0x0A81
BIOS_CHPUT_TAB: equ 0x0A71

; MSX2
BIOS_CHKNEW:    equ 0x0165      ; C-flag set if screenmode = 5, 6, 7 or 8
BIOS_EXTROM:	equ	0x015F
BIOS_SCALXY2:   equ 0x008D      ; in BC = X, DE = Y
BIOS_MAPXYC2:   equ 0x0091      ; in BC = X, DE = Y
BIOS_SETC2:     equ 0x009D      ; set current pixel to color from SETATR
BIOS_READC2:    equ 0x0095      ; out A = color of current pixel
BIOS_CHGMOD2:   equ 0x00D1      ; in A = screenmode
BIOS_DOBOXF:    equ 0x0079      ; hl = basic text pointer
BIOS_GRPPRT2:   equ 0x0089      ; a = character
BIOS_CHGCLR2:   equ 0x0111      ; change color, a = screen mode
BIOS_CALPAT2:   equ 0x00F9
BIOS_CALATR2:   equ 0x00FD
BIOS_GSPSIZ2:   equ 0x0101
BIOS_CLRSPR2:   equ 0x00F5

;---------------------------------------------------------------------------------------------------------
; BIOS WORK AREAS
;---------------------------------------------------------------------------------------------------------

BIOS_VERSION:   equ 0x002D   ; 0 = MSX1, 1 = MSX2, 2 = MSX2+, 3 = MSXturboR
BIOS_FORCLR:    equ 0xF3E9
BIOS_BAKCLR:    equ 0xF3EA
BIOS_BDRCLR:    equ 0xF3EB
BIOS_ATRBYT:    equ 0xF3F2
BIOS_INTFLG:    equ 0xFC9B
BIOS_EXPTBL:    equ 0xFCC1
BIOS_JIFFY:     equ 0xFC9E
BIOS_BOTTOM:    equ 0xFC48
BIOS_HIMEM:     equ 0xFC4A
BIOS_SCRMOD:    equ 0xFCAF   ; 0=40x24 Text Mode, 1=32x24 Text Mode, 2=Graphics Mode, 3=Multicolour Mode.
BIOS_CLIKSW:    equ 0xF3DB   ; 0=keyboard click off, 1=keyboard click on
BIOS_GRPACX:    equ 0xFCB7
BIOS_GRPACY:    equ 0xFCB9
BIOS_DATLIN:    equ 0xF6A3   ; 2 - line number of DATA statement read by READ statement
BIOS_DATPTR:    equ 0xF6C8   ; 2 - address of data read by executing READ statement
BIOS_FLGINP:    equ 0xF6A6   ; 1 - flag used in INPUT or READ
BIOS_TEMP:      equ 0xF6A7   ; 2
BIOS_TEMP2:     equ 0xF6BC   ; 2
BIOS_TEMP3:     equ 0xF69D   ; 2
BIOS_TEMP8:     equ 0xF69F   ; 2
BIOS_TEMP9:     equ 0xF7B8   ; 2
BIOS_OLDSCR:    equ 0xFCB0   ; screen mode of the last text mode set
BIOS_LINL40:    equ 0xF3AE   ; width for 40 columns screen mode
BIOS_LINL32:    equ 0xF3AF   ; width for 32 columns screen mode
BIOS_LINLEN:    equ 0xF3B0   ; current width for text screen mode
BIOS_CLMLST:    equ 0xF3B2   ; minimum number of columns that must still be available on a line for a CRLF
BIOS_TXTNAM:    equ 0xF3B3   ; characters table name

BIOS_VOICEN:    equ 0xFB38   ; PSG voice number
BIOS_MCLTAB:    equ 0xF956
BIOS_PRSCNT:    equ 0xFB35
BIOS_SAVSP:     equ 0xFB36
BIOS_QUEUEN:    equ 0xFB3E
BIOS_MUSICF:    equ 0xFB3F   ;contains 3 bit flags set by the STRTMS. Bits 0, 1 and 2 correspond to VOICAQ, VOICBQ and VOICCQ.
BIOS_PLYCNT:    equ 0xFB40

BIOS_DRWFLG:    equ 0xFCBB
BIOS_MCLFLG:    equ 0xF958

BIOS_SLTROM:    equ 0xFCC1
BIOS_RAMAD0:	equ	0xF341	; Main-RAM Slot (00000h~03FFFh)
BIOS_RAMAD1:	equ	0xF342	; Main-RAM Slot (04000h~07FFFh)
BIOS_RAMAD2:	equ	0xF343	; Main-RAM Slot (08000h~0BFFFh)
BIOS_RAMAD3:	equ	0xF344	; Main-RAM Slot (0C000h~0FFFFh)
)";

const char * Assembler::math_pack_functions = R"(
;---------------------------------------------------------------------------------------------------------
; MATH PACK FUNCTIONS
;---------------------------------------------------------------------------------------------------------

MATH_PACK:  equ 1

MATH_DECSUB:equ 0x268C
MATH_DECADD:equ 0x269A
MATH_DECMUL:equ 0x27E6
MATH_DECDIV:equ 0x289F
MATH_SNGEXP:equ 0x37C8
MATH_DBLEXP:equ 0x37D7
MATH_COS:   equ 0x2993
MATH_SIN:   equ 0x29AC
MATH_TAN:   equ 0x29FB
MATH_ATN:   equ 0x2A14
MATH_SQR:   equ 0x2AFF
MATH_LOG:   equ 0x2A72
MATH_EXP:   equ 0x2B4A
MATH_DECNRM:equ 0x26FA
MATH_SIGN:  equ 0x2E71
MATH_ABSFN: equ 0x2E82
MATH_NEG:   equ 0x2E8D
MATH_SGN:   equ 0x2E97
MATH_RND:   equ 0x2BDF
MATH_UMULT: equ 0x314A
MATH_ISUB:  equ 0x3167
MATH_IADD:  equ 0x3172
MATH_IMULT: equ 0x3193
MATH_IDIV:  equ 0x31E6
MATH_IMOD:  equ 0x323A
MATH_INTEXP:equ 0x383F
MATH_FRCINT:equ 0x2F8A
MATH_FRCSGL:equ 0x2FB2
MATH_FRCDBL:equ 0x303A
MATH_FIXER: equ 0x30BE
MATH_MAF:   equ 0x2C4D
MATH_MAM:   equ 0x2C50
MATH_MOV8DH:equ 0x2C53
MATH_MFA:   equ 0x2C59
MATH_MFM:   equ 0x2C5C
MATH_MMF:   equ 0x2C67
MATH_MOV8HD:equ 0x2C6A
MATH_XTF:   equ 0x2C6F
MATH_PHA:   equ 0x2CC7
MATH_PHF:   equ 0x2CCC
MATH_PPA:   equ 0x2CDC
MATH_PPF:   equ 0x2CE1
MATH_PUSHF: equ 0x2EB1
MATH_MOVFM: equ 0x2EBE
MATH_MOVFR: equ 0x2EC1
MATH_MOVRF: equ 0x2ECC
MATH_MOVRMI:equ 0x2ED6
MATH_MOVRM: equ 0x2EDF
MATH_MOVMF: equ 0x2EE8
MATH_MOVE:  equ 0x2EEB
MATH_VMOVAM:equ 0x2EEF
MATH_MOVVFM:equ 0x2EF2
MATH_VMOVE: equ 0x2EF3
MATH_VMOVFA:equ 0x2F05
MATH_VMOVFM:equ 0x2F08
MATH_VMOVAF:equ 0x2F0D
MATH_VMOVMF:equ 0x2F10
MATH_ICOMP: equ 0x2F4D
MATH_DCOMP: equ 0x2F21
MATH_XDCOMP:equ 0x2F5C
MATH_FIN:   equ 0x3299
MATH_FOUT:  equ 0x3425
MATH_PUFOUT:equ 0x3426
MATH_FIOUT: equ 0x3441
MATH_FBOUT: equ 0x371A
MATH_FOOUT: equ 0x371E
MATH_FHOUT: equ 0x3722
)";

const char * Assembler::basic_functions = R"(
;---------------------------------------------------------------------------------------------------------
; MSX BASIC FUNCTIONS
;---------------------------------------------------------------------------------------------------------

BASIC_AUTO:   equ 0x3973
BASIC_AND:    equ 0x3A18
BASIC_ATTR:   equ 0x39FE
BASIC_BASE:   equ 0x39BE
BASIC_BSAVE:  equ 0x39CC
BASIC_BLOAD:  equ 0x39CA
BASIC_BEEP:   equ 0x39AC
BASIC_CALL:   equ 0x39C0
BASIC_CLOSE:  equ 0x3994
BASIC_COPY:   equ 0x39D8
BASIC_CONT:   equ 0x395E
BASIC_CLEAR:  equ 0x3950
BASIC_CLOAD:  equ 0x3962
BASIC_CSAVE:  equ 0x3960
BASIC_CSRLIN: equ 0x39FC
BASIC_CIRCLE: equ 0x39A4
BASIC_COLOR:  equ 0x39A6
BASIC_CLS:    equ 0x396A
BASIC_CMD:    equ 0x39DA
BASIC_DELETE: equ 0x397C
BASIC_DATA:   equ 0x3934
BASIC_DIM:    equ 0x3938
BASIC_DEFSTR: equ 0x3982
BASIC_DEFINT: equ 0x3984
BASIC_DEFSNG: equ 0x3986
BASIC_DEFDBL: equ 0x3988
BASIC_DSKO:   equ 0x39CE
BASIC_DEF:    equ 0x395A
BASIC_DSKI:   equ 0x3A00
BASIC_DRAW:   equ 0x39A8
BASIC_ELSE:   equ 0x396E
BASIC_END:    equ 0x392E
BASIC_ERASE:  equ 0x3976
BASIC_ERROR:  equ 0x3978
BASIC_ERL:    equ 0x39EE
BASIC_ERR:    equ 0x39F0
BASIC_EQU:    equ 0x3A1E
BASIC_FOR:    equ 0x3920
BASIC_FIELD:  equ 0x398E
BASIC_FILES:  equ 0x39AA
BASIC_FN:     equ 0x39E8
BASIC_GOTO:   equ 0x393E
BASIC_GOSUB:  equ 0x3948
BASIC_GET:    equ 0x3990
BASIC_INPUT:  equ 0x3936
BASIC_IF:     equ 0x3942
BASIC_INSTR:  equ 0x39F6
BASIC_IMP:    equ 0x3A20
BASIC_INKEY:  equ 0x3A04
BASIC_IPL:    equ 0x39D6
BASIC_KILL:   equ 0x39D4
BASIC_KEY:    equ 0x3964
BASIC_LPRINT: equ 0x394C
BASIC_LLIST:  equ 0x3968
BASIC_LET:    equ 0x393C
BASIC_LOCATE: equ 0x39DC
BASIC_LINE:   equ 0x398A
BASIC_LOAD:   equ 0x3996
BASIC_LSET:   equ 0x399C
BASIC_LIST:   equ 0x3952
BASIC_LFILES: equ 0x39A2
BASIC_MOTOR:  equ 0x39C8
BASIC_MERGE:  equ 0x3998
BASIC_MOD:    equ 0x3A22
BASIC_MAX:    equ 0x39C6
BASIC_NEXT:   equ 0x3932
BASIC_NAME:   equ 0x39D2
BASIC_NEW:    equ 0x3954
BASIC_NOT:    equ 0x39EC
BASIC_OPEN:   equ 0x398C
BASIC_OUT:    equ 0x3964
BASIC_ON:     equ 0x3956
BASIC_OR:     equ 0x3A1A
BASIC_OFF:    equ 0x3A02
BASIC_PRINT:  equ 0x394E
BASIC_PUT:    equ 0x3992
BASIC_POKE:   equ 0x395C
BASIC_PSET:   equ 0x39B0
BASIC_PRESET: equ 0x39B2
BASIC_POINT:  equ 0x3A06
BASIC_PAINT:  equ 0x39AA
BASIC_PLAY:   equ 0x39AE
BASIC_RETURN: equ 0x3948
BASIC_READ:   equ 0x393A
BASIC_RUN:    equ 0x3940
BASIC_RESTORE:equ 0x3944
BASIC_REM:    equ 0x394A
BASIC_RESUME: equ 0x397A
BASIC_RSET:   equ 0x399E
BASIC_RENUM:  equ 0x3980
BASIC_SCREEN: equ 0x39B6
BASIC_SPRITE: equ 0x39BA
BASIC_STOP:   equ 0x394C
BASIC_SWAP:   equ 0x3974
BASIC_SET:    equ 0x39D0
BASIC_SAVE:   equ 0x39A0
BASIC_SPC:    equ 0x39EA
BASIC_STEP:   equ 0x39E4
BASIC_STRING: equ 0x39F2
BASIC_SPACE1: equ 0x397E
BASIC_SOUND:  equ 0x39B4
BASIC_THEN:   equ 0x39E0
BASIC_TRON:   equ 0x3970
BASIC_TROFF:  equ 0x3972
BASIC_TAB:    equ 0x39E2
BASIC_TO:     equ 0x39DE
BASIC_TIME:   equ 0x39C2
BASIC_USING:  equ 0x39F4
BASIC_USR:    equ 0x39E6
BASIC_VARPTR: equ 0x39FA
BASIC_VDP:    equ 0x39BC
BASIC_VPOKE:  equ 0x39B8
BASIC_WIDTH:  equ 0x396C
BASIC_WAIT:   equ 0x3958
BASIC_XOR:    equ 0x3A1C
BASIC_ABS:    equ 0x39E8
BASIC_ATN:    equ 0x39F8
BASIC_ASC:    equ 0x3A06
BASIC_BIN:    equ 0x3A16
BASIC_CINT:   equ 0x3A18
BASIC_CSNG:   equ 0x3A1A
BASIC_CDBL:   equ 0x3A1C
BASIC_CVI:    equ 0x3A2C
BASIC_CVS:    equ 0x3A2E
BASIC_CVD:    equ 0x3A30
BASIC_COS:    equ 0x39F4
BASIC_CHR:    equ 0x3A08
BASIC_DSKF:   equ 0x3A28
BASIC_EXP:    equ 0x39F2
BASIC_EOF:    equ 0x3A32
BASIC_FRE:    equ 0x39FA
BASIC_FIX:    equ 0x3A1E
BASIC_FPOS:   equ 0x3A2A
BASIC_HEX:    equ 0x3A12
BASIC_INT:    equ 0x39E6
BASIC_INP:    equ 0x39FC
BASIC_LPOS:   equ 0x3A14
BASIC_LOG:    equ 0x39F0
BASIC_LOC:    equ 0x3A34
BASIC_LEN:    equ 0x3A00
BASIC_LEFT:   equ 0x39DE
BASIC_LOF:    equ 0x3A36
BASIC_MKI:    equ 0x3A38
BASIC_MKS:    equ 0x3A3A
BASIC_MKD:    equ 0x3A3C
BASIC_MID:    equ 0x39E2
BASIC_OCT:    equ 0x3A10
BASIC_POS:    equ 0x39FE
BASIC_PEEK:   equ 0x3A0A
BASIC_PDL:    equ 0x3A24
BASIC_PAD:    equ 0x3A26
BASIC_RIGHT:  equ 0x39E0
BASIC_RND:    equ 0x39EC
BASIC_SGN:    equ 0x39E4
BASIC_SQR:    equ 0x39EA
BASIC_SIN:    equ 0x39EE
BASIC_STR:    equ 0x3A02
BASIC_SPACE2: equ 0x3A0E
BASIC_STICK:  equ 0x3A20
BASIC_STRIG:  equ 0x3A22
BASIC_TAN:    equ 0x39F6
BASIC_VAL:    equ 0x3A04
BASIC_VPEEK:  equ 0x3A0C

BASIC_TRAP_ENABLE:  equ 0x631B    ; ON INTERVAL/KEY/SPRITE/STOP/STRIG - hl = pointer to trap block
BASIC_TRAP_DISABLE: equ 0x632B    ; hl = pointer to trap block
BASIC_TRAP_ACKNW:   equ 0x6358    ; hl, acknowledge trap (handle trap: sts=5? has handler? ackn, pause, run trap, sts=1? unpause)
BASIC_TRAP_PAUSE:   equ 0x6331    ; hl
BASIC_TRAP_UNPAUSE: equ 0x633E    ; hl
BASIC_TRAP_CLEAR:   equ 0x636E

BASIC_PLAY_DIRECT:  equ 0x744C
BASIC_DRAW_DIRECT:  equ 0x568C

BASIC_READYR:       equ 0x409B
BASIC_READYC:       equ 0x7D17
BASIC_FACEVAL:      equ 0x4DC7

BASIC_ERROR_HANDLER:equ 0x406F
BASIC_ERROR_SYNTAX: equ 0x4055
BASIC_ERROR_DIVZER: equ 0x4058
BASIC_ERROR_OVRFLW: equ 0x4067
BASIC_ERROR_ARRAY:  equ 0x405E
BASIC_ERROR_TYPMIS: equ 0x406D

; BASIC ERROR CODES TO BASIC_ERROR_HANDLER
; 01 NEXT without FOR             19 Device I/O error
; 02 Syntax error                 20 Verify error
; 03 RETURN without GOSUB         21 No RESUME
; 04 Out of DATA                  22 RESUME without error
; 05 Illegal function call        23 Unprintable error
; 06 Overflow                     24 Missing operand
; 07 Out of memory                25 Line buffer overflow
; 08 Undefined line number        50 FIELD overflow
; 09 Subscript out of range       51 Internal error
; 10 Redimensioned array          52 Bad file number
; 11 Division by zero             53 File not found
; 12 Illegal direct               54 File already open
; 13 Type mismatch                55 Input past end
; 14 Out of string space          56 Bad file name
; 15 String too long              57 Direct statement in file
; 16 String formula too complex   58 Sequential I/O only
; 17 Can't CONTINUE               59 File not OPEN
; 18 Undefined user function

;---------------------------------------------------------------------------------------------------------
; MSX BASIC WORK AREAS
;---------------------------------------------------------------------------------------------------------

BASIC_DAC:    equ 0xF7F6    ; 16
BASIC_ARG:    equ 0xF847    ; 16
BASIC_VALTYP: equ 0xF663
BASIC_RNDX:   equ 0xF857
BASIC_FBUFFR: equ 0xF7C5
BASIC_BUF:    equ 0xF55E    ; 259
BASIC_KBUF:   equ 0xF41F    ; 318
BASIC_SWPTMP: equ 0xF7BC    ; 8
BASIC_STRBUF: equ 0xF7C5    ; 43
BASIC_TXTTAB: equ 0xF676
BASIC_VARTAB: equ 0xF6C2
BASIC_ARYTAB: equ 0xF6C4
BASIC_STREND: equ 0xF6C6
BASIC_STKTOP: equ 0xF674
BASIC_FRETOP: equ 0xF69B
BASIC_MEMSIZ: equ 0xF672

BASIC_TEMPPT: equ 0xF678    ; 2	Starting address of unused area of temporary descriptor.
BASIC_TEMPST: equ 0xF67A    ; 30 Temporary descriptors.

BASIC_DATPTR: equ 0xF6C8    ; 2 Pointer to next data to read from the instruction DATA. Modified by RESTORE.
BASIC_DATLIN: equ 0xF6A3    ; 2 Número de linha do comando DATA para o comando READ.
BASIC_DORES:  equ 0xF664    ; 1 Usada pelo comando DATA para manter o texto no formato ASCII.
BASIC_DEFTBL: equ 0xF6CA    ; 26 table of variables defined by DEFINT, DEFSTR, DEFSNG and DEFDBL for each alphabet letter (2 = integer, 3 = String, 4 = Simple precision, 8 = Double precision).

BASIC_CURLIN: equ 0xF41C    ; BASIC current line number
BASIC_INTVAL: equ 0xFCA0    ; interval value
BASIC_INTCNT: equ 0xFCA2    ; interval current count

BASIC_PRMPRV: equ 0xF74C    ; Pointer to previous parameter block in PARM1

BASIC_TRPTBL: equ 0xFC4C    ; 78 trap table - array of 3 bytes - state[1] (bit 0=on, bit 1=stop, bit 2=active) + address[2]

BASIC_TRPTBL_KEY:        equ 0xFC4C  ; 30 ON KEY GOSUB
BASIC_TRPTBL_STOP:       equ 0xFC6A  ; 3  ON STOP GOSUB
BASIC_TRPTBL_SPRITE:     equ 0xFC6D  ; 3  ON SPRITE GOSUB
BASIC_TRPTBL_STRIG:      equ 0xFC70  ; 15 ON STRIG GOSUB
BASIC_TRPTBL_INTERVAL:   equ 0xFC7F  ; 3  ON INTERVAL GOSUB
BASIC_TRPTBL_OTHER:      equ 0xFC82  ; 24 reserved for expansion

BASIC_ONGSBF:            equ 0xFBD8  ; 1  trap occurred counter (0=not occurred)
)";

const char * Assembler::support_macros = R"(
MACRO __call_basic,CALL_PARM
    ld ix, CALL_PARM
    call BIOS_CALBAS
ENDM

if defined COMPILE_TO_DOS

  MACRO __call_bios,CALL_PARM
    ;ld iy,(BIOS_EXPTBL-1)
    ld ix, CALL_PARM
    call BIOS_CALBAS   ; BIOS_CALSLT
  ENDM

else

  MACRO __call_bios,CALL_PARM
    call CALL_PARM
  ENDM

endif

if defined BCD_MATH
  MACRO __call_math,CALL_PARM
    ld ix, CALL_PARM
    ;call BIOS_CALBAS
    call BCD_MATH_CALL
  ENDM
else
  MACRO __call_math,CALL_PARM
    call CALL_PARM
  ENDM
endif

MACRO push.parm
    push hl ; save parameter
ENDM

MACRO pop.parm
    pop iy  ; restore PC of caller
    pop hl  ; get next parameter
    push iy ; save PC of caller
ENDM

MACRO push.ret.parm
    pop iy  ; restore PC of caller
    push hl ; save return parameter
    push iy ; save PC of caller
ENDM

MACRO ret.parm
    pop iy         ; restore PC of caller
    push hl        ; save return parameter
    push iy        ; save PC of caller
    ret            ; return
ENDM

MACRO set.line.number, line_number
    ld bc, line_number          ; current line number
    ld (BASIC_CURLIN), bc
ENDM

MACRO verify.break
    ld a, (BIOS_INTFLG)         ; verify CTRL+BREAK
    or a
    jp nz, end_pgm
ENDM

MACRO check.traps
     ld a, (BASIC_ONGSBF)       ; trap occured counter
     or a
     call nz, RUN_TRAPS
ENDM
)";

const char * Assembler::program_header = R"(
;---------------------------------------------------------------------------------------------------------
; PROGRAM HEADER
;---------------------------------------------------------------------------------------------------------

    romSize:       equ  0x8000                 ; ROM size (32k)
    pageSize:      equ  0x4000                 ; Page size (16k)
    lowLimitSize:  equ  0x400                  ; 10% of a page size

if defined COMPILE_TO_BIN

    pgmArea:    equ  0x8000                 ; page 2 - program area
    ramArea:    equ  0xc000                 ; page 3 - free RAM start area

    org  pgmArea                ; program binary type start address
    db	 0FEh		            ; binary file ID
    dw	 start_pgm	            ; begin address
    dw	 end_file - 1	        ; end address
    dw	 start_pgm	            ; program execution address (for ,R option)

else
if defined COMPILE_TO_ROM

    pgmArea:    equ  0x4000                 ; page 1 and 2 - program area
    ramArea:    equ  0xc000                 ; page 3 - free RAM start area

    org  pgmArea                ; program rom type start address
    db   'AB'                   ; rom file ID
    dw   start_pgm              ; INIT
    dw   0x0000                 ; STATEMENT
    dw   0x0000                 ; DEVICE
    dw   0x0000                 ; TEXT
    ds   6,0                    ; RESERVED

else

    pgmArea:    equ  0x8000                 ; page 2 - program area
    ramArea:    equ  0xc000                 ; page 3 - free RAM start area

    org  pgmArea                ; program DOS type start address    ; 0x0100

endif
endif
)";

const char * Assembler::program_footer = R"(
;---------------------------------------------------------------------------------------------------------
; PROGRAM BASIC ROM HOOKS
;---------------------------------------------------------------------------------------------------------

if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS

BASIC_SLOT_ENABLE:
    push af
    push hl
    push de
    push bc
      ld a, (BIOS_EXPTBL)
      ld hl,04000h
      __call_bios BIOS_ENASLT ; Select main ROM on page 1 (4000h~7FFFh)
    pop bc
    pop de
    pop hl
    pop af
    ret

PROGRAM_SLOT_1_ENABLE:
    push af
    push hl
    push de
    push bc
      call BIOS_RSLREG
  	  rrca
	  rrca
	  rrca
	  rrca
	  and	3	;Keep bits corresponding to the page
	  ld	c,a
	  ld	b,0
	  ld	hl,BIOS_EXPTBL
	  add	hl,bc
	  ld	a,(hl)
	  and	80h
	  or	c
	  ld	c,a
	  inc	hl
	  inc	hl
	  inc	hl
	  inc	hl
	  ld	a,(hl)
	  and	0Ch
	  or	c
	  ld	h,040h
      call BIOS_ENASLT	; Select the ROM on page 4000h-7FFFh
    pop bc
    pop de
    pop hl
    pop af
    ret

else

BASIC_SLOT_ENABLE:
PROGRAM_SLOT_1_ENABLE: ret

endif

if defined DO_DRAW

DRAW_HOOK:
   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     ld de, PROGRAM_SLOT_1_ENABLE
     push de
   endif

   ld bc, 0
   push bc
   push bc
   push bc
   push hl                 ; address of string
   push af                 ; size of string

   ld de, 0x5D83
   ld (BIOS_MCLTAB),de
   xor a
   ld (BIOS_DRWFLG),a
   ld (BIOS_MCLFLG),a

   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     call BASIC_SLOT_ENABLE
   endif

   jp BASIC_DRAW_DIRECT

endif

if defined SET_PLAY_VOICE_1 or defined SET_PLAY_VOICE_2 or defined SET_PLAY_VOICE_3 or defined DO_PLAY or defined MUSIC_PLAY or defined MUSIC_NEXT or defined MUSIC_STOP

PLAY_HOOK:
   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     ld de, PROGRAM_SLOT_1_ENABLE
     push de
   endif

   ld HL, 0x752E
   ld (BIOS_MCLTAB),HL
   ld a, 1
   ld (BIOS_MCLFLG),A
   ld hl, BIOS_TEMP      ; voice count
   ld a, 3
   sub (hl)
   dec a
   ld (BIOS_PRSCNT), a
   xor a
   ld (BIOS_VOICEN), a
   ld (BIOS_QUEUEN), a
   ld (BIOS_MUSICF), a
   ld HL,-12 ; -10
   add HL,SP
   ld (BIOS_SAVSP),HL
   ld HL, BIOS_PLYCNT
   ld (HL), 0

   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     call BASIC_SLOT_ENABLE
   endif

   jp BASIC_PLAY_DIRECT

endif

if defined gfxBorderFill

PAINT_HOOK:
   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     ld de, PROGRAM_SLOT_1_ENABLE
     push de
   endif

   ld bc, (BIOS_GRPACX)
   ld (BIOS_GXPOS), bc
   ld de, (BIOS_GRPACY)
   ld (BIOS_GYPOS), de
   push bc
   push de

   xor a
   ld hl, BASIC_BUF
   ld (hl), a

   ld a, (BIOS_BDRATR)
   xor b
   ld e, a
   ld a, (BIOS_ATRBYT)
   xor d
   ld c, a

   push af
   push bc
   push hl
   push de

   call gfxIsScreenModeMSX2
   jr nc, PAINT_HOOK.1  ; if MSX2 and screen mode above 3, jump (BASIC_SUB_PAINT2)
     ld a, (BIOS_BDRATR)
     ld (BIOS_ATRBYT), a
     ld e, a
PAINT_HOOK.1:
   if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS
     call BASIC_SLOT_ENABLE
   endif

   pop de
   pop hl
   pop bc
   pop af

   jp BASIC_SUB_PAINT1

endif

if defined BCD_MATH

BCD_MATH_CALL:
  ld iy, BCD_MATH_CALL.1
  push iy
  push ix
  call BASIC_SLOT_ENABLE
  ret
BCD_MATH_CALL.1:
  jp PROGRAM_SLOT_1_ENABLE

endif

;---------------------------------------------------------------------------------------------------------
; PROGRAM FOOTER
;---------------------------------------------------------------------------------------------------------

    if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS

        romPad:

        pgmPage2.pad: equ romSize - (romPad - pgmArea)

        if pgmPage2.pad >= 0
           ds pgmPage2.pad, 0

           if pgmPage2.pad < lowLimitSize
                .WARNING "There's only less than 5% free space on this ROM"
           endif
        else
           .ERROR "There's no free space left on this ROM"
        endif

    endif

    end_file: end start_pgm           ; label start is the entry point
)";

const char * Assembler::program_main_code = R"(
;---------------------------------------------------------------------------------------------------------
; PROGRAM MAIN CODE
;---------------------------------------------------------------------------------------------------------

start_pgm:                               ; start of the program

    if defined COMPILE_TO_DOS

       call BIOS_SLOT_ENABLE   ; enable bios on page 0
       ;call BASIC_SLOT_ENABLE  ; enable basic on page 1

    endif

    if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS

       call PROGRAM_SLOT_2_SAVE     ; save slot on page 2
       call PROGRAM_SLOT_2_ENABLE   ; enable program on page 2

    endif

    __call_bios BIOS_ERAFNK      ; turn off function keys display
    __call_bios BIOS_GICINI      ; initialize sound system
    __call_bios BIOS_INITXT      ; initialize text screen
    xor a
    ld (BIOS_CLIKSW), a          ; disable keyboard click
    ld bc, 0xFFFF                ;
    ld (BASIC_CURLIN), bc        ; interpreter in direct mode
    __call_basic BASIC_TRAP_CLEAR        ; clear traps work space
    ;call INITIALIZE_PARAMETERS   ; initialize parameters stack
    call memory.init             ; initialize memory allocation
    call INITIALIZE_VARIABLES    ; initialize variables
)";

const char * Assembler::program_end_code = R"(
;---------------------------------------------------------------------------------------------------------
; PROGRAM END CODE
;---------------------------------------------------------------------------------------------------------

    end_pgm:    __call_bios BIOS_DSPFNK      ; turn on function keys display
                ld a, 1                      ;
                ld (BIOS_CLIKSW), a          ; enable keyboard click

                if defined COMPILE_TO_ROM
                    jp PROGRAM_TO_BASIC
                else
                   if defined COMPILE_TO_DOS
                      ; go back to DOS
                   else
                      __call_basic BASIC_READYR    ; warm start Basic
                   endif
                endif

                ret                          ; end of the program

                ;__call_bios BIOS_GICINI      ; initialize sound system
                ;if defined COMPILE_TO_DOS or defined COMPILE_TO_ROM
                ;   __call_bios BIOS_RESET       ; restart Basic
                ;else
                ;   __call_basic BASIC_END       ; end to Basic
                ;endif
)";

const char * Assembler::program_routines = R"(
;---------------------------------------------------------------------------------------------------------
; PROGRAM ROUTINES
;---------------------------------------------------------------------------------------------------------

if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS

PROGRAM_TO_BASIC:
    call PROGRAM_SLOT_2_RESTORE
    __call_basic BASIC_READYR    ; warm start Basic
    ret

PROGRAM_SLOT_2_SAVE:
    ld h,080h
    call PROGRAM_SLOT_GET
    ld (BIOS_RAMAD2), a ; Save RAM slot of page 8000h-BFFFh
    ret

PROGRAM_SLOT_2_RESTORE:
    ld a, (BIOS_RAMAD2)
	ld h,080h
	jp BIOS_ENASLT	    ; Select the RAM on page 8000h-BFFFh

PROGRAM_SLOT_2_ENABLE:
    call BIOS_RSLREG
	rrca
	rrca
	and	3	;Keep bits corresponding to the page
	ld	c,a
	ld	b,0
	ld	hl,BIOS_EXPTBL
	add	hl,bc
	ld	a,(hl)
	and	80h
	or	c
	ld	c,a
	inc	hl
	inc	hl
	inc	hl
	inc	hl
	ld	a,(hl)
	and	0Ch
	or	c
	ld	h,080h
	jp BIOS_ENASLT		; Select the ROM on page 8000h-BFFFh

; h = memory page
; a <- slot ID formatted FxxxSSPP
; Modifies: af, bc, de, hl
; ref: https://www.msx.org/forum/msx-talk/development/fusion-c-and-htimi#comment-366469
PROGRAM_SLOT_GET:
	call BIOS_RSLREG
	bit 7,h
	jr z,PrimaryShiftContinue
	rrca
	rrca
	rrca
	rrca
PrimaryShiftContinue:
	bit 6,h
	jr z,PrimaryShiftDone
	rrca
	rrca
PrimaryShiftDone:
	and 00000011B
	ld c,a
	ld b,0
	ex de,hl
	ld hl,BIOS_EXPTBL
	add hl,bc
	ld c,a
	ld a,(hl)
	and 80H
	or c
	ld c,a
	inc hl  ; move to SLTTBL
	inc hl
	inc hl
	inc hl
	ld a,(hl)
	ex de,hl
	bit 7,h
	jr z,SecondaryShiftContinue
	rrca
	rrca
	rrca
	rrca
SecondaryShiftContinue:
	bit 6,h
	jr nz,SecondaryShiftDone
	rlca
	rlca
SecondaryShiftDone:
	and 00001100B
	or c
	ret

endif

if defined COMPILE_TO_DOS

BIOS_SLOT_ENABLE:
               ld a, (BIOS_EXPTBL)
               ld hl,0
               __call_bios BIOS_ENASLT ; Select main ROM on page 0 (0000h~3FFFh)
               ret

endif
)";

const char * Assembler::support_basic_begin = R"(
;---------------------------------------------------------------------------------------------------------
; MSX BASIC KEYWORDS
;---------------------------------------------------------------------------------------------------------
)";

const char * Assembler::support_basic_end = R"(
;---------------------------------------------------------------------------------------------------------
; MSX BASIC SUPPORT CODE
;---------------------------------------------------------------------------------------------------------

if defined CHR or defined INKEY.STR

; in a = char
; out hl = string
COPY_CHAR_TO_STR:
              push af
                ld bc, 2                     ; string size
                call memory.alloc          ; in bc size, out ix new memory address, nz=OK
                jr z, COPY_CHAR_TO_STR.ERROR
              pop af
              ld (ix), a
              xor a
              ld (ix+1), a
              push ix
              pop hl                       ; HL = string address
              ld a, 1                      ; A = size
              ret
COPY_CHAR_TO_STR.ERROR:
              pop af
              jp memory.error

endif

;if defined INPUT or defined LINE_INPUT or defined SPC or defined SPACE or defined INPUT.FUNCTION.STR

; bc = size
; out: a = size, hl = address
GET_STRING_SPACE:
              ld hl, LIT_NULL_STR
              ld a, c
              or a
              ret z
              push de
              push bc
                inc bc
                call memory.alloc       ; in bc = size, out ix = address, nz=OK
              pop bc
              pop de
              jp z, memory.error
              ld a, 32
              ld b, c
              push ix
GET_STRING_SPACE.1:
                ld (ix), a
                inc ix
                djnz GET_STRING_SPACE.1
                xor a
                ld (ix), a
              pop hl
              ld a, c   ; size
              ret

; in hl = string, out hl = temp string
COPY_TO.TEMP_STR:
              call GET_STR.LENGTH         ; get string length, in hl = address, out a = size
              ex de, hl
              ld c, a
              ld b, 0
              call GET_STRING_SPACE       ; in bc = size, out hl = address, out a = string size
              or 0
              ret z
              push af
              push hl
                push hl
                  call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                pop hl
                ex de, hl
                ld c, a
                ld b, 0
                ldir                      ; copy bc bytes from hl to de
              pop hl
              pop af
              ret

;endif

if defined INPUT or defined LINE_INPUT

INPUT.CONTINUE:
              jr c, INPUT.EXIT            ; exit if CTRL+STOP

              inc hl                      ; string start
              call COPY_TO.TEMP_STR
              call COPY_TO.VAR_DUMMY.STR  ; make a fake string variable from HL
              push.parm                   ; LET parameter 2 - fake string variable as right operand
              ld hl, (BIOS_TEMP)
              push.parm                   ; LET parameter 1 - input variable as left operand
              call LET                    ; put string into variable
INPUT.EXIT:
              xor a                       ;
              ld (BIOS_CLIKSW), a         ; disable keyboard click
              ret                         ;

endif

if defined ON_ERROR or defined ON_INTERVAL or defined ON_KEY_START or defined ON_SPRITE or defined ON_STOP or defined ON_STRIG_START or defined TRAP_ENABLED or defined TRAP_DISABLED or defined TRAP_PAUSE or defined TRAP_UNPAUSE

RUN_TRAPS:    ld b, 26
              ld hl, BASIC_TRPTBL
RUN_TRAPS.1:  push hl
                push bc
                  call TRAP_HANDLER
                pop bc
              pop hl
              inc hl
              inc hl
              inc hl
              djnz RUN_TRAPS.1
              ret

; in hl = trap block address (handle trap: sts=5? has handler? ackn, pause, run trap, sts=1? unpause)
TRAP_HANDLER:
                ld a, (hl)    ; trap status
                cp 5          ; trap occured AND trap not paused AND trap enabled ?
                ret nz        ; return if false
                inc hl
                ld e, (hl)    ; get trap address
                inc hl
                ld d, (hl)
                dec hl
                dec hl
                ld a, d
                or e
                ret z         ; return if address zero
                push hl
                  __call_basic BASIC_TRAP_ACKNW
                  __call_basic BASIC_TRAP_PAUSE
                  ld hl, TRAP_HANDLER.1
                  ld a, (BASIC_ONGSBF)  ; save traps execution
                  push af
                  xor a
                  ld (BASIC_ONGSBF), a  ; disable traps execution
                  push hl  ; next return will be to trap handler
                  push de  ; indirect jump to trap address
                  ret
TRAP_HANDLER.1: pop af
                ld (BASIC_ONGSBF), a    ; restore traps execution
                pop hl
                ld a, (hl)
                cp 1       ; trap enabled?
                ret z
                __call_basic BASIC_TRAP_UNPAUSE
                ret

; hl = trap block, de = trap handler
SET_TRAP:       xor a
                ld (hl), a                  ; trap block status
                inc hl
                ld (hl), e                  ; trap block handler (pointer)
                inc hl
                ld (hl), d
                ret

endif

if defined SET_PLAY_VOICE_1 or defined SET_PLAY_VOICE_2 or defined SET_PLAY_VOICE_3 or defined DO_PLAY or defined MUSIC_PLAY or defined MUSIC_NEXT or defined MUSIC_STOP

   SET_PLAY_VOICE:
        ld (BIOS_TEMP), a       ; save voice number
        pop.parm
		ld a, (hl)
		cp 3
		ret nz                  ; return if not string
        call GET_STR.ADDR
    SET_PLAY_VOICE.1:
		ld (BIOS_TEMP2), a      ; save string size
        push hl                 ; string address
		  ld a, (BIOS_TEMP)     ; restore voice number
		  call BIOS_GETVCP      ; get PSG voice buffer address (in A = voice number, out HL = address of byte 2)
		pop de
		ld a, (BIOS_TEMP2)      ; restore string size
		ld (hl), a              ; string size
		inc hl
		ld (hl), e              ; string address
		inc hl
		ld (hl), d
		inc hl
        ld D,H                  ; voice stack
        ld E,L
        ld BC,001CH
        add HL,BC
        ex DE,HL
        ld (HL),E
        inc HL
        ld (HL),D
		ret

endif
)";

const char * Assembler::variables_workarea_begin = R"(
;---------------------------------------------------------------------------------------------------------
; MAIN WORK AREA - LITERALS / VARIABLES / CONFIGURATIONS
;---------------------------------------------------------------------------------------------------------

if defined COMPILE_TO_ROM or defined COMPILE_TO_DOS

   workAreaPad:
   pgmPage1.pad: equ pageSize - (workAreaPad - pgmArea)

   if pgmPage1.pad >= 0
      ds pgmPage1.pad, 0
   ;else
   ;   .WARNING "There's no free space left on program page 1"
   endif

endif

VAR_STACK.START:     equ ramArea
    ;VAR_STACK.END:       equ VAR_STACK.START + 0x800   ; 2kb (~200 variables)

VAR_STACK.POINTER:   equ VAR_STACK.START

PRINT.CRLF:      db 3, 0, 0, 2
                 dw PRINT.CRLF.DATA, 0, 0, 0
PRINT.CRLF.DATA: db 13,10,0

PRINT.TAB:       db 3, 0, 0, 1
                 dw PRINT.TAB.DATA, 0, 0, 0
PRINT.TAB.DATA:  db 09,0

; null double
LIT_NULL_DBL: dw 0, 0, 0, 0

; null string
LIT_NULL_STR: db 0

; quote string
LIT_QUOTE_CHAR: db '\"'

; logical true
LIT_TRUE: db 2, 0, 0
          dw 0, 0xFFFF, 0, 0

; logical false
LIT_FALSE: db 2, 0, 0
           dw 0, 0, 0, 0
)";

const char * Assembler::variables_workarea_end = R"(
VAR_DUMMY.START:       equ AFTER_LAST_VARIABLE    ; variable dummy circular queue area
VAR_DUMMY.COUNTER:     equ VAR_DUMMY.START        ; variable dummy circular queue count
VAR_DUMMY.POINTER:     equ VAR_DUMMY.COUNTER + 1  ; pointer to next variable dummy
VAR_DUMMY.DATA:        equ VAR_DUMMY.POINTER + 2  ; first variable dummy

VAR_DUMMY.SIZE:        equ 8
VAR_DUMMY.LENGTH:      equ (11 * VAR_DUMMY.SIZE)
VAR_DUMMY.END:         equ VAR_DUMMY.DATA + VAR_DUMMY.LENGTH
VAR_STACK.END:         equ VAR_DUMMY.END + 1
)";

const char * Assembler::variables_initialize = R"(
;---------------------------------------------------------------------------------------------------------
; VARIABLES INITIALIZE
;---------------------------------------------------------------------------------------------------------

INITIALIZE_DUMMY:
    xor a
    ld (VAR_DUMMY.COUNTER), a                    ; max circular queue = 8 dummys
    ld hl, VAR_DUMMY.DATA                        ; start of variable dummy circular queue
    ld (VAR_DUMMY.POINTER), hl
    ld b, VAR_DUMMY.LENGTH
    ld c, 0
INITIALIZE_DUMMY.1:
    ld (hl), a
    inc hl
    djnz INITIALIZE_DUMMY.1
    ret

INITIALIZE_DATA:
    ld hl, DATA_ITEMS
    ld (BASIC_DATPTR), hl        ; next DATA pointer to use by READ command
    ld hl, 0
    ld (BASIC_DATLIN), hl        ; index of DATA item to use by READ command
    ret

INITIALIZE_VARIABLES:
    call INITIALIZE_DATA
    call INITIALIZE_DUMMY

    if defined SCREEN
       call gfxInitSpriteCollisionTable
    endif

    ;if defined COMPILE_TO_ROM
    ;   ld ix, BIOS_JIFFY            ; initialize rom clock
    ;   di
    ;     ld (ix), 0
    ;     ld (ix+1), 0
    ;   ei
    ;endif
)";

const char * Assembler::variables_routines = R"(
;---------------------------------------------------------------------------------------------------------
; VARIABLES ROUTINES
;---------------------------------------------------------------------------------------------------------

; input hl = variable address
; input bc = variable name
; input d =  variable type
INIT_VAR:          ld (hl), d    ; variable type
                   inc hl
                   ld (hl), c    ; variable name 1
                   inc hl
                   ld (hl), b    ; variable name 2
                   ld a, d
                   cp 3
                   jp nz, CLEAR.VAR
                   ld de, LIT_NULL_STR
                   inc hl
                   ld (hl), 0
                   inc hl
                   ld (hl), e
                   inc hl
                   ld (hl), d
                   ld b, 5
                   jr CLEAR.VAR.LOOP
CLEAR.VAR:         ld b, 8
CLEAR.VAR.LOOP:    inc hl
                   ld (hl), 0    ; data address/value
                   djnz CLEAR.VAR.LOOP
                   ret
; input HL = variable address
; input A = variable output type
; output HL = casted data address
CAST_TO:           cp 2
                   jp z, CAST_TO.INT
                   cp 3
                   jp z, CAST_TO.STR
                   cp 4
                   jp z, CAST_TO.SGL
                   cp 8
                   jp z, CAST_TO.DBL
                   ret
; input HL = variable address
; output HL = variable address
CAST_TO.INT:       ;push af
                     ld a, (HL)
                     cp 2
                     jp z, GET_INT.ADDR
                     cp 3
                     jp z, CAST_STR_TO.INT
                     cp 4
                     jp z, CAST_SGL_TO.INT
                     cp 8
                     jp z, CAST_DBL_TO.INT
                   ;pop af
                   ret
; input HL = variable address
; output HL = variable address
CAST_TO.STR:       ;push af
                     ld a, (HL)
                     cp 2
                     jp z, CAST_INT_TO.STR
                     cp 3
                     jp z, GET_STR.ADDR
                     cp 4
                     jp z, CAST_SGL_TO.STR
                     cp 8
                     jp z, CAST_DBL_TO.STR
                   ;pop af
                   ret
; input HL = variable address
; output HL = variable address
CAST_TO.SGL:       ;push af
                     ld a, (HL)
                     cp 2
                     jp z, CAST_INT_TO.SGL
                     cp 3
                     jp z, CAST_STR_TO.SGL
                     cp 4
                     jp z, GET_SGL.ADDR
                     cp 8
                     jp z, CAST_DBL_TO.SGL
                   ;pop af
                   ret
; input HL = variable address
; output HL = variable address
CAST_TO.DBL:       ;push af
                     ld a, (hl)
                     cp 2
                     jp z, CAST_INT_TO.DBL
                     cp 3
                     jp z, CAST_STR_TO.DBL
                     cp 4
                     jp z, CAST_SGL_TO.DBL
                     cp 8
                     jp z, GET_DBL.ADDR
                   ;pop af
                   ret
CAST_SGL_TO.STR:                           ; same as CAST_DBL_TO.STR
CAST_DBL_TO.STR:   call COPY_TO.DAC
                   __call_math MATH_FOUT    ; convert DAC to string
                   jr CAST_INT_TO.STR.1
CAST_INT_TO.STR:   call COPY_TO.DAC
                   if defined BCD_MATH
                     ld hl, (BASIC_DAC+2)
                     ld (BASIC_DAC), hl
                     ld hl, BASIC_FBUFFR      ; integer as string
                     __call_math MATH_FOUT
                   else
                     __call_math MATH_FOUT    ; convert DAC to string
                   endif
CAST_INT_TO.STR.1: call COPY_TO.TEMP_STR
                   ret
CAST_INT_TO.SGL:   call COPY_TO.DAC
                   __call_math MATH_FRCSGL
                   ld hl, BASIC_DAC
                   ret
CAST_INT_TO.DBL:   call COPY_TO.DAC
                   __call_math MATH_FRCDBL
                   ld hl, BASIC_DAC
                   ret
CAST_SGL_TO.INT:                           ; same as CAST_DBL_TO.INT
CAST_DBL_TO.INT:   call COPY_TO.DAC
                   __call_math MATH_FRCINT
                   ld hl, BASIC_DAC
                   ret
CAST_STR_TO.INT:   ld a, 2
                   call CAST_STR_TO.VAL    ;
                   if defined BCD_MATH
                      __call_math MATH_FRCINT
                   endif
                   ld hl, BASIC_DAC        ;
                   ret                     ;
CAST_STR_TO.SGL:   ld a, 4
                   call CAST_STR_TO.VAL    ;
                   if defined BCD_MATH
                      __call_math MATH_FRCSGL
                   endif
                   ld hl, BASIC_DAC        ;
                   ret                     ;
CAST_STR_TO.DBL:   ld a, 8
                   call CAST_STR_TO.VAL    ;
                   if defined BCD_MATH
                      __call_math MATH_FRCDBL
                   endif
                   ld hl, BASIC_DAC        ;
                   ret                     ;
CAST_STR_TO.VAL:   ld (BASIC_VALTYP), a
                   call GET_STR.ADDR       ;
                   ld a, (hl)              ;
                   __call_math MATH_FIN    ; convert string to a value type
                   ld hl, BASIC_DAC        ;
                   ret                     ;
GET_INT.VALUE:     inc hl                  ; output BC with integer value
                   inc hl                  ;
                   ld c, (hl)              ;
                   inc hl                  ;
                   ld b, (hl)              ;
                   ret                     ;
CAST_SGL_TO.DBL:                           ; same as GET_DBL.ADDR
CAST_DBL_TO.SGL:                           ; same as GET_DBL.ADDR
GET_INT.ADDR:                              ; same as GET_DBL.ADDR
GET_SGL.ADDR:                              ; same as GET_DBL.ADDR
GET_DBL.ADDR:      inc hl
                   inc hl
                   inc hl
                   ;pop af
                   ret
GET_STR.ADDR:      push hl
                   pop ix
				   ld a, (ix + 3)
                   ld l, (ix + 4)
                   ld h, (ix + 5)
                   ret
; input hl = string address
; output a = string length
GET_STR.LENGTH:    push hl
                     ld bc, 255
                     xor a
                     cpir      ; hl = source, a = compare char, bc = count
                   pop hl
                   ret nz      ; not found
                   ld a, c
                   cpl
                   dec a
                   ret
STRING.COMPARE:    ld ix, (BASIC_DAC+1)     ; string 1
                   ld iy, (BASIC_ARG+1)     ; string 2
STRING.COMPARE.NX: ld a, (ix)               ; next char from string 1
                   cp (iy)                  ; char s1 = char s2?
                   jr nz, STRING.COMPARE.NE ; if not equal...
                   cp 0                     ;
                   jr z, STRING.COMPARE.F1  ; if string 1 has finished...
                   ld a, (iy)               ; next char from string 2
                   cp 0                     ;
                   jr z, STRING.COMPARE.GT  ; if s2 has finished, s1 has not finished yet, so s1 is greater than s2
                   inc ix                   ;
                   inc iy                   ;
                   jr STRING.COMPARE.NX     ; get next char pair
STRING.COMPARE.F1: ld a, (iy)               ; verify if string 2 has finished too
                   cp 0                     ;
                   jr z, STRING.COMPARE.EQ  ; if s2 has finished, then they are equals
                   jr STRING.COMPARE.LT     ; else, result = s1 is less than s2
STRING.COMPARE.NE: jr c, STRING.COMPARE.GT  ; verify if s1 is greater than s2...
STRING.COMPARE.LT: ld a, 1                  ; ...else, result = s1 less than s2
                   ret                      ;
STRING.COMPARE.GT: ld a, 0xFF               ; result = s1 is greater than s2
                   ret                      ;
STRING.COMPARE.EQ: xor a                    ; result = s1 is equal to s2
                   ret                      ;
STRING.CONCAT:     ld ix, BASIC_DAC           ; s1 size
				   ld a, (BASIC_ARG)          ; s2 size
				   add a, (ix)                ; s3 size = s1 size + s2 size
				   push af
                     ld b, 0
                     ld c, a                    ;
                     inc bc                     ; add 1 byte to size
                     call memory.alloc          ; in bc size, out ix new memory address, nz=OK
                     jp z, memory.error         ;
                     push ix                    ; save ix
                       push ix                  ; save ix
                       pop de                   ; de = ix
					   ld a, (BASIC_DAC)        ; s1 size
                       ld hl, (BASIC_DAC + 1)   ; string 1
                       call COPY_TO.STR         ; copy to new memory
					   ld a, (BASIC_ARG)        ; s2 size
                       ld hl, (BASIC_ARG + 1)   ; string 2
                       call COPY_TO.STR         ; copy to new memory
					   xor a
					   ld (de), a               ; null terminated
                     pop hl                     ; hl = ix
                   pop af
                   call COPY_TO.VAR_DUMMY.STR ;
                   ret.parm                   ; WARNING - VERIFY STRING MEMORY LEAKs
STRING.PRINT:      ld a, (BIOS_SCRMOD)        ; 0=40x24 Text Mode, 1=32x24 Text Mode, 2=Graphics Mode, 3=Multicolour Mode
                   cp 5                       ;
                   jr nc, STRING.PRINT.G2     ; jump if graphic screen mode MSX2 (>=5)
                   cp 2                       ;
                   jr nc, STRING.PRINT.G1     ; jump if graphic screen mode MSX1 (>=2)
STRING.PRINT.T:    ld a, (hl)                 ; get a char from a string parameter
                   or a                       ; cp 0 - is it the string end?
                   ret z                      ; exit if yes
                   __call_bios BIOS_CHPUT     ; put the char (a) into text screen
                   inc hl                     ; next char
                   jr STRING.PRINT.T          ; repeat
STRING.PRINT.G1:   ld a, (hl)                 ; get a char from a string parameter
                   or a                       ; cp 0 - is it the string end?
                   ret z                      ; exit if yes
                   __call_bios BIOS_GRPPRT    ; put the char (a) into graphical screen
                   inc hl                     ; next char
                   jr STRING.PRINT.G1         ; repeat
STRING.PRINT.G2:   ld a, (hl)                 ; get a char from a string parameter
                   or a                       ; cp 0 - is it the string end?
                   ret z                      ; exit if yes
                   ld ix, BIOS_GRPPRT2        ; put the char (a) into graphical screen
                   call BIOS_EXTROM
                   inc hl                     ; next char
                   jr STRING.PRINT.G2         ; repeat

; a = string size to copy
; input hl = string from
; input de = string to
COPY_TO.STR:       or a
                   ret z                      ; avoid copy if size = zero
                   ld b, 0
                   ld c, a                    ; string size
                   ldir                       ; copy bc bytes from hl to de
                   ret                        ;
COPY_TO.BASIC_BUF: ld bc, BASIC_BUF
                   ld a, (LIT_QUOTE_CHAR)
                   ld (bc), a
                   inc bc
COPY_BAS_BUF.LOOP: ld a, (hl)
                   or a                      ; cp 0
                   jr z, COPY_BAS_BUF.EXIT
                   ld (bc), a
                   inc bc
                   inc hl
                   jr COPY_BAS_BUF.LOOP
COPY_BAS_BUF.EXIT: ld a, (LIT_QUOTE_CHAR)
                   ld (bc), a
                   inc bc
                   xor a
                   ld (bc), a
                   ld hl, BASIC_BUF
                   ret
COPY_TO.VAR_DUMMY:     ld a, (BASIC_VALTYP)    ; create dummy variable from VALTYPE
                       cp 3                    ;
                       jr nz, COPY_TO.VAR_DUMMY.DBL
                       call GET_STR.LENGTH     ; get string length, in hl = address, out a = size
COPY_TO.VAR_DUMMY.STR: call GET_VAR_DUMMY.ADDR ; create dummy string variable from HL
                       ld (ix), 3            ; data type string
                       ld (ix+1), 0          ;
                       ld (ix+2), 255        ; var type fixed
                       ld (ix+3), a          ; string length
                       ld (ix+4), l          ; data address low
                       ld (ix+5), h          ; data address high
                       push ix               ; output var address...
                       pop hl                ; ...into hl
                       ret                   ;
COPY_TO.VAR_DUMMY.INT: call GET_VAR_DUMMY.ADDR ; create dummy integer variable from BC
                       ld (ix),    2           ; data type string
                       ld (ix+1),  0           ;
                       ld (ix+2),  0           ;
                       ld (ix+3),  0           ;
                       ld (ix+4),  0           ;
                       ld (ix+5),  c           ;
                       ld (ix+6),  b           ;
                       ld (ix+7),  0           ;
                       ld (ix+8),  0           ;
                       ld (ix+9),  0           ;
                       ld (ix+10), 0           ;
                       push ix                 ; output var address...
                       pop hl                  ; ...into hl
                       ret                     ;
COPY_TO.VAR_DUMMY.DBL: call GET_VAR_DUMMY.ADDR  ; create dummy value variable from DAC
                       ld (ix), a            ; data type
                       ld (ix+1), 0          ;
                       ld (ix+2), 0          ;
                       ld bc, 8              ;
                       ld hl, BASIC_DAC      ;
                       push ix               ; just to copy ix to de
                       pop de                ;
                       inc de                ;
                       inc de                ;
                       inc de                ;
                       ldir                  ; copy bc bytes from hl (data address) to de (variable address)
                       push ix               ; output var address...
                       pop hl                ; ...into hl
                       ret                   ;
GET_VAR_DUMMY.ADDR:    push af                       ;
                       push de
                         ld de, 11                   ;
                         ld ix, (VAR_DUMMY.POINTER)  ;
                         ld a, (VAR_DUMMY.COUNTER)   ;
GET_VAR_DUMMY.NEXT:      add ix, de                  ;
                         inc a                       ;
                         cp VAR_DUMMY.SIZE           ;
                         jr nz, GET_VAR_DUMMY.EXIT   ;
                           xor a                     ;
                           ld ix, VAR_DUMMY.DATA     ;
GET_VAR_DUMMY.EXIT:      ld (VAR_DUMMY.POINTER), ix  ;
                         ld (VAR_DUMMY.COUNTER), a   ;
						 ld a, (ix)                  ; get last var dummy type
						 cp 3                        ; is it string?
						 call z, GET_VAR_DUMMY.FREE  ; free string memory
                       pop de
                       pop af                        ;
                       ret                           ;
GET_VAR_DUMMY.FREE:
                   push hl
                   push ix
					 ld a, (ix+3)                    ; string size
				     ld l, (ix+4)                    ; get string data address
					 ld h, (ix+5)
					 push hl
					 pop ix
					 or a
                     call nz, memory.free            ; free memory
				   pop ix
				   pop hl
				   ret
; input hl = variable address
COPY_TO.DAC:       ld de, BASIC_DAC
COPY_TO.DAC.DATA:  ld a, (hl)
                   ld (BASIC_VALTYP), a
                   inc hl
                   inc hl
                   inc hl
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret
COPY_TO.ARG:       ld de, BASIC_ARG        ;
                   jr COPY_TO.DAC.DATA     ;
COPY_TO.DAC_ARG:   ld hl, BASIC_DAC        ;
                   ld de, BASIC_ARG        ;
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret                     ;
COPY_TO.ARG_DAC:   ld hl, BASIC_ARG        ;
                   ld de, BASIC_DAC        ;
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret                     ;
COPY_TO.DAC_TMP:   ld hl, BASIC_DAC        ;
                   ld de, BASIC_SWPTMP     ;
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret                     ;
COPY_TO.TMP_DAC:   ld hl, BASIC_SWPTMP     ;
                   ld de, BASIC_DAC        ;
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret                     ;
SWAP.DAC.ARG:      di
                   exx                     ; save registers
                     ld bc, 8              ;
                     ld hl, BASIC_DAC      ;
                     ld de, BASIC_SWPTMP   ;
                     ldir                  ; copy bc bytes from hl to de
                     ld bc, 8              ;
                     ld hl, BASIC_ARG      ;
                     ld de, BASIC_DAC      ;
                     ldir                  ; copy bc bytes from hl to de
                     ld bc, 8              ;
                     ld hl, BASIC_SWPTMP   ;
                     ld de, BASIC_ARG      ;
                     ldir                  ; copy bc bytes from hl to de
                   exx                     ; restore registers
                   ei
                   ret                     ;
CLEAR.DAC:         ld de, BASIC_DAC
CLEAR.DAC.DATA:    ld hl, BASIC_VALTYP
                   ld (hl), 2
                   ld hl, LIT_NULL_DBL
                   ld bc, 8                ; data = 8 bytes
                   ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                   ret
CLEAR.ARG:         ld de, BASIC_ARG
                   jr CLEAR.DAC.DATA

;--------------
; array struct:
;   vartype with array flag 0xF0 (byte) - 2nd dimension factor (word) - elements count (word) - data pointer (word)
;--------------

if defined ERASE or defined DIM1 or defined DIM2 or defined REDIM1 or defined REDIM2

; hl = array variable
ARRAY.ERASE:       ld a, (hl)
                   ld c, a
                   and 0xF0                    ; test if its an array
                   ret z

                   ld a, c
                   and 0x0F      ; var type

                   ; free data
                   push hl
                     inc hl      ; 2nd dimension factor
                     inc hl
                     inc hl
                     ld c, (hl)
                     inc hl
                     ld b, (hl)  ; elements count
                     inc hl
                     ld e, (hl)
                     inc hl
                     ld d, (hl)  ; data pointer
                     ex af, af'
                     ld a, d
                     or e
                     jr z, ARRAY.ERASE.continue
                       ex af, af'
                       push de
                         cp 3                    ; test if its a string array
                         call z, ARRAY.ERASE.string
                       pop ix
                       call memory.free
ARRAY.ERASE.continue:
                   pop hl

                   ; clear variable
                   ld a, (hl)
                   and 0x0F
                   ld (hl), a

                   inc hl
                   xor a
                   ld (hl), a

                   ld e, l
                   ld d, h
                   inc de
                   ld bc, 9
                   ldir        ; (de) = (hl), until bc = 0

                   ret

; BC = elements count
; DE = array data
ARRAY.ERASE.string:
                   push de
                     ld l, e
                     ld h, d

ARRAY.ERASE.str.next:
                     ld a, (hl)   ; string size
                     inc hl
                     ld e, (hl)
                     inc hl
                     ld d, (hl)   ; string pointer
                     inc hl
                     inc hl       ; next element
                     push hl
                     push bc
                       or 0
                       jr z, ARRAY.ERASE.str.cont
                         push de
                         pop ix
                         ld a, d
                         or e
                         call nz, memory.free
ARRAY.ERASE.str.cont:
                     pop bc
                     pop hl
                     dec bc
                     ld a, b
                     or c
                     jr nz, ARRAY.ERASE.str.next
                   pop de
                   ret

endif

if defined DIM1 or defined DIM2 or defined REDIM1 or defined REDIM2

; hl = array variable
; de = 2nd dimension factor
; bc = elements count
ARRAY.DIM:         push hl
                   push de
                   push bc
                     call ARRAY.ERASE
                   pop bc
                   pop de
                   pop hl
ARRAY.DIM.init:    ld a, (hl)
                   or 0xF0
                   ld (hl), a     ; set array flag on vartype

                   inc hl
                   ld (hl), e
                   inc hl
                   ld (hl), d     ; 2nd dimension factor
                   inc hl
                   ld (hl), c
                   inc hl
                   ld (hl), b     ; elements count
                   inc hl
                   push hl
                     ld l, c
                     ld h, b      ; elements count
                     add hl, hl   ; size * 2
                     add hl, hl   ; size * 4   ; data size (elements count * 4)
                     ld c, l
                     ld b, h
                     push bc
                       push bc
                         call memory.alloc ; in bc = size, out ix = address, nz=OK
		                 jp z, memory.error
                       pop bc
                       push ix
                         push ix  ; clear array data
                         pop de
                         ld l, e
                         ld h, d
                         xor a
                         ld (hl), a
                         inc de
                         dec bc
                         ldir
                       pop de
                     pop bc
                   pop hl
                   ld (hl), e   ; data address
                   inc hl
                   ld (hl), d
                   ret

endif

if defined REDIM1 or defined REDIM2

; hl = array variable
; de = 2nd dimension factor
; bc = elements count
ARRAY.REDIM:       push hl
                   push de
                     inc hl       ; 2nd dimension factor
                     inc hl
                     inc hl
                     inc hl       ; elements count
                     inc hl
                     inc hl
                     ld e, (hl)
                     inc hl
                     ld d, (hl)   ; old array data
                     ld (BIOS_TEMP), de
                   pop de
                   pop hl
                   call ARRAY.DIM.init
                   ld hl, (BIOS_TEMP)
                   ldir    ; de = destination, bc = bytes count
                   ld ix, (BIOS_TEMP)
                   call memory.free   ; memory leak here on string arrays shortened
                                      ; bug at bidimensional arrays resized (data order change)
                   ret

endif

)";

const char * Assembler::math16bits_routines = R"(
;---------------------------------------------------------------------------------------------------------
; MATH 16 BITS ROUTINES
;---------------------------------------------------------------------------------------------------------

MATH.PARM.POP:  pop af                       ; get PC from caller stack
                ex af, af'                   ; save PC to temp
                  pop.parm                   ; get first parameter
                  call COPY_TO.ARG           ; put HL in ARG (return var type in A)
                  pop.parm                   ; get second parameter
                ex af, af'                   ; restore PC from temp
                push af                      ; put again PC from caller in stack
                ex af, af'                   ; restore 1st data type
                push af                      ; save 1st data type
                  call COPY_TO.DAC           ; put HL in DAC (return var type in A)
                pop bc                       ; restore 1st data type (ARG) in B
                cp b                         ; test if data type in A (DAC) = data type in B (ARG)
                ret z                        ; return if is equal data types
MATH.PARM.CAST: push bc                      ; else cast both to double
                  and 12                     ; test if single/double
                  jr nz, MATH.PARM.CST1      ; avoid cast if already single/double
                  __call_math MATH_FRCDBL    ; convert DAC to double
MATH.PARM.CST1: pop af                       ;
                and 12                       ; test if single/double
                jr nz, MATH.PARM.CST2        ; avoid cast if already single/double
                ld (BASIC_VALTYP), a         ;
                call COPY_TO.DAC_TMP         ;
                call COPY_TO.ARG_DAC         ;
                __call_math MATH_FRCDBL      ; convert ARG to double
                call COPY_TO.DAC_ARG         ;
                call COPY_TO.TMP_DAC         ;
MATH.PARM.CST2: ld a, 8                      ;
                ld (BASIC_VALTYP), a         ;
                ret                          ;
MATH.PARM.POP.INT:                           ; return result in DAC/ARG as integer
                pop af                       ; get PC from caller stack
                  ex af, af'                 ; save PC to temp
                    pop.parm                 ; get first parameter
                    ld a, (hl)               ; get parameter type
                    and 2                    ; test if integer
                    jr z, MATH.PARM.POP.I1   ; do cast if not integer
                    call COPY_TO.ARG         ; put HL in ARG (return var type in A)
                    jr MATH.PARM.POP.I2      ; go to next parameter
MATH.PARM.POP.I1:   call COPY_TO.DAC         ; put HL in DAC (return var type in A)
                    __call_math MATH_FRCINT  ; convert DAC to int
                    call COPY_TO.DAC_ARG     ; copy DAC to ARG
MATH.PARM.POP.I2:   pop.parm                 ; get second parameter
                    call COPY_TO.DAC         ; put HL in DAC (return var type in A)
                    and 2                    ; test if integer
                    jr nz, MATH.PARM.POP.I3  ; avoid cast if already integer
                    __call_math MATH_FRCINT  ; convert DAC to int
                    ld a, 2                  ;
                    ld (BASIC_VALTYP), a     ;
MATH.PARM.POP.I3:
                    ex af, af'                 ; restore PC from temp
                push af                      ; put again PC from caller in stack
                ret                          ;
MATH.PARM.PUSH: call COPY_TO.VAR_DUMMY       ;
                ret.parm                     ;

if defined MATH.ADD

; input DAC, ARG
; output in parm stack
; http://www.z80.info/zip/zaks_book.pdf - page 104
MATH.ADD.INT:  ld hl, (BASIC_DAC+2)  ;
               ld bc, (BASIC_ARG+2)  ;
               add hl, bc            ;
               ld (BASIC_DAC+2), hl  ;
               jp MATH.PARM.PUSH     ;

endif

if defined MATH.SUB or defined MATH.NEG

; input DAC, ARG
; output in parm stack
; http://www.z80.info/zip/zaks_book.pdf - page 104
MATH.SUB.INT:  ld hl, (BASIC_DAC+2)  ;
               ld de, (BASIC_ARG+2)  ;
               and a                 ; clear carry
               sbc hl, de            ;
               ld (BASIC_DAC+2), hl  ;
               jp MATH.PARM.PUSH     ;

endif

if defined MATH.MULT or defined ARRAY.DIM

; input DAC, ARG
; output in parm stack
MATH.MULT.INT: ld hl, (BASIC_DAC+2)  ;
               ld bc, (BASIC_ARG+2)  ;
               call MATH.MULT.16     ;
               ld (BASIC_DAC+2), hl  ;
               jp MATH.PARM.PUSH     ;

; input HL = multiplicand
; input BC = multiplier
; output HL = result
; http://www.z80.info/zip/zaks_book.pdf - page 131
MATH.MULT.16:  ld a, c				; low multiplier
               ld c, b				; high multiplier
               ld b, 16
               ld d, h 			; multiplicand
               ld e, l
               ld hl, 0
MULT16LOOP:    srl c 				; right shift multiplier high
               rra					; rotate right multiplier low
               jr nc, MULT16NOADD	; test carry
               add hl, de			; add multiplicand to result
MULT16NOADD:   ex de, hl
               add hl, hl			; double - shift multiplicand
               ex de, hl
               djnz MULT16LOOP
               ret

endif

if defined MATH.DIV or defined MATH.IDIV or defined MATH.MOD

; input AC = dividend
; input DE = divisor
; output AC = quotient
; output HL = remainder
; http://www.z80.info/zip/zaks_book.pdf - page 140
MATH.DIV.16:   ld hl, 0				; clear accumulator
               ld b, 16				; set counter
DIV16LOOP:     rl c					; rotate accumulator result left
               rla
               adc hl, hl				; left shift
               sbc hl, de				; trial subtract divisor
               jr nc, $ + 3			; subtract was OK ($ = current location)
               add hl, de				; restore accumulator
               ccf						; calculate result bit
               djnz DIV16LOOP			; counter not zero
               rl c 					; shift in last result bit
               rla
               ret

endif

if defined GFX_FAST or defined LINE

; compare two signed 16 bits integers
; HL < DE: Carry flag
; HL = DE: Zero flag
; http://www.z80.info/zip/zaks_book.pdf - page 531
MATH.COMP.S16: ld a, h                       ; test high order byte
               and 0x80                      ; test sign, clear carry
			   jr nz, MATH.COMP.S16.NEGM1    ; jump if hl is negative
			   bit 7, d
			   ret nz                        ; de is negative (and hl is positive)
			   ld a, h
			   cp d                          ; signs are both positive, so normal compare
			   ret nz
			   ld a, l                       ; test low order byte
			   cp e
               ret
MATH.COMP.S16.NEGM1:
               xor d
               rla                           ; sign bit into carry
               ret c                         ; signs different
               ld a, h
               cp d                          ; both signs negative
			   ret nz
			   ld a, l
			   cp e
			   ret

endif

if defined MATH.ADD

MATH.ADD.SGL:  ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.ADD.DBL:  __call_math MATH_DECADD  ;
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.SUB or defined MATH.NEG

MATH.SUB.SGL:  ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.SUB.DBL:  __call_math MATH_DECSUB  ;
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.MULT

MATH.MULT.SGL: ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.MULT.DBL: __call_math MATH_DECMUL  ;
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.DIV

; input DAC, ARG
; output in parm stack
MATH.DIV.INT:  __call_math MATH_FRCDBL  ; convert DAC to double
               call SWAP.DAC.ARG        ;
               ld a, 2                  ;
               ld (BASIC_VALTYP), a     ;
               __call_math MATH_FRCDBL  ; convert ARG to double
               call SWAP.DAC.ARG        ;
MATH.DIV.SGL:  ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.DIV.DBL:  __call_math MATH_DECDIV  ;
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.IDIV

; input DAC, ARG
; output in parm stack
MATH.IDIV.SGL: ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.IDIV.DBL: __call_math MATH_FRCINT  ; convert DAC to integer
               call SWAP.DAC.ARG        ;
               ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
               __call_math MATH_FRCINT  ; convert ARG to integer
               call SWAP.DAC.ARG        ;
MATH.IDIV.INT: ld hl, (BASIC_DAC+2)     ;
               ld a, h                  ;
               ld c, l                  ;
               ld de, (BASIC_ARG+2)     ;
               call MATH.DIV.16         ;
               ld h, a                  ;
               ld l, c                  ;
               ld (BASIC_DAC+2), hl     ; quotient
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.POW

MATH.POW.INT:  ld (BASIC_VALTYP), a     ;
               __call_math MATH_FRCDBL  ; convert DAC to double
               call SWAP.DAC.ARG        ;
               ld a, 2                  ;
               ld (BASIC_VALTYP), a     ;
               __call_math MATH_FRCDBL  ; convert ARG to double
               call SWAP.DAC.ARG        ;
MATH.POW.SGL:  ld a, 8                  ;
               ld (BASIC_VALTYP), a     ;
MATH.POW.DBL:  __call_math MATH_DBLEXP  ;
               jp MATH.PARM.PUSH        ;

endif

if defined MATH.MOD

;MATH.MOD.SGL:  ld a, 8                  ;
;               ld (BASIC_VALTYP), a     ;
;MATH.MOD.DBL:  __call_bios MATH_FRCINT  ; convert DAC to integer
;               call SWAP.DAC.ARG        ;
;		         ld a, 8                  ;
;               ld (BASIC_VALTYP), a     ;
;               __call_bios MATH_FRCINT  ; convert ARG to integer
;               call SWAP.DAC.ARG        ;
MATH.MOD.INT:  ld hl, (BASIC_DAC+2)     ;
               ld a, h                  ;
               ld c, l                  ;
               ld de, (BASIC_ARG+2)     ;
               call MATH.DIV.16         ;
               ld (BASIC_DAC+2), hl     ; remainder
               jp MATH.PARM.PUSH        ;

endif

if defined ISQR

; fast 16-bit integer square root
; http://www.retroprogramming.com/2017/07/a-fast-z80-integer-square-root.html
; 92 bytes, 344-379 cycles (average 362)
; v2 - 3 t-state optimization spotted by Russ McNulty
; call with hl = number to square root
; returns    a = square root
; corrupts  hl, de

MATH.INT.SQR:
  ld a,h
  ld de,0B0C0h
  add a,e
  jr c,sq7
  ld a,h
  ld d,0F0h
sq7:
  add a,d
  jr nc,sq6
  res 5,d
  db 254
sq6:
  sub d
  sra d
  set 2,d
  add a,d
  jr nc,sq5
  res 3,d
  db 254
sq5:
  sub d
  sra d
  inc d
  add a,d
  jr nc,sq4
  res 1,d
  db 254
sq4:
  sub d
  sra d
  ld h,a
  add hl,de
  jr nc,sq3
  ld e,040h
  db 210
sq3:
  sbc hl,de
  sra d
  ld a,e
  rra
  or 010h
  ld e,a
  add hl,de
  jr nc,sq2
  and 0DFh
  db 218
sq2:
  sbc hl,de
  sra d
  rra
  or 04h
  ld e,a
  add hl,de
  jr nc,sq1
  and 0F7h
  db 218
sq1:
  sbc hl,de
  sra d
  rra
  inc a
  ld e,a
  add hl,de
  jr nc,sq0
  and 0FDh
sq0:
  sra d
  rra
  cpl
  ret

endif

if defined RANDOMIZE or defined SEED

MATH.RANDOMIZE:    di                          ;
                     ld bc, (BIOS_JIFFY)       ;
                   ei                          ;

MATH.SEED:         ld (BASIC_RNDX), bc         ; seed to IRND
                   push bc                     ; in bc = new integer seed
                     call CLEAR.DAC            ;
                   pop bc                      ;
                   ;ld ix, BASIC_DAC            ;
                   ld (BASIC_DAC+2), bc        ; copy bc to dac
                   ld a, 2                     ; type integer
                   ld (BASIC_VALTYP), a        ;
                   __call_math MATH_FRCDBL     ; convert DAC integer to DAC double
                   __call_math MATH_NEG        ; DAC = -DAC
                   __call_math MATH_RND        ; put in DAC a new random number from previous DAC parameter
                   ret                         ;

endif

MATH.ERROR:        ld e, 13                          ; type mismatch
                   __call_basic BASIC_ERROR_HANDLER  ;
                   ret


;---------------------------------------------------------------------------------------------------------
; BOOLEAN ROUTINES
;---------------------------------------------------------------------------------------------------------

BOOLEAN.RET.TRUE:  ld hl, LIT_TRUE             ;
                   ret.parm                    ;
BOOLEAN.RET.FALSE: ld hl, LIT_FALSE            ;
                   ret.parm                    ;
BOOLEAN.CMP.INT:   ld hl, (BASIC_DAC+2)        ;
                   ld de, (BASIC_ARG+2)        ;
                   __call_math MATH_ICOMP      ;
                   ret                         ;
BOOLEAN.CMP.SGL:   ld bc, (BASIC_ARG)          ;
                   ld de, (BASIC_ARG+2)        ;
                   __call_math MATH_DCOMP      ;
                   ret                         ;
BOOLEAN.CMP.DBL:   __call_math MATH_XDCOMP     ;
                   ret                         ;
BOOLEAN.CMP.STR:   call STRING.COMPARE         ;
                   ret                         ;

if defined BOOLEAN.GT

BOOLEAN.GT.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.GT.RET           ;
BOOLEAN.GT.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.GT.RET           ;
BOOLEAN.GT.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.GT.RET           ;
BOOLEAN.GT.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.GT.RET           ;
BOOLEAN.GT.RET:    cp 0x01                     ;
                   jp z, BOOLEAN.RET.TRUE      ;
                   jp BOOLEAN.RET.FALSE        ;
endif

if defined BOOLEAN.LT

BOOLEAN.LT.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.LT.RET           ;
BOOLEAN.LT.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.LT.RET           ;
BOOLEAN.LT.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.LT.RET           ;
BOOLEAN.LT.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.LT.RET           ;
BOOLEAN.LT.RET:    cp 0xFF                     ;
                   jp z, BOOLEAN.RET.TRUE      ;
                   jp BOOLEAN.RET.FALSE        ;

endif

if defined BOOLEAN.GE

BOOLEAN.GE.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.GE.RET           ;
BOOLEAN.GE.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.GE.RET           ;
BOOLEAN.GE.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.GE.RET           ;
BOOLEAN.GE.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.GE.RET           ;
BOOLEAN.GE.RET:    cp 0x01                     ;
                   jp z, BOOLEAN.RET.TRUE      ;
                   or a                        ; cp 0
                   jp z, BOOLEAN.RET.TRUE      ;
                   jp BOOLEAN.RET.FALSE        ;

endif

if defined BOOLEAN.LE

BOOLEAN.LE.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.LE.RET           ;
BOOLEAN.LE.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.LE.RET           ;
BOOLEAN.LE.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.LE.RET           ;
BOOLEAN.LE.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.LE.RET           ;
BOOLEAN.LE.RET:    cp 0xFF                     ;
                   jp z, BOOLEAN.RET.TRUE      ;
                   or a                        ; cp 0
                   jp z, BOOLEAN.RET.TRUE      ;
                   jp BOOLEAN.RET.FALSE        ;

endif

if defined BOOLEAN.NE

BOOLEAN.NE.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.NE.RET           ;
BOOLEAN.NE.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.NE.RET           ;
BOOLEAN.NE.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.NE.RET           ;
BOOLEAN.NE.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.NE.RET           ;
BOOLEAN.NE.RET:    or a                        ; cp 0
                   jp nz, BOOLEAN.RET.TRUE     ;
                   jp BOOLEAN.RET.FALSE        ;

endif

if defined BOOLEAN.EQ

BOOLEAN.EQ.INT:    call BOOLEAN.CMP.INT        ;
                   jr BOOLEAN.EQ.RET           ;
BOOLEAN.EQ.STR:    call BOOLEAN.CMP.STR        ;
                   jr BOOLEAN.EQ.RET           ;
BOOLEAN.EQ.SGL:    call BOOLEAN.CMP.SGL        ;
                   jr BOOLEAN.EQ.RET           ;
BOOLEAN.EQ.DBL:    call BOOLEAN.CMP.DBL        ;
                   jr BOOLEAN.EQ.RET           ;
BOOLEAN.EQ.RET:    or a                        ; cp 0
                   jp z, BOOLEAN.RET.TRUE      ;
                   jp BOOLEAN.RET.FALSE        ;

endif

if defined BOOLEAN.AND

BOOLEAN.AND.INT:   ld a, (BASIC_DAC+2)         ;
                   ld hl, BASIC_ARG+2          ;
                   and (hl)                    ;
                   ld (BASIC_DAC+2), a         ;
                   inc hl                      ;
                   ld a, (BASIC_DAC+3)         ;
                   and (hl)                    ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif

if defined BOOLEAN.OR

BOOLEAN.OR.INT:    ld a, (BASIC_DAC+2)         ;
                   ld hl, BASIC_ARG+2          ;
                   or (hl)                     ;
                   ld (BASIC_DAC+2), a         ;
                   inc hl                      ;
                   ld a, (BASIC_DAC+3)         ;
                   or (hl)                     ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif

if defined BOOLEAN.XOR

BOOLEAN.XOR.INT:   ld a, (BASIC_DAC+2)         ;
                   ld hl, BASIC_ARG+2          ;
                   xor (hl)                    ;
                   ld (BASIC_DAC+2), a         ;
                   inc hl                      ;
                   ld a, (BASIC_DAC+3)         ;
                   xor (hl)                    ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif

if defined BOOLEAN.EQV

BOOLEAN.EQV.INT:   ld a, (BASIC_DAC+2)         ;
                   ld hl, BASIC_ARG+2          ;
                   xor (hl)                    ;
                   cpl                         ;
                   ld (BASIC_DAC+2), a         ;
                   inc hl                      ;
                   ld a, (BASIC_DAC+3)         ;
                   xor (hl)                    ;
                   cpl                         ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif

if defined BOOLEAN.IMP

BOOLEAN.IMP.INT:   ld a, (BASIC_DAC+2)         ;
                   ld hl, BASIC_ARG+2          ;
                   cpl                         ;
                   or (hl)                     ;
                   ld (BASIC_DAC+2), a         ;
                   inc hl                      ;
                   ld a, (BASIC_DAC+3)         ;
                   cpl                         ;
                   or (hl)                     ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif

if defined BOOLEAN.SHR

BOOLEAN.SHR.INT:   ld ix, BASIC_DAC+2          ; shift DAC integer to right (bits 15...0-->)
                   ld a, (BASIC_ARG+2)         ;
                   or a                        ; clear carry
                   jp z, MATH.PARM.PUSH        ; return if not shift
                   ld b, a                     ; shift count
BOOLEAN.SHR.INT.N: rr (ix+1)                   ;
                   rr (ix)                     ;
                   or a                        ; clear carry
                   djnz BOOLEAN.SHR.INT.N      ; next shift
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ; return DAC

endif

if defined BOOLEAN.SHL

BOOLEAN.SHL.INT:   ld ix, BASIC_DAC+2          ; shift DAC integer to left (<--bits 15...0)
                   ld a, (BASIC_ARG+2)         ;
                   or a                        ; clear carry
                   jp z, MATH.PARM.PUSH        ; return if not shift
                   ld b, a                     ; shift count
BOOLEAN.SHL.INT.N: rl (ix)                     ;
                   rl (ix+1)                   ;
                   or a                        ; clear carry
                   djnz BOOLEAN.SHL.INT.N      ; next shift
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ; return DAC

endif

if defined BOOLEAN.NOT

BOOLEAN.NOT.INT:   ld a, (BASIC_DAC+2)         ;
                   cpl                         ;
                   ld (BASIC_DAC+2), a         ;
                   ld a, (BASIC_DAC+3)         ;
                   cpl                         ;
                   ld (BASIC_DAC+3), a         ;
                   ld a, 2                     ;
                   jp MATH.PARM.PUSH           ;

endif
)";

// memory.heap_end: equ memory.heap_start+0x2000    ; ~8kb of free memory

const char * Assembler::memory_allocation_routines = R"(
;---------------------------------------------------------------------------------------------------------
; MEMORY ALLOCATION ROUTINES
;---------------------------------------------------------------------------------------------------------
; Adapted from memory allocator code by SamSaga2, Spain, 2015
; https://www.msx.org/forum/msx-talk/development/asm-memory-allocator
; https://www.msx.org/users/samsaga2
;---------------------------------------------------------------------------------------------------------
memory.heap_start: equ VAR_STACK.END + 1    ; start at end of variable stack
memory.heap_end:   equ 0xF0A0 - 100         ; end at start of work area for stack (100 bytes reserved), BIOS and BASIC interpreter
block.next:        equ 0                    ; next free block address
block.size:        equ 2                    ; size of block including header
block:             equ 4                    ; block.next + block.size

;; init
memory.init:
       ld ix,memory.heap_start              ; first block
       ld hl,memory.heap_start+block        ; second block
       ;; first block NEXT=secondblock, SIZE=0
       ;; with this block we have a fixed start location
       ;; because never will be allocated
       ld (ix+block.next),l
       ld (ix+block.next+1),h
       ld (ix+block.size),0
       ld (ix+block.size+1),0
       ;; second block NEXT=0, SIZE=all
       ;; the first and only free block have all available memory
       ld (ix+block.next+block),0
       ld (ix+block.next+block+1),0
       xor a
       ;ld hl,memory.heap_end          ; size = @heap_end (stack) - heap_start - block_header * 2 - 100 (buffer for stack)
	   ld (BIOS_TEMP), sp
	   ld hl, (BIOS_TEMP)
       ld de, memory.heap_start + (block * 2) + 100
       sbc hl,de
	   ;ld de, block * 2 + 100
	   ;sbc hl, de
       ld (ix+block.size+block),l
       ld (ix+block.size+block+1),h
       ret

;; alloc
;; IN BC=size, OUT IX=memptr, NZ=ok
memory.alloc:
       ld hl,block
       add hl,bc
       ;push hl
       ;pop bc
	   ld b, h
	   ld c, l
       ld ix,memory.heap_start       ; this
       ld iy,0                       ; prev
memory.alloc.find:
       ld l,(ix+block.size)
       ld h,(ix+block.size+1)
       xor a
       sbc hl,bc
       jp z, memory.alloc.exactfit
       jp c, memory.alloc.nextblock
;; split found block
memory.alloc.splitfit:
       ;; free space must allow at least two blocks headers (current + next)
	   or h
	   jr nz, memory.alloc.splitfit.do   ; if free space > 0xFF, do split
	     ld a, l
	     cp 4
	     jr c, memory.alloc.nextblock    ; if free space < 4, skip to next block
memory.alloc.splitfit.do:
       ;; newfreeblock = this + BC
       push ix
       pop hl
       add hl,bc
       ;; prevblock->next = newfreeblock
       ld (iy+block.next),l
       ld (iy+block.next+1),h
       ;; newfreeblock->next = this->next
       push hl
       pop iy                        ; iy = newfreeblock
       ld l,(ix+block.next)
       ld h,(ix+block.next+1)
       ld (iy+block.next),l
       ld (iy+block.next+1),h
       ;; newfreeblock->size = this->size - BC
       ld l,(ix+block.size)
       ld h,(ix+block.size+1)
       xor a
       sbc hl,bc
       ld (iy+block.size),l
       ld (iy+block.size+1),h
       ;; this->size = BC
       ld (ix+block.size),c
       ld (ix+block.size+1),b
       jr memory.alloc.ok
;; use whole found block
memory.alloc.exactfit:
       ;; prevblock->next = this->next - remove block from free list
       ld l,(ix+block.next)
       ld h,(ix+block.next+1)
       ld (iy+block.next),l
       ld (iy+block.next+1),h
memory.alloc.ok:
       ;; ix = first byte
       ld de,block
       add ix,de
       ;; enable z-flag
       ld a,1
       or a
       ret
memory.alloc.nextblock:
       ld l,(ix+block.next)
       ld h,(ix+block.next+1)
       ld a,l
       cp h
       ret z
         ;; prevblock = this
         push ix
         pop iy
         ;; this = this->next
         push hl
         pop ix
         jp memory.alloc.find

;; free
;; IN IX=memptr
memory.free:
       ;; HL = IX - block_header_size
       push ix
       pop hl
       ld de, block
	   xor a
       sbc hl,de
       ;; start of search
       ld ix,memory.heap_start
memory.free.find:
       ld e,(ix+block.next)
       ld d,(ix+block.next+1)
       ld a,d
       or e
       jp z, memory.free.passedend
         sbc hl,de                     ; test this (HL) against next (DE)
         jr c, memory.free.found       ; if DE > HL
           add hl,de                     ; restore hl value
	       push de
	       pop ix                        ; current = next
           jr memory.free.find

;; ix=prev, hl=this, de=next
memory.free.found:
       add hl,de                     ; restore hl value
	   ld (ix+block.next), l
	   ld (ix+block.next+1), h       ; prev->next = this
	   push hl
	   pop iy
	   ld (iy+block.next), e
	   ld (iy+block.next+1), d       ; this->next = next
	   push ix 					     ; prev x this
	   pop iy
	   push hl
	   pop ix
	   push de
	     call memory.free.coalesce
	   pop ix                        ; this x next
       jr memory.free.coalesce

;; parm1 = *next
;; parm2 = *this
memory.free.coalesce:
	   ld c, (iy+block.size)
	   ld b, (iy+block.size+1)  ; bc = this->size
       push iy
	   pop hl
	   xor a
	   adc hl, bc     ; hl = this + this->size
	   push ix
	   pop de
	   xor a
	   sbc hl, de     ; if this + this->size == next, then this->size += next->size, this->next = next->next
	   jr z, memory.free.coalesce.do
	     push ix                ; else, new *this = *next
         pop iy
		 ret
memory.free.coalesce.do:
       ld l, (ix+block.size)
	   ld h, (ix+block.size+1)  ; hl = next->size
	   xor a
	   adc hl, bc               ; hl += this->size
	   ld (iy+block.size), l
	   ld (iy+block.size+1), h  ; this->size = hl
	   ld l, (ix+block.next)
	   ld h, (ix+block.next+1)  ; hl = next->next
       ld (iy+block.next), l
	   ld (iy+block.next+1), h  ; this->next = hl
	   ret

memory.free.passedend:
       ;; append block at the end of the free list
       ld (ix+block.next),l
       ld (ix+block.next+1),h
       push hl
       pop iy
       ld (iy+block.next),0
       ld (iy+block.next+1),0
	   ret

;; get_free
;; OUT BC=freespace
memory.get_free:
       ld ix,memory.heap_start
       ld bc,0
memory.get_free.count:
       ld a,c
       add a,(ix+block.size)
       ld c,a
       ld a,b
       adc a,(ix+block.size+1)
       ld b,a
       ld l,(ix+block.next)
       ld h,(ix+block.next+1)
       ld a,h
       or l
       ret z
       push hl
       pop ix
       jr memory.get_free.count

memory.error:  ld e, 7                           ; out of memory
               __call_basic BASIC_ERROR_HANDLER  ;
               ret
)";

const char * Assembler::compression_routines = R"(
;---------------------------------------------------------------------------------------------------------
; ENCODER/DECODER ROUTINES
;---------------------------------------------------------------------------------------------------------
; RLE adapted from Wonder Boy algorithm
; https://en.wikipedia.org/wiki/Run-length_encoding
;
; Pletter v0.5c1 adapted from XL2S Entertainment 2008
; https://github.com/nanochess/Pletter
;---------------------------------------------------------------------------------------------------------

if defined RLE_DECODER

; RLE - run length encoder
; structure: {b0 b1 b2... b0 b1 b2... b0 b1 b2...}
;   byte 0 = repeat data n times
;   byte 1 = data size
;   byte 2... = data

rle.out.size:     equ BIOS_PARM2+50
rle.out.addr:     equ rle.out.size+2
rle.block.repeat: equ rle.out.addr+2
rle.block.size:   equ rle.block.repeat+1
rle.block.addr:   equ rle.block.size+2
rle.data.size:    equ rle.block.addr+2
rle.data.addr:    equ rle.data.size+2


; hl = rle data
; de = destination
; bc = rle data size
rle.unpack:
  ld a, b
  cp c
  ret z

  xor a
  ld (rle.out.size), a
  ld (rle.out.size+1), a
  ld (rle.out.addr), de
  ld (rle.data.addr), hl
  ld (rle.data.size), bc

rle.unpack.next:
  ld hl, (rle.data.addr)
  ld a, (hl)
  ld (rle.block.repeat), a
  inc hl
  ld a, (hl)
  ld (rle.block.size), a
  xor a
  ld (rle.block.size+1), a
  inc hl
  ld (rle.block.addr), hl

rle.unpack.loop:
  ld hl, (rle.block.addr)
  ld bc, (rle.block.size)
  ldir
  ld bc, (rle.block.size)
  ld hl, (rle.out.size)
  add hl, bc
  ld (rle.out.size), hl
  ld a, (rle.block.repeat)
  dec a
  ld (rle.block.repeat), a
  jr nz, rle.unpack.loop

  ld hl, (rle.block.addr)
  ld bc, (rle.block.size)
  add hl, bc
  ld (rle.data.addr), hl

  ld hl, (rle.data.size)
  inc bc
  inc bc
  xor a
  sbc hl, bc
  ld (rle.data.size), hl
  jr c, rle.unpack.end
  jp nz, rle.unpack.next

rle.unpack.end:
  ld bc, (rle.out.size)
  ld de, (rle.out.addr)
  ld h, d
  ld l, e

  ld a, b
  cp c
  ret

endif

if defined PLETTER_DECODER

; pletter v0.5c msx unpacker
; call unpack with hl pointing to some pletter5 data, and de pointing to the destination.
; changes all registers (de' point to decoded data end address)
; define lengthindata when the original size is written in the pletter data
; define LENGTHINDATA

  macro PLETTER.GETBIT
    add a,a
    call z,pletter.getbit
  endm

  macro PLETTER.GETBITEXX
    add a,a
    call z,pletter.getbitexx
  endm

pletter.modes
  dw pletter.offsok
  dw pletter.mode2
  dw pletter.mode3
  dw pletter.mode4
  dw pletter.mode5
  dw pletter.mode6

pletter.unpack:

  if defined LENGTHINDATA
    inc hl
    inc hl
  endif

  ld a,(hl)
  inc hl
  exx
  ld de,0
  add a,a
  inc a
  rl e
  add a,a
  rl e
  add a,a
  rl e
  rl e
  ld hl,pletter.modes
  add hl,de
  ld e,(hl)
  ld ixl,e
  inc hl
  ld e,(hl)
  ld ixh,e
  ld e,1
  exx
  ld iy,pletter.loop
pletter.literal:
  ldi
pletter.loop:
  PLETTER.GETBIT
  jr nc,pletter.literal
  exx
  ld h,d
  ld l,e
pletter.getlen:
  PLETTER.GETBITEXX
  jr nc,pletter.lenok
pletter.lus:
  PLETTER.GETBITEXX
  adc hl,hl
  ret c
  PLETTER.GETBITEXX
  jr nc,pletter.lenok
  PLETTER.GETBITEXX
  adc hl,hl
  ret c
  PLETTER.GETBITEXX
  jp c,pletter.lus
pletter.lenok
  inc hl
  exx
  ld c,(hl)
  inc hl
  ld b,0
  bit 7,c
  jp z,pletter.offsok
  ;jp ix
  push ix
  ret

pletter.mode6
  PLETTER.GETBIT
  rl b
pletter.mode5
  PLETTER.GETBIT
  rl b
pletter.mode4
  PLETTER.GETBIT
  rl b
pletter.mode3
  PLETTER.GETBIT
  rl b
pletter.mode2
  PLETTER.GETBIT
  rl b
  PLETTER.GETBIT
  jr nc,pletter.offsok
  or a
  inc b
  res 7,c
pletter.offsok
  inc bc
  push hl
  exx
  push hl
  exx
  ld l,e
  ld h,d
  sbc hl,bc
  pop bc
  ldir
  pop hl
  ;jp iy
  push iy
  ret

pletter.getbit
  ld a,(hl)
  inc hl
  rla
  ret

pletter.getbitexx
  exx
  ld a,(hl)
  inc hl
  exx
  rla
  ret

;eof

endif
)";

const char * Assembler::font_data_routines = R"(
;---------------------------------------------------------------------------------------------------------
; MSX FONT DATA
;---------------------------------------------------------------------------------------------------------
; 0 - BIOS FONT
; https://www.msx.org/wiki/MSX_font
;
; 1 - COMPILE ZANAC STYLE
; https://www.msx.org/forum/msx-talk/general-discussion/creating-new-fonts-and-spirits
;
; 2 - KONAMI GRADIUS STYLE 1
; https://fontstruct.com/fontstructions/show/712920/gradius_nes
;
; 3 - KONAMI GRADIUS STYLE 2
; https://fontstruct.com/fontstructions/show/1080449/gradius_gradius_ii
;---------------------------------------------------------------------------------------------------------

if defined FONT_STYLE_1

; pletter data
FONT_STYLE_1_SIZE: equ 453
FONT_STYLE_1_DATA:
db 0x3e,0x00,0x32,0x00,0x20,0x00,0x05,0x96,0x02,0x50,0x00,0x92,0x0f,0x06,0xf8,0x01
db 0xa0,0x0b,0x11,0x78,0xa0,0x70,0x28,0x40,0xf0,0x17,0xc0,0xc8,0x10,0x20,0x40,0x00
db 0x98,0x18,0x00,0x40,0xa0,0x40,0xa8,0x90,0x19,0x98,0x60,0x00,0x0d,0xdd,0x27,0x07
db 0x14,0x00,0x20,0x10,0x17,0x03,0x10,0xe0,0x13,0x20,0xa8,0x70,0x20,0x1c,0x70,0xa8
db 0x20,0x50,0xf8,0xef,0x51,0x38,0x5c,0x2c,0x78,0x00,0xf9,0x00,0x60,0x4c,0x3f,0x00
db 0x08,0x2e,0x00,0x80,0x00,0x3c,0x66,0x66,0x00,0xcc,0xcc,0x80,0x1a,0x04,0x06,0x06
db 0x00,0x0c,0x0c,0x50,0x08,0x0f,0x07,0x38,0xc0,0xc0,0xba,0x0f,0x07,0x0f,0x95,0x07
db 0x24,0x1f,0x07,0xd6,0x17,0x33,0x9d,0x0f,0x07,0xcf,0x2f,0x3c,0xad,0x2f,0x1f,0xb4
db 0x0f,0x27,0xee,0x62,0x7b,0xd3,0x74,0x40,0x77,0x18,0x30,0x60,0xc0,0x18,0x60,0x30
db 0x18,0x0f,0xf8,0xcb,0x01,0x00,0xb0,0x0c,0x12,0x00,0x70,0x65,0x88,0x7f,0x23,0xc4
db 0x07,0x48,0xa8,0xb3,0x44,0x18,0x3e,0x71,0x00,0x7f,0x45,0x71,0x07,0x7e,0x71,0x01
db 0x70,0x02,0x0f,0x70,0x70,0x71,0x75,0x3e,0x0f,0x00,0xc2,0x0f,0x7f,0x70,0x7e,0x10
db 0x75,0x7f,0x07,0x05,0x68,0x02,0x1f,0x77,0x67,0x73,0x1f,0x53,0x35,0x00,0x07,0x1c
db 0x49,0x00,0x07,0x0e,0x86,0x00,0x4e,0x3c,0x05,0x17,0x72,0x7c,0x74,0x72,0x17,0x75
db 0x60,0x00,0x37,0x0e,0x0f,0x7b,0x7f,0x75,0x27,0x13,0x71,0x79,0x7d,0x37,0x71,0x6f
db 0x6f,0x1b,0x5f,0x7e,0x06,0x4f,0x75,0x75,0x72,0x3d,0xc7,0x0f,0x73,0x73,0x0a,0x0f
db 0x3c,0x0e,0x43,0x1f,0x6f,0xf1,0x57,0x71,0xe6,0x2f,0x0e,0x05,0x73,0x76,0x7c,0x07
db 0x1c,0x75,0x7f,0x7b,0x4f,0x3a,0x9e,0x1d,0x3a,0x70,0x07,0x27,0x7f,0x0e,0x1c,0x53
db 0x38,0x6f,0x76,0x40,0xa9,0x00,0x57,0x00,0x80,0x92,0x9a,0x08,0x0f,0x74,0x10,0x00
db 0x0f,0x20,0x3e,0x50,0x88,0x86,0xb9,0x93,0xef,0x1d,0x24,0x0b,0x89,0xf8,0x88,0xe0
db 0x1b,0xf0,0x48,0x70,0x48,0x74,0xf0,0xae,0x80,0x9a,0x00,0x78,0x3a,0x0f,0x48,0x0f
db 0x2b,0x0f,0x80,0xf0,0x80,0xf8,0x1a,0x07,0x80,0x1c,0x1f,0xb8,0x88,0x5f,0x88,0xf2
db 0x37,0xee,0xae,0x00,0x1f,0x1b,0x38,0x08,0x08,0x0e,0x17,0x90,0xe0,0x90,0x17,0xce
db 0x46,0x37,0x8e,0x0f,0xd8,0xa8,0x27,0x86,0x07,0xc8,0xa8,0x98,0x95,0x2f,0x0d,0x39
db 0xcd,0x5f,0x88,0xab,0x4f,0x0f,0xd8,0x74,0xe8,0x07,0x15,0xa0,0x68,0x90,0x5f,0x70
db 0x3c,0x08,0xf8,0x57,0xe3,0xa3,0x3c,0x2e,0x6f,0x1d,0x90,0xa0,0x40,0x07,0x34,0x2f
db 0xd8,0x4f,0x47,0x50,0xd3,0x9d,0x07,0x27,0x67,0x2f,0xd7,0x32,0x37,0x18,0xfb,0xab
db 0x02,0xf7,0xb9,0x13,0x03,0xc0,0x2a,0x03,0x10,0x02,0xf7,0x1a,0x40,0xa8,0x10,0x1f
db 0xed,0xff,0xff,0xff,0xf8

endif

if defined FONT_STYLE_2

; pletter data
FONT_STYLE_2_SIZE: equ 345
FONT_STYLE_2_DATA:
db 0x3e, 0x00, 0x00, 0x00, 0x06, 0x0e, 0x0c, 0x18, 0x00, 0x60, 0x60, 0x2d, 0x00, 0x6c, 0x00, 0x26
db 0x0f, 0x06, 0xfe, 0x40, 0x01, 0x0b, 0x10, 0x7c, 0xd0, 0x7c, 0x16, 0x20, 0x7c, 0x10, 0x10, 0x62
db 0x64, 0x08, 0x10, 0x00, 0x26, 0x46, 0x00, 0x70, 0x88, 0x50, 0x20, 0x54, 0x23, 0x88, 0x76, 0x0f
db 0x18, 0x30, 0x4b, 0x27, 0x06, 0xab, 0x00, 0x3a, 0x02, 0xb0, 0x00, 0x13, 0x24, 0x18, 0x1d, 0x7e
db 0x18, 0x24, 0x18, 0x2d, 0x07, 0x7e, 0x03, 0x96, 0x23, 0x18, 0x3a, 0x5e, 0x06, 0x0f, 0x74, 0x6b
db 0x16, 0x6f, 0x1c, 0x01, 0x38, 0x70, 0xe0, 0xc0, 0x00, 0x7c, 0xc6, 0xc9, 0x00, 0x7c, 0x13, 0x38
db 0xd4, 0x3f, 0x21, 0x0f, 0x0e, 0x0a, 0x3c, 0x78, 0xe0, 0xfe, 0x2c, 0x8e, 0x0c, 0x3c, 0x06, 0x17
db 0x00, 0x1c, 0x2c, 0x4c, 0x8c, 0xfe, 0x0c, 0x0c, 0x00, 0x0e, 0xfe, 0xc0, 0xfc, 0x06, 0x0f, 0x2e
db 0x3c, 0x60, 0x08, 0x2f, 0x1c, 0xfe, 0xc6, 0x0c, 0x7a, 0x00, 0x01, 0x78, 0xc4, 0xe4, 0x7c, 0x9e
db 0x86, 0x7c, 0xc1, 0x47, 0x7e, 0x06, 0x0c, 0x78, 0xdd, 0x5c, 0x5f, 0xe5, 0x07, 0x77, 0x88, 0x25
db 0x60, 0x9b, 0xbb, 0x3f, 0x7d, 0x8c, 0x80, 0x0b, 0xa8, 0x26, 0x13, 0x47, 0x66, 0x66, 0xac, 0x06
db 0x21, 0x37, 0x48, 0xde, 0x00, 0xc0, 0x3f, 0x38, 0x6c, 0x93, 0x40, 0xfe, 0x02, 0x00, 0x3b, 0x5c
db 0x02, 0x10, 0x1f, 0xc0, 0x00, 0x66, 0x3c, 0x00, 0x30, 0xf8, 0xcc, 0x9e, 0x32, 0xcc, 0xf8, 0x7f
db 0x77, 0x9e, 0x02, 0xfe, 0x42, 0x07, 0xb7, 0x3e, 0x87, 0x00, 0xce, 0xc6, 0x66, 0x3e, 0x00, 0xc6
db 0xea, 0x36, 0x37, 0x2f, 0xd0, 0xff, 0x91, 0x2f, 0x0e, 0xa5, 0xa1, 0xa7, 0x00, 0x33, 0xd8, 0xf0
db 0xf8, 0xdc, 0xce, 0x00, 0xc0, 0xea, 0x00, 0x37, 0x0f, 0x0a, 0xee, 0xfe, 0xfe, 0xd6, 0x27, 0x07
db 0x0b, 0xe6, 0xf6, 0xfe, 0xde, 0x39, 0xe3, 0xf7, 0xac, 0x6f, 0x6d, 0x4f, 0xe1, 0x0f, 0xde, 0xcc
db 0x7a, 0xa7, 0x0f, 0x51, 0x37, 0x0f, 0x94, 0xd8, 0x47, 0xc3, 0xf7, 0x00, 0x77, 0x67, 0x2f, 0x06
db 0x07, 0x44, 0x6c, 0x38, 0x10, 0x25, 0x07, 0xd6, 0x50, 0xee, 0x4f, 0x0c, 0x57, 0x7c, 0x38, 0x7c
db 0x07, 0x87, 0xce, 0x24, 0x11, 0x82, 0x9f, 0xa3, 0xd7, 0x34, 0x6f, 0x3c, 0x97, 0x90, 0x3c, 0x7f
db 0xe0, 0x70, 0x38, 0x44, 0x1c, 0x94, 0x0f, 0x0c, 0xe9, 0x00, 0x0f, 0x18, 0x0f, 0xd8, 0x00, 0xbc
db 0x00, 0x90, 0x74, 0x0a, 0xff, 0xff, 0xff, 0xff, 0xc0

endif

if defined FONT_STYLE_3

; pletter data
FONT_STYLE_3_SIZE: equ 326
FONT_STYLE_3_DATA:
db 0x3f, 0x00, 0x01, 0x00, 0x0c, 0x18, 0x30, 0x20, 0x00, 0x40, 0x0d, 0x07, 0x66, 0x66, 0x22, 0x06
db 0x0f, 0x24, 0x7e, 0x24, 0x00, 0x02, 0x00, 0x10, 0x7c, 0xd0, 0x7c, 0x16, 0x7c, 0x40, 0x10, 0x0f
db 0x62, 0x64, 0x08, 0x10, 0x26, 0x00, 0x46, 0x00, 0x70, 0x88, 0x50, 0x20, 0x54, 0x88, 0x56, 0x76
db 0x0f, 0x2e, 0x99, 0x26, 0x35, 0x09, 0x00, 0x30, 0x18, 0x00, 0x02, 0x08, 0x70, 0x00, 0x13, 0x24
db 0x18, 0x7e, 0x29, 0x18, 0x24, 0x18, 0x38, 0x10, 0x5b, 0x35, 0x36, 0x5c, 0x23, 0x1b, 0x09, 0x3c
db 0xfa, 0x6a, 0x0f, 0x98, 0x03, 0x06, 0x6f, 0x60, 0x00, 0xc0, 0x80, 0x00, 0xfe, 0x86, 0x8a, 0x92
db 0xa2, 0x23, 0xc2, 0xfe, 0x2e, 0x30, 0x10, 0xa0, 0x00, 0x0f, 0x82, 0x02, 0xfe, 0x80, 0x38, 0x80
db 0xfe, 0x07, 0x3e, 0x02, 0x52, 0x82, 0x07, 0x0b, 0x88, 0x00, 0x5d, 0xfe, 0x5a, 0x14, 0xdc, 0x0f
db 0x07, 0x82, 0xe1, 0x07, 0x02, 0x02, 0x04, 0x47, 0x55, 0xb1, 0x77, 0x0c, 0x0f, 0x2e, 0x04, 0x02
db 0x3a, 0x5c, 0xef, 0x5f, 0x2e, 0x07, 0x21, 0x66, 0x86, 0x9b, 0x0c, 0x34, 0x10, 0x7e, 0x02, 0xce
db 0x0b, 0x7a, 0xe5, 0x2f, 0x0c, 0x6d, 0x08, 0x01, 0x7c, 0xc4, 0x9c, 0x00, 0xc0, 0x7c, 0x00, 0x00
db 0x02, 0x06, 0x0a, 0x12, 0x3e, 0x42, 0x82, 0x06, 0x00, 0xf8, 0x84, 0x82, 0xfc, 0x76, 0x02, 0x67
db 0x44, 0x00, 0x4f, 0xf0, 0x88, 0x0d, 0x84, 0x39, 0x88, 0xf0, 0x0f, 0xfc, 0xdb, 0x0f, 0x39, 0x07
db 0x80, 0x07, 0x9e, 0xd7, 0x77, 0x03, 0x51, 0x7f, 0x37, 0xbd, 0xe0, 0xbf, 0x08, 0xe5, 0x00, 0xf8
db 0x17, 0x32, 0x34, 0x90, 0x3a, 0xbf, 0x3c, 0x47, 0x0e, 0x82, 0xc6, 0xaa, 0x92, 0x27, 0x02, 0x82
db 0xc2, 0xa2, 0x92, 0x8a, 0x86, 0x07, 0xef, 0x34, 0x07, 0xb7, 0x9e, 0x74, 0x4f, 0x0f, 0x18, 0xfa
db 0xd3, 0x0f, 0xb3, 0x37, 0x5f, 0x8b, 0xcf, 0x07, 0xc7, 0x57, 0x82, 0x9c, 0x2f, 0x57, 0xa0, 0xc4
db 0xb7, 0x89, 0x99, 0x04, 0xaa, 0xaa, 0xcc, 0xcc, 0x88, 0x0f, 0x44, 0x07, 0x28, 0x10, 0x28, 0x44
db 0x82, 0x6d, 0x07, 0x7f, 0xca, 0x95, 0x40, 0x27, 0x68, 0x1c, 0x36, 0x1c, 0x1c, 0x00, 0xc0, 0x60
db 0xf1, 0x0d, 0x02, 0x00, 0x38, 0x04, 0x97, 0x38, 0x93, 0x9a, 0x3c, 0x66, 0x97, 0xf2, 0x8e, 0x90
db 0x9f, 0x09, 0xff, 0xff, 0xff, 0xf8

endif
)";

/*
; FONT 1 - RAW - 760 bytes
; 32 - 0x20
db 0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00   ; space
db 0x20,0x20,0x20,0x20,0x00,0x00,0x20,0x00   ; !
db 0x50,0x50,0x50,0x00,0x00,0x00,0x00,0x00   ; "
db 0x50,0x50,0xf8,0x50,0xf8,0x50,0x50,0x00   ; #
db 0x20,0x78,0xa0,0x70,0x28,0xf0,0x20,0x00   ; $
db 0xc0,0xc8,0x10,0x20,0x40,0x98,0x18,0x00   ; %
db 0x40,0xa0,0x40,0xa8,0x90,0x98,0x60,0x00   ; &
db 0x10,0x20,0x40,0x00,0x00,0x00,0x00,0x00   ; Ž
db 0x10,0x20,0x40,0x40,0x40,0x20,0x10,0x00   ; (
db 0x40,0x20,0x10,0x10,0x10,0x20,0x40,0x00   ; )
db 0x20,0xa8,0x70,0x20,0x70,0xa8,0x20,0x00   ; *
db 0x00,0x20,0x20,0xf8,0x20,0x20,0x00,0x00   ; +
db 0x00,0x00,0x00,0x00,0x00,0x20,0x20,0x40   ; ,
db 0x00,0x00,0x00,0x78,0x00,0x00,0x00,0x00   ; -
db 0x00,0x00,0x00,0x00,0x00,0x60,0x60,0x00   ; .
db 0x00,0x00,0x08,0x10,0x20,0x40,0x80,0x00   ; /
db 0x3c,0x66,0x66,0x00,0xcc,0xcc,0x78,0x00   ; 0
db 0x04,0x06,0x06,0x00,0x0c,0x0c,0x08,0x00   ; 1
db 0x3c,0x06,0x06,0x38,0xc0,0xc0,0x78,0x00   ; 2
db 0x3c,0x06,0x06,0x38,0x0c,0x0c,0x78,0x00   ; 3
db 0x24,0x66,0x66,0x38,0x0c,0x0c,0x08,0x00   ; 4
db 0x3c,0x60,0x60,0x38,0x0c,0x0c,0x78,0x00   ; 5
db 0x3c,0x60,0x60,0x38,0xcc,0xcc,0x78,0x00   ; 6
db 0x3c,0x06,0x06,0x00,0x0c,0x0c,0x08,0x00   ; 7
db 0x3c,0x66,0x66,0x38,0xcc,0xcc,0x78,0x00   ; 8
db 0x3c,0x66,0x66,0x38,0x0c,0x0c,0x78,0x00   ; 9
db 0x00,0x00,0x20,0x00,0x00,0x20,0x00,0x00   ; :
db 0x00,0x00,0x20,0x00,0x00,0x20,0x20,0x40   ; ;
db 0x18,0x30,0x60,0xc0,0x60,0x30,0x18,0x00   ; <
db 0x00,0x00,0xf8,0x00,0xf8,0x00,0x00,0x00   ; =
db 0xc0,0x60,0x30,0x18,0x30,0x60,0xc0,0x00   ; >
db 0x70,0x88,0x08,0x10,0x20,0x00,0x20,0x00   ; ?
db 0x70,0x88,0x08,0x48,0xa8,0xa8,0x70,0x00   ; @
db 0x00,0x3E,0x71,0x71,0x71,0x7F,0x71,0x00   ; A
db 0x00,0x7E,0x71,0x7E,0x71,0x71,0x7E,0x00   ; B
db 0x00,0x3E,0x71,0x70,0x70,0x71,0x3E,0x00   ; C
db 0x00,0x7e,0x71,0x71,0x71,0x71,0x7e,0x00   ; D
db 0x00,0x7f,0x70,0x7e,0x70,0x70,0x7f,0x00   ; E
db 0x00,0x7f,0x70,0x70,0x7f,0x70,0x70,0x00   ; F
db 0x00,0x3e,0x71,0x70,0x77,0x73,0x3e,0x00   ; G
db 0x00,0x71,0x71,0x7f,0x71,0x71,0x71,0x00   ; H
db 0x00,0x1c,0x1c,0x1c,0x1c,0x1c,0x1c,0x00   ; I
db 0x00,0x0e,0x0e,0x0e,0x0e,0x4e,0x3c,0x00   ; J
db 0x00,0x71,0x72,0x7c,0x74,0x72,0x71,0x00   ; K
db 0x00,0x70,0x70,0x70,0x70,0x70,0x7f,0x00   ; L
db 0x00,0x71,0x7b,0x7f,0x75,0x71,0x71,0x00   ; M
db 0x00,0x71,0x79,0x7d,0x77,0x73,0x71,0x00   ; N
db 0x00,0x3e,0x71,0x71,0x71,0x71,0x3e,0x00   ; O
db 0x00,0x7e,0x71,0x71,0x7e,0x70,0x70,0x00   ; P
db 0x00,0x3e,0x71,0x75,0x75,0x72,0x3d,0x00   ; Q
db 0x00,0x7e,0x71,0x71,0x7e,0x73,0x73,0x00   ; R
db 0x00,0x3e,0x71,0x3c,0x0e,0x43,0x3e,0x00   ; S
db 0x00,0x7f,0x1c,0x1c,0x1c,0x1c,0x1c,0x00   ; T
db 0x00,0x71,0x71,0x71,0x71,0x71,0x3e,0x00   ; U
db 0x00,0x71,0x71,0x71,0x73,0x76,0x7c,0x00   ; V
db 0x00,0x71,0x71,0x75,0x7f,0x7b,0x71,0x00   ; W
db 0x00,0x71,0x3a,0x1c,0x1c,0x3a,0x71,0x00   ; X
db 0x00,0x71,0x3a,0x1c,0x1c,0x1c,0x1c,0x00   ; Y
db 0x00,0x7f,0x0e,0x1c,0x38,0x70,0x7f,0x00   ; Z
db 0x70,0x40,0x40,0x40,0x40,0x40,0x70,0x00   ; [
db 0x00,0x00,0x80,0x40,0x20,0x10,0x08,0x00   ; \
db 0x70,0x10,0x10,0x10,0x10,0x10,0x70,0x00   ; ]
db 0x20,0x50,0x88,0x00,0x00,0x00,0x00,0x00   ; ^
db 0x00,0x00,0x00,0x00,0x00,0x00,0xf8,0x00   ; _
db 0x80,0x40,0x20,0x00,0x00,0x00,0x00,0x00   ; `
;db 0x7e,0xc1,0xdd,0xd9,0xdd,0xc1,0x7e,0x00   ; ????
db 0x00,0x00,0x70,0x88,0xF8,0x88,0x88,0x00   ; a
db 0x00,0x00,0xF0,0x48,0x70,0x48,0xF0,0x00   ; b
db 0x00,0x00,0x78,0x80,0x80,0x80,0x78,0x00   ; c
db 0x00,0x00,0xF0,0x48,0x48,0x48,0xF0,0x00   ; d
db 0x00,0x00,0xF8,0x80,0xF0,0x80,0xF8,0x00   ; e
db 0x00,0x00,0xF8,0x80,0xF0,0x80,0x80,0x00   ; f
db 0x00,0x00,0x78,0x80,0xb8,0x88,0x70,0x00   ; g
db 0x00,0x00,0x88,0x88,0xF8,0x88,0x88,0x00   ; h
db 0x00,0x00,0xF8,0x20,0x20,0x20,0xF8,0x00   ; i
db 0x00,0x00,0x38,0x08,0x08,0x88,0x70,0x00   ; j
db 0x00,0x00,0x88,0x90,0xE0,0x90,0x88,0x00   ; k
db 0x00,0x00,0x80,0x80,0x80,0x80,0xF8,0x00   ; l
db 0x00,0x00,0x88,0xD8,0xA8,0x88,0x88,0x00   ; m
db 0x00,0x00,0x88,0xC8,0xA8,0x98,0x88,0x00   ; n
db 0x00,0x00,0xF8,0x88,0x88,0x88,0xF8,0x00   ; o
db 0x00,0x00,0xF0,0x88,0xF0,0x80,0x80,0x00   ; p
db 0x00,0x00,0xF8,0x88,0xA8,0x90,0xE8,0x00   ; q
db 0x00,0x00,0xF8,0x88,0xF8,0xA0,0x90,0x00   ; r
db 0x00,0x00,0x78,0x80,0x70,0x08,0xF8,0x00   ; s
db 0x00,0x00,0xF8,0x20,0x20,0x20,0x20,0x00   ; t
db 0x00,0x00,0x88,0x88,0x88,0x88,0x70,0x00   ; u
db 0x00,0x00,0x88,0x88,0x90,0xA0,0x40,0x00   ; v
db 0x00,0x00,0x88,0x88,0xA8,0xD8,0x88,0x00   ; w
db 0x00,0x00,0x88,0x50,0x20,0x50,0x88,0x00   ; x
db 0x00,0x00,0x88,0x50,0x20,0x20,0x20,0x00   ; y
db 0x00,0x00,0xF8,0x10,0x20,0x40,0xF8,0x00   ; z
db 0x18,0x20,0x20,0x40,0x20,0x20,0x18,0x00   ; {
db 0x20,0x20,0x20,0x00,0x20,0x20,0x20,0x00   ; |
db 0xc0,0x20,0x20,0x10,0x20,0x20,0xc0,0x00   ; }
db 0x40,0xa8,0x10,0x00,0x00,0x00,0x00,0x00   ; ~
;126 - 0x7E

; FONT 2 - RAW - 520 bytes
; char 32 - 96 (65)
	db 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x6, 0xe, 0xc, 0x18, 0x0, 0x60, 0x60, 0x0
	db 0x6c, 0x6c, 0x6c, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x6c, 0x6c, 0xfe, 0x6c, 0xfe, 0x6c, 0x6c, 0x0
	db 0x10, 0x7c, 0xd0, 0x7c, 0x16, 0x7c, 0x10, 0x0
	db 0x0, 0x62, 0x64, 0x8, 0x10, 0x26, 0x46, 0x0
	db 0x70, 0x88, 0x50, 0x20, 0x54, 0x88, 0x76, 0x0
	db 0x0, 0x18, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x18, 0x30, 0x30, 0x30, 0x30, 0x30, 0x18, 0x0
	db 0x30, 0x18, 0x18, 0x18, 0x18, 0x18, 0x30, 0x0
	db 0x0, 0x24, 0x18, 0x7e, 0x18, 0x24, 0x0, 0x0
	db 0x0, 0x18, 0x18, 0x7e, 0x7e, 0x18, 0x18, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x8, 0x10
	db 0x0, 0x0, 0x0, 0x7e, 0x7e, 0x0, 0x0, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0
	db 0x6, 0xe, 0x1c, 0x38, 0x70, 0xe0, 0xc0, 0x0
	db 0x7c, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0x7c, 0x0
	db 0x18, 0x38, 0x18, 0x18, 0x18, 0x18, 0x7e, 0x0
	db 0x7c, 0xc6, 0xe, 0x3c, 0x78, 0xe0, 0xfe, 0x0
	db 0x7e, 0xc, 0x18, 0x3c, 0x6, 0xc6, 0x7c, 0x0
	db 0x1c, 0x2c, 0x4c, 0x8c, 0xfe, 0xc, 0xc, 0x0
	db 0xfe, 0xc0, 0xfc, 0x6, 0x6, 0xc6, 0x7c, 0x0
	db 0x3c, 0x60, 0xc0, 0xfc, 0xc6, 0xc6, 0x7c, 0x0
	db 0xfe, 0xc6, 0xc, 0x18, 0x30, 0x30, 0x30, 0x0
	db 0x78, 0xc4, 0xe4, 0x7c, 0x9e, 0x86, 0x7c, 0x0
	db 0x7c, 0xc6, 0xc6, 0x7e, 0x6, 0xc, 0x78, 0x0
	db 0x0, 0x18, 0x18, 0x0, 0x18, 0x18, 0x0, 0x0
	db 0x0, 0x18, 0x18, 0x0, 0x18, 0x18, 0x8, 0x10
	db 0xc, 0x18, 0x30, 0x60, 0x30, 0x18, 0xc, 0x0
	db 0x0, 0x7e, 0x7e, 0x0, 0x7e, 0x7e, 0x0, 0x0
	db 0x30, 0x18, 0xc, 0x6, 0xc, 0x18, 0x30, 0x0
	db 0x3c, 0x66, 0x66, 0xc, 0x18, 0x0, 0x18, 0x0
	db 0x7c, 0xc6, 0xde, 0xde, 0xde, 0xc0, 0x7c, 0x0
	db 0x38, 0x6c, 0xc6, 0xc6, 0xfe, 0xc6, 0xc6, 0x0
	db 0xfc, 0xc6, 0xc6, 0xfc, 0xc6, 0xc6, 0xfc, 0x0
	db 0x3c, 0x66, 0xc0, 0xc0, 0xc0, 0x66, 0x3c, 0x0
	db 0xf8, 0xcc, 0xc6, 0xc6, 0xc6, 0xcc, 0xf8, 0x0
	db 0xfe, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xfe, 0x0
	db 0xfe, 0xc0, 0xc0, 0xfc, 0xc0, 0xc0, 0xc0, 0x0
	db 0x3e, 0x60, 0xc0, 0xce, 0xc6, 0x66, 0x3e, 0x0
	db 0xc6, 0xc6, 0xc6, 0xfe, 0xc6, 0xc6, 0xc6, 0x0
	db 0x3c, 0x18, 0x18, 0x18, 0x18, 0x18, 0x3c, 0x0
	db 0xe, 0x6, 0x6, 0x6, 0x6, 0xc6, 0x7c, 0x0
	db 0xc6, 0xcc, 0xd8, 0xf0, 0xf8, 0xdc, 0xce, 0x0
	db 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xc0, 0xfe, 0x0
	db 0xc6, 0xee, 0xfe, 0xfe, 0xd6, 0xc6, 0xc6, 0x0
	db 0xc6, 0xe6, 0xf6, 0xfe, 0xde, 0xce, 0xc6, 0x0
	db 0x7c, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0x7c, 0x0
	db 0xfc, 0xc6, 0xc6, 0xc6, 0xfc, 0xc0, 0xc0, 0x0
	db 0x7c, 0xc6, 0xc6, 0xc6, 0xde, 0xcc, 0x7a, 0x0
	db 0xfc, 0xc6, 0xc6, 0xc6, 0xf8, 0xdc, 0xce, 0x0
	db 0x7c, 0xc6, 0xc0, 0x7c, 0x6, 0xc6, 0x7c, 0x0
	db 0x7e, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x0
	db 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0xc6, 0x7c, 0x0
	db 0xc6, 0xc6, 0xc6, 0x44, 0x6c, 0x38, 0x10, 0x0
	db 0xc6, 0xc6, 0xd6, 0xfe, 0xfe, 0xee, 0xc6, 0x0
	db 0xc6, 0xee, 0x7c, 0x38, 0x7c, 0xee, 0xc6, 0x0
	db 0x66, 0x66, 0x24, 0x3c, 0x18, 0x18, 0x18, 0x0
	db 0xfe, 0xe, 0x1c, 0x38, 0x70, 0xe0, 0xfe, 0x0
	db 0x3c, 0x30, 0x30, 0x30, 0x30, 0x30, 0x3c, 0x0
	db 0xc0, 0xe0, 0x70, 0x38, 0x1c, 0xe, 0x6, 0x0
	db 0x3c, 0xc, 0xc, 0xc, 0xc, 0xc, 0x3c, 0x0
	db 0x18, 0x3c, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x0, 0x7e, 0x7e, 0x0
	db 0x0, 0x30, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0

; FONT 3 - RAW - 520 bytes
; char 32 - 96 (65)
	db 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x0, 0xc, 0x18, 0x30, 0x20, 0x0, 0x40, 0x0
	db 0x0, 0x66, 0x66, 0x22, 0x0, 0x0, 0x0, 0x0
	db 0x0, 0x24, 0x7e, 0x24, 0x24, 0x7e, 0x24, 0x0
	db 0x10, 0x7c, 0xd0, 0x7c, 0x16, 0x7c, 0x10, 0x0
	db 0x0, 0x62, 0x64, 0x8, 0x10, 0x26, 0x46, 0x0
	db 0x70, 0x88, 0x50, 0x20, 0x54, 0x88, 0x76, 0x0
	db 0x0, 0x18, 0x30, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x18, 0x30, 0x20, 0x20, 0x20, 0x30, 0x18, 0x0
	db 0x30, 0x18, 0x8, 0x8, 0x8, 0x18, 0x30, 0x0
	db 0x0, 0x24, 0x18, 0x7e, 0x18, 0x24, 0x0, 0x0
	db 0x0, 0x10, 0x10, 0x7c, 0x10, 0x10, 0x0, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x8, 0x10
	db 0x0, 0x0, 0x0, 0x3c, 0x0, 0x0, 0x0, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x18, 0x18, 0x0, 0x0
	db 0x6, 0xc, 0x18, 0x30, 0x60, 0xc0, 0x80, 0x0
	db 0xfe, 0x86, 0x8a, 0x92, 0xa2, 0xc2, 0xfe, 0x0
	db 0x10, 0x30, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0
	db 0xfe, 0x82, 0x2, 0xfe, 0x80, 0x80, 0xfe, 0x0
	db 0xfe, 0x82, 0x2, 0x3e, 0x2, 0x82, 0xfe, 0x0
	db 0x80, 0x80, 0x88, 0x88, 0x88, 0xfe, 0x8, 0x8
	db 0xfe, 0x80, 0x80, 0xfe, 0x2, 0x82, 0xfe, 0x0
	db 0xfe, 0x80, 0x80, 0xfe, 0x82, 0x82, 0xfe, 0x0
	db 0xfe, 0x2, 0x2, 0x4, 0x8, 0x10, 0x20, 0x0
	db 0xfe, 0x82, 0x82, 0xfe, 0x82, 0x82, 0xfe, 0x0
	db 0xfe, 0x82, 0x82, 0xfe, 0x2, 0x2, 0xfe, 0x0
	db 0x0, 0x18, 0x18, 0x0, 0x18, 0x18, 0x0, 0x0
	db 0x0, 0x18, 0x18, 0x0, 0x18, 0x18, 0x8, 0x10
	db 0xc, 0x18, 0x30, 0x60, 0x30, 0x18, 0xc, 0x0
	db 0x0, 0x0, 0x7e, 0x0, 0x0, 0x7e, 0x0, 0x0
	db 0x30, 0x18, 0xc, 0x6, 0xc, 0x18, 0x30, 0x0
	db 0xfe, 0x82, 0x82, 0xc, 0x10, 0x0, 0x10, 0x0
	db 0x7c, 0xc4, 0x9c, 0x9c, 0x9c, 0xc0, 0x7c, 0x0
	db 0x2, 0x6, 0xa, 0x12, 0x3e, 0x42, 0x82, 0x0
	db 0xf8, 0x84, 0x82, 0xfc, 0x84, 0x82, 0xfc, 0x0
	db 0xfe, 0x80, 0x80, 0x80, 0x80, 0x80, 0xfe, 0x0
	db 0xf0, 0x88, 0x84, 0x82, 0x84, 0x88, 0xf0, 0x0
	db 0xfe, 0x80, 0x80, 0xfc, 0x80, 0x80, 0xfe, 0x0
	db 0xfe, 0x80, 0x80, 0xfc, 0x80, 0x80, 0x80, 0x0
	db 0xfe, 0x80, 0x80, 0x9e, 0x82, 0x82, 0xfe, 0x0
	db 0x82, 0x82, 0x82, 0xfe, 0x82, 0x82, 0x82, 0x0
	db 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0
	db 0xfe, 0x8, 0x8, 0x8, 0x8, 0x8, 0xf8, 0x0
	db 0x82, 0x84, 0x88, 0x90, 0x88, 0x84, 0x82, 0x0
	db 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0xfe, 0x0
	db 0x82, 0xc6, 0xaa, 0x92, 0x82, 0x82, 0x82, 0x0
	db 0x82, 0xc2, 0xa2, 0x92, 0x8a, 0x86, 0x82, 0x0
	db 0xfe, 0x82, 0x82, 0x82, 0x82, 0x82, 0xfe, 0x0
	db 0xfe, 0x82, 0x82, 0x9e, 0x80, 0x80, 0x80, 0x0
	db 0xfe, 0x82, 0x82, 0x82, 0x92, 0x8a, 0xfa, 0x0
	db 0xfe, 0x82, 0x82, 0x9e, 0x88, 0x84, 0x82, 0x0
	db 0xfe, 0x80, 0x80, 0xfe, 0x2, 0x2, 0xfe, 0x0
	db 0xfe, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x0
	db 0x82, 0x82, 0x82, 0x82, 0x82, 0x82, 0xfe, 0x0
	db 0x82, 0x84, 0x88, 0x90, 0xa0, 0xc0, 0x80, 0x0
	db 0x89, 0x99, 0xaa, 0xaa, 0xcc, 0xcc, 0x88, 0x0
	db 0x82, 0x44, 0x28, 0x10, 0x28, 0x44, 0x82, 0x0
	db 0x82, 0x44, 0x28, 0x10, 0x10, 0x10, 0x10, 0x0
	db 0xfe, 0x4, 0x8, 0x10, 0x20, 0x40, 0xfe, 0x0
	db 0x1c, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1c, 0x0
	db 0xc0, 0x60, 0x30, 0x18, 0xc, 0x6, 0x2, 0x0
	db 0x38, 0x8, 0x8, 0x8, 0x8, 0x8, 0x38, 0x0
	db 0x18, 0x3c, 0x66, 0x0, 0x0, 0x0, 0x0, 0x0
	db 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x7e, 0x0
	db 0x0, 0x30, 0x18, 0x0, 0x0, 0x0, 0x0, 0x0

*/

#endif // KERNEL_H_INCLUDED
