#ifndef LEX_H_INCLUDED
#define LEX_H_INCLUDED

#include <string>
#include <vector>
#include <stack>

using namespace std;

class Lexeme {
  public:
    enum LexemeType {
        type_unknown = 0,
        type_identifier, // x, y, up
        type_keyword,    // if, print, color
        type_separator,  // { } ( ) , : ;
        type_operator,   // + - * / < > =
        type_literal,    // true, 6, 1.0, "text"
        type_comment     // text
    } type;

    enum LexemeSubType {
        subtype_any = 0,
        subtype_string,
        subtype_numeric,
        subtype_single_decimal,
        subtype_double_decimal,
        subtype_basic_string,
		subtype_boolean_operator,
		subtype_function,
		subtype_internal,
    } subtype;

    string name, value;
    int indent;
    int tag;

    void print();
    char * getTypeName();
    char * getSubTypeName();

    bool isKeyword();
	bool isBooleanOperator();
	bool isFunction();

    Lexeme();

};

class LexerLine {
  private:
    int lexemeIndex;
    vector<Lexeme*> lexemes;
    stack<int> lexemeStack;

  public:
    string line;

    bool evaluate();
    void print();

    void clearLexemes();
    void addLexeme(Lexeme *lexeme);

    Lexeme *getFirstLexeme();
    Lexeme *getCurrentLexeme();
    Lexeme *getNextLexeme();
    Lexeme *getPreviousLexeme();
    Lexeme *getLastLexeme();
    Lexeme *getLexeme(int i);
    void setLexemeBOF();
    int getLexemeCount();
    void pushLexeme();
    void popLexeme();
    void popLexemeDiscarting();

	bool isNumeric(char c);
	bool isDecimal(char c);
	bool isHexDecimal(char c);
	bool isSeparator(char c);
	bool isOperator(char c);
	bool isIdentifier(char c, bool start);
	bool isComment(char c);

    LexerLine();
};

class Lexer {
  public:
    int lineNo=0;
    vector<LexerLine*> lines;

    bool load(char *filename);
    bool evaluate();

    void print();
    void error();

};

#endif // LEX_H_INCLUDED
