#ifndef MAIN_H_INCLUDED
#define MAIN_H_INCLUDED

#define app_version "0.4.7.1"

char inputFilename[255]="",
     outputFilename[255]="",
     compileFilename[255]="";

bool parmHelp=false,  parmDebug=false,
     parmQuiet=false, parmError=false,
     parmBin=false, parmRom=true, parmDOS=false,
     parmCompile=false, parmNoCtrlBreak=false,
	 parmBCD=false, parmFGX=false, parmDoc=false, parmVer=false, parmLineNumber=false;

bool FileExists(char *filename);

#ifdef __WINDOWS__
    char pasmoFilename[] = ".\\pasmo.exe";
    char pasmoError[] = "compiler not found in current folder.\nTry download it from http://pasmo.speccy.org/";
#else
    char pasmoFilename[] = "/usr/bin/pasmo";
    char pasmoError[] = "compiler not found.\nTry download it from http://pasmo.speccy.org/\nOr, try to install via command below:\nsudo apt-get install parmo";
#endif // defined

const char * info_splash = R"(
MSXBAS2ASM - MSX BASIC TO Z80 ASSEMBLY CONVERTER
Created by Amaury Carvalho (2019)
Version:)";

const char * info_usage = R"(
Usage: MSXBAS2ASM [options] <filename.bas>
Options:
       -q  = quiet (no verbose)
       -d  = debug mode (show details)
       -tb = target to binary
       -tr = target to rom (default)
       -td = target to MSX-DOS
       -c  = compile with Pasmo
       -h or -? = help
       --doc = display documentation
       --ver = display version history
       --ncb = disable control+break
       --bcd = enable BCD math pack (slow, but more precise)
       --flt = enable float point math pack (fast, but less precise)
	   --fgx = fast graphical routines (MSX1 screen modes only)
	   --lin = update BASIC line number (for error handler)
Output: <filename.asm>

See more information at http://launchpad.net/msxbas2asm

)";

const char * info_documentation = R"(
control-commands
    IF..THEN..ELSE: DONE
    FOR..NEXT: DONE

data
    DATA: DONE
    READ: DONE
    RESTORE: DONE

expansion
    ASM <assembly code>: TODO
    SEED <n>: DONE
    RANDOMIZE: DONE
    IRND(): DONE
    SLEEP <n>: DONE
    PAUSE: DONE
    DATA SET <name>: DONE
    DATA DECODER <RAW|RLE|PLETTER>: DONE
    DATA TILE BANK <0-2|3=0+1, 4=1+2, 5=0+2, 6=0+1+2=all|ALL>: DONE
    DATA TILE FONT <0=bios|1=zanac|2=gradius 1|3=gradius 2>: DONE
    DATA TILE <SET|COLOR> <n>,...: DONE
    DATA TILE INCLUDE <SET|COLOR> <"filename.bmp">: INPROGRESS (import file to bin/rom)
    DATA TILE LOAD <SET|COLOR> <"filename.bmp">: INPROGRESS (import file from disk)
    DATA TILE WALL <tile number>, ...: DONE
    DATA TILE HOTSPOT <tile number>, ...: DONE
    DATA TILE LIMIT <ON|OFF>: DONE
    DATA SPRITE <SET|COLOR|MULTI> <n>,...: DONE
    DATA SPRITE INCLUDE <SET|COLOR|MULTI> <"filename.bmp">: INPROGRESS (import file to bin/rom)
    DATA SPRITE LOAD <SET|COLOR|MULTI> <"filename.bmp">: INPROGRESS (import file from disk)
    DATA SCREEN SET <n>,...: DONE
    DATA SCREEN INCLUDE SET <"filename.bmp">: INPROGRESS (import file to bin/rom)
    DATA SCREEN LOAD SET <"filename.bmp">: INPROGRESS (import file from disk)
    DATA MUSIC <string>, ...: DONE
    MUSIC <PLAY|STOP>: DONE
    PUT TILE <tile number>, (<x:0-31>, <y:0-23>)[, <color>|, (<color>,...)]: INPROGRESS
    PUT SPRITE <sprite number>, (<x>, <y>)[, (<color>,...)]: TODO
    <var:tile number> = TILE(x,y): TODO
    <var:true/false> = COLLISION(): DONE
    <var:sprite number> = COLLIDER(): DONE
    <var:true/false> = HOTSPOT(): INPROGRESS
    <var:tile number> = TOUCHER(): INPROGRESS
    <var> = SPRITE_X(<sprite number>): TODO
    <var> = SPRITE_Y(<sprite number>): TODO
    <var> = SPRITE_COLOR(<sprite number>): TODO
    <var> = SPRITE_PATTERN(<sprite number>): TODO
    <var> = INKEY: DONE
    <n> SHL <n>: DONE
    <n> SHR <n>: DONE

expression-evaluator
    Literals: DONE
    Variables: DONE
    Conditionals: DONE
    Math operators: DONE
    String concatenate: DONE
    Function assignment: INPROGRESS

file-commands
    OPEN: TODO
    CLOSE: TODO
    EOF: TODO
    ERR: TODO
    ERROR: TODO
    FILES: TODO
    MAXFILES: TODO

graphics
    SCREEN: DONE (first 2 parameters only)
    WIDTH: DONE
    LOCATE: DONE
    DRAW: DONE
    POS: TODO
    CSRLIN: TODO
    COLOR: DONE
	COLOR SPRITE: DONE
    SPRITE: DONE
    SPRITE$: DONE
    BASE: DONE
    CIRCLE: DONE (arc and aspect ratio not implemented)
    POINT: TODO
    LINE: DONE
    PAD: DONE
    PAINT: DONE
    PDL: DONE
    PRESET: TODO
    PSET: DONE
    PUT SPRITE: DONE
    STICK: DONE
    STRIG: DONE
    Graphic mode CLS: DONE
    Graphic mode PRINT: DONE
    Graphic mode INPUT: INPROGRESS

input-commands
    INPUT: DONE
    INPUT$: DONE
    LINE INPUT: DONE
    INKEY$: DONE

low-level-io
    INP: TODO
    OUT: TODO
    PEEK: DONE
    POKE: DONE
    SWAP: TODO
    VARPTR: DONE
    DEF USR: TODO
    USR: TODO
    VDP: DONE
    VPEEK: DONE
    VPOKE: DONE
    WAIT: TODO

math-functions
    ABS: TODO
    ASC: DONE
    ATN: DONE
    CDBL: TODO
    CINT: TODO
    COS: DONE
    CSNG: TODO
    EXP: DONE
    FIX: TODO
    INT: DONE
    LEN: TODO
    LOG: DONE
    MOD: DONE
    RND: DONE
    SGN: TODO
    SIN: DONE
    SQR: DONE
    TAN: DONE
    VAL: TODO

miscellaneous
    CLEAR: INPROGRESS
    DEF DBL: DONE
    DEF FN: DONE (ignored)
    DEF INT: DONE
    DEF SNG: DONE
    DEF STR: DONE
    FRE: TODO
    ERL: TODO
    GET DATE: TODO
    GET TIME: TODO
    ON...GOSUB: TODO
    ON...GOTO: TODO
    ON ERROR GOTO: TODO
    RESUME: TODO
    ON INTERVAL GOSUB: DONE
    ON KEY GOSUB: TODO
    ON SPRITE GOSUB: DONE
    ON STOP GOSUB: TODO
    ON STRIG GOSUB: TODO
    INTERVAL: DONE
    KEY: TODO
    STOP: TODO
    LPOS: TODO
    LPRINT: TODO

sound-commands
    BEEP: DONE
    PLAY A[,B[,C]]: DONE (only 3 parameters)
    PLAY(): DONE
    SOUND: DONE

string-functions
    ATTR$: TODO
    BIN$: TODO
    CHR$: DONE
    HEX$: TODO
    INSTR: TODO
    LEFT$: TODO
    MID$: TODO
    OCT$: TODO
    RIGHT$: TODO
    SPACE$: DONE
    SPC: DONE
    STR$: TODO
    STRING$: TODO
    TAB: TODO

variables
    LET: DONE
    String: DONE
    Integer: DONE
    Single: DONE
    Double: DONE
    Array: INPROGRESS
    DIM: INPROGRESS
    REDIM: INPROGRESS
    ERASE: INPROGRESS


)";


const char * info_history = R"(
Version history

0.5.4 - (2021/xx/xx) ON GOTO, ON GOSUB, IF GOTO
0.5.3 - (2021/xx/xx) ON STRIG, STRIG() ON/OFF, KEY() ON/OFF
0.5.2 - (2021/xx/xx) DATA LOAD
0.5.1 - (2021/xx/xx) PUT TILE multicolor
0.5.0 - (2021/xx/xx) TILE(), SPRITE_X(), SPRITE_Y(), SPRITE_COLOR(), SPRITE_PATTERN()
0.4.9 - (2021/xx/xx) HOTSPOT(), TOUCHER()
0.4.8 - (2021/xx/xx) DIM, REDIM, ERASE, array variables
0.4.7 - (2020/11/03) BCD mode bugs fix
0.4.6 - (2020/04/10) compile to 32kb ROM format
0.4.5 - (2020/04/05) DATA MUSIC, MUSIC PLAY/STOP, DATA TILE HOTSPOT
0.4.4 - (2020/04/02) COLLISION(), COLLIDER(), DATA TILE WALL, DATA TILE LIMIT, PUT SPRITE STEP
0.4.3 - (2020/03/16) DATA INCLUDE (RAW/PLETTER), DATA SPRITE COLOR/MULTI
0.4.2 - (2020/03/14) DATA TILE BANK, DATA TILE FONT
0.4.1 - (2020/03/11) DATA enhancements basics, PUT TILE
0.4.0 - (2020/02/22) ON INTERVAL, ON SPRITE, INTERVAL ON/OFF, SPRITE ON/OFF, PLAY()
0.3.9 - (2020/02/20) several memory management bug fixes, sprites routines fixed
0.3.8 - (2020/02/01) DRAW and PLAY remodelled (direct call to ROM BASIC internal routines)
0.3.7 - (2020/01/20) VARPTR, DEF INT/DBL/SNG/STR
0.3.6 - (2020/01/18) POKE, PEEK, BASE, VPOKE, VPEEK, VDP, update line number (for error handler) command line option
0.3.5 - (2020/01/02) LINE INPUT, string concatenate fixed;
0.3.4 - (2019/12/30) SPRITE$, COLOR SPRITE, PUT SPRITE;
0.3.3 - (2019/12/23) version, documentation and fast graphical routines (MSX1 only) command line options implemented;
0.3.2 - (2019/12/22) floating point math bug fixed (multiply a value stored at rom halts real hardware);
0.3.1 - (2019/12/10) SCREEN command sprite parameter implemented and sprite support routines enhanced;
0.3.0 - (2019/12/04) data / read commands implemented;
0.2.9 - (2019/12/01) data´s command memory management implemented;
0.2.8 - (2019/11/28) graphical commands enhanced for MSX2 screen mode;
0.2.7 - (2019/11/24) graphical support routines enhanced;
0.2.6 - (2019/11/01) functions INT, SPC, SPACE$, command WIDTH, COLOR with missing parameters, and corrections in IF/THEN/ELSE command;
0.2.5 - (2019/08/26) draft for sprite commands;
0.2.4 - (2019/08/25) screen command adjusts (MSX1/MSX2 compatibility), ASC, CHR$, INKEY$, INPUT$, STICK, STRIG, PAD, PDL;
0.2.3 - (2019/08/15) graphics commands PSET, LINE, PAINT and CIRCLE implemented;
0.2.2 - (2019/08/10) graphics library included as a ROM Basic interpreter functions substitute, 16k rom size automatic detection (by code line count);
0.2.1 - (2019/07/28) float point math pack library included as a default ROM Basic BCD math alternative (can disable it via command line option);
0.2.0 - (2019/07/21) parser refactoring (individualized command parse), FOR...NEXT, LOG, SIN, COS, TAN, ATN, SQR, EXP, print in graphical mode, LOCATE in graphical mode, PSET, function parameters;
0.1.9 - (2019/07/14) parser refactoring and information better structured in analysis display (-d option), INKEY function (with integer result), IF command;
0.1.8 - (2019/07/04) parameters refactoring (10% speed optimization), conditionals operators;
0.1.7 – (2019/06/26) expression evaluator, time, randomize, rnd(), irnd();
0.1.6 – (2019/06/25) expression evaluator, math pack, variables;
0.1.5 – (2019/06/16) expression evaluator, float point math, literals;
0.1.4 – (2019/06/16) expression evaluator, parser, integer math;
0.1.3 – (2019/06/10) variables, input;
0.1.2 – (2019/06/07) literals, jump commands, print;
0.1.1 – (2019/06/02) basic kernel (lexer, parser, assembler);
0.1.0 – (2019/05/30) proof of concept.
)";


#endif // MAIN_H_INCLUDED
