#ifndef PARSE_H_INCLUDED
#define PARSE_H_INCLUDED

#include <string>
#include <vector>
#include <stack>
#include "lex.h"

using namespace std;

class ActionNode {
  public:
    string name;
    bool isBasicAction;
    bool isDirectParm;
    bool isMultiParm;
    bool isRemark;
    Lexeme* lastLexeme;
    int parmNo, indent;
    vector<Lexeme*> lexemes;
    vector<ActionNode*> actions;
    void print();
    ActionNode();
};

class TagNode {
  public:
    string name;
    vector<ActionNode*> actions;
    void print();
};

class Parser {
  private:
    bool eval_line(LexerLine *lexerLine);
    bool eval_phrase(LexerLine *phrase);
    bool eval_statement(LexerLine *statement);
    bool eval_assignment(LexerLine *assignment);
    bool eval_expression(LexerLine *parm);
    bool eval_array_identifier(LexerLine *statement);

    bool eval_cmd_generic(LexerLine *statement, int parmCount);
    bool eval_cmd_let(LexerLine *statement);
    bool eval_cmd_rem(LexerLine *statement);
    bool eval_cmd_def(LexerLine *statement, int vartype);
    bool eval_cmd_data(LexerLine *statement);
    bool eval_cmd_data_set(LexerLine *statement);
    bool eval_cmd_data_tile(LexerLine *statement);
    bool eval_cmd_data_tile_limit(LexerLine *statement);
    bool eval_cmd_data_include(LexerLine *statement, string name);
    bool eval_cmd_data_load(LexerLine *statement);
    bool eval_cmd_data_sprite(LexerLine *statement);
    bool eval_cmd_data_screen(LexerLine *statement);
    bool eval_cmd_data_decoder(LexerLine *statement);
    bool eval_cmd_data_music(LexerLine *statement);
    bool eval_cmd_dim(LexerLine *statement);
    bool eval_cmd_redim(LexerLine *statement);
    bool eval_cmd_erase(LexerLine *statement);
    bool eval_cmd_load(LexerLine *statement);
    bool eval_cmd_read(LexerLine *statement);
    bool eval_cmd_restore(LexerLine *statement);
    bool eval_cmd_cls(LexerLine *statement);
    bool eval_cmd_end(LexerLine *statement);
    bool eval_cmd_beep(LexerLine *statement);
    bool eval_cmd_goto(LexerLine *statement);
    bool eval_cmd_gosub(LexerLine *statement);
    bool eval_cmd_return(LexerLine *statement);
    bool eval_cmd_print(LexerLine *statement);
    bool eval_cmd_input(LexerLine *statement);
    bool eval_cmd_line_input(LexerLine *statement);
    bool eval_cmd_locate(LexerLine *statement);
    bool eval_cmd_draw(LexerLine *statement);
    bool eval_cmd_play(LexerLine *statement);
    bool eval_cmd_music(LexerLine *statement);
    bool eval_cmd_put(LexerLine *statement);
    bool eval_cmd_put_sprite(LexerLine *statement);
    bool eval_cmd_put_tile(LexerLine *statement);
    bool eval_cmd_put_kanji(LexerLine *statement);
    bool eval_cmd_poke(LexerLine *statement);
    bool eval_cmd_vpoke(LexerLine *statement);
    bool eval_cmd_base(LexerLine *statement);
    bool eval_cmd_vdp(LexerLine *statement);
    bool eval_cmd_time(LexerLine *statement);
    bool eval_cmd_sprite(LexerLine *statement);
    bool eval_cmd_sound(LexerLine *statement);
    bool eval_cmd_screen(LexerLine *statement);
    bool eval_cmd_color(LexerLine *statement);
    bool eval_cmd_color_rgb(LexerLine *statement);
    bool eval_cmd_color_sprite(LexerLine *statement);
    bool eval_cmd_if(LexerLine *statement);
    bool eval_cmd_then(LexerLine *statement);
    bool eval_cmd_else(LexerLine *statement);
    bool eval_cmd_for(LexerLine *statement);
    bool eval_cmd_next(LexerLine *statement);
    bool eval_cmd_sleep(LexerLine *statement);
    bool eval_cmd_randomize(LexerLine *statement);
    bool eval_cmd_seed(LexerLine *statement);
    bool eval_cmd_pset(LexerLine *statement);
    bool eval_cmd_line(LexerLine *statement);
    bool eval_cmd_circle(LexerLine *statement);
    bool eval_cmd_paint(LexerLine *statement);
    bool eval_cmd_on(LexerLine *statement);
    bool eval_cmd_on_goto_gosub(LexerLine *statement);
    bool eval_cmd_on_error(LexerLine *statement);
    bool eval_cmd_on_interval(LexerLine *statement);
    bool eval_cmd_on_key(LexerLine *statement);
    bool eval_cmd_on_sprite(LexerLine *statement);
    bool eval_cmd_on_stop(LexerLine *statement);
    bool eval_cmd_on_strig(LexerLine *statement);
    bool eval_cmd_interval(LexerLine *statement);
    bool eval_cmd_stop(LexerLine *statement);
    bool eval_cmd_key(LexerLine *statement);
    bool eval_cmd_strig(LexerLine *statement);

    int readFromFile(char *filename, char *data, int maxlen);

    int getOperatorPrecedence(Lexeme *lexeme);
    ActionNode * pushActionFromLexeme(Lexeme *lexeme);

    void pushActionRoot(ActionNode *action);
    void popActionRoot();

    TagNode *tag;
    ActionNode *actionRoot;
    LexerLine *error_line;

    stack<ActionNode*> actionStack;

    bool eval_expr_error, line_comment;
    int parmNo=0;
    int if_count=0;
    string error_message;
	//string data_set_name = "default";

  public:
    int lineNo=0;
    vector<TagNode*> tags;      // abstract syntax tree in list form implementation

    int deftbl[26];            // DEFTBL
    bool has_traps;

    bool evaluate(Lexer *lexer);

    void print();
    void error();

};

#endif // PARSE_H_INCLUDED
