#ifndef SYMBOLS_H_INCLUDED
#define SYMBOLS_H_INCLUDED

const char *Assembler::symbol_keywords[][2] = {
    {
        "_NOP_",
        R"(
		nop                     ; unknown keyword
        )"
    },

    {
        "CLS",
        R"(
        if defined EXIST_DATA_SET
           call gfxIsTileMode
           jp z, gfxClearTileScreen
        endif
        xor a                   ; reset Z flag
        __call_bios BIOS_CLS    ; clear screen
        ret                     ;
        )"
    },

    {
        "LET",
        R"(
                    ; out IX = variable assigned address
                    pop.parm                ; get variable address parameter
                    push hl                 ; just to transfer hl to ix
                    pop ix                  ;
                    ld a, (ix)              ; get variable type
                    cp 3                    ; test if string
                    jr nz, LET.PARM         ; if not a string, it isn't necessary to free memory
                    ld a, (ix + 3)          ; get variable string length
                    or a                    ; cp 0
                    jr z, LET.PARM          ; if zero, it isn't necessary to free memory
                    ld c, (ix + 4)          ; get old string address low
                    ld b, (ix + 5)          ; get old string address high
                    push ix                 ; save variable address
                      push bc               ; just to transfer bc (old string address) to ix
                      pop ix                ;
                      call memory.free      ; free memory
                    pop ix                  ; restore variable address
        LET.PARM:   pop.parm                ; get data address parameter (out hl = data address)
                    ld a, (ix + 2)          ; get variable type flag
                    or a                    ; cp 0 - test type flag (0=any, 255=fixed)
                    jr nz, LET.FIXED        ; if type flag is fixed, so casting is necessary
        LET.ANY:    push ix                 ; just to transfer ix (variable address) to de
                    pop de                  ;
                    ldi                     ; copy 1 byte from hl (data address) to de (variable address)
                    inc de                  ; go to variable data area
                    inc de                  ;
                    inc hl                  ; go to data data area
                    inc hl                  ;
                    ld bc, 8                ; data = 8 bytes
                    ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                    ld a, (ix)              ; get variable type
                    cp 3                    ; test if string
                    ret nz                  ; if not string, return
                    jp LET.STRING           ; else do string treatment (in ix = variable address)
        LET.FIXED:  push ix                 ; save variable destination address
		            push hl                 ; save variable source address
                      ld a, (ix)            ; get variable fixed type, and hl has parameter data address
                      call CAST_TO          ; cast data to type (in hl = variable address, a = type to, out hl = casted data address)
					pop de
                    pop ix                  ; restore variable address
                    ld a, (ix)              ; get variable destination type again
                    cp 3                    ; test if string
                    jr nz, LET.VALUE        ; if not string, do value treatment
					ld a, (de)              ; get variable source type again
                    cp 3                    ; test if string
                    jr nz, LET.FIX1         ; if not string, get casted string size
					inc de
					inc de
					inc de
					ld a, (de)
					ld (ix + 3), a          ; source string size
					jr LET.FIX2
		LET.FIX1:   call GET_STR.LENGTH     ; get string length (in HL, out a)
                    ld (ix + 3), a          ; set variable length
		LET.FIX2:   ld (ix + 4), l          ; casted data address low
                    ld (ix + 5), h          ; casted data address high
                    jp LET.STRING           ; do string treatment (in ix = variable address)
        LET.VALUE:  push ix                 ; just to transfer ix (variable address) to de
                    pop de                  ;
                    inc de                  ; go to variable data area (and the data from its casted)
                    inc de                  ;
                    inc de                  ;
                    ld bc, 8                ; data = 8 bytes
                    ldir                    ; copy bc bytes from hl (data address) to de (variable address)
                    ret                     ;
        LET.STRING: ld a, (ix + 3)          ; string size
                    or a                    ; cp 0 - test if null
                    jr nz, LET.ALLOC        ; if not null, allocate new string (in ix = variable address)
                    ld bc, LIT_NULL_STR     ; else, set to a null string literal
                    ld (ix + 4), c          ; variable address low
                    ld (ix + 5), b          ; variable address high
                    ret                     ;
        LET.ALLOC:  push ix                 ; save variable address
                      ld l, (ix + 4)        ; source string address low
                      ld h, (ix + 5)        ; source string address high
                      push hl               ; save copy from address
                        ld c, (ix + 3)      ; get variable length
                        ld b, 0             ;
                        inc bc              ; string length have one more byte from zero terminator
                        push bc             ; save variable lenght + 1
                          call memory.alloc ; in bc = size, out ix = address, nz=OK
		                  jp z, memory.error
                          push ix           ; just to transfer memory address from ix to de
                          pop de            ;
                        pop bc              ; restore bytes to be copied
                      pop hl                ; restore copy from string address
                      push de               ; save copy to address
                        ldir                ; copy bc bytes from hl (data address) to de (variable address)
						;xor a
						;ld (de), a
                      pop de                ; restore copy to address
                    pop ix                  ; restore variable address
                    ld (ix + 4), e          ; put memory address low into variable
                    ld (ix + 5), d          ; put memory address high into variable
                    ret                     ; variable assigned
        )"
    },

    {
        "ERASE",
        R"(
            pop.parm                    ; get parameter (array variable)
            jp ARRAY.ERASE
        )"
    },

    {
        "DIM1",
        R"(
            pop.parm                    ; get parameter (array variable)
            ld (BIOS_TEMP), hl
            pop.parm                    ; get parameter (dimension x size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            ld hl, (BIOS_TEMP)          ; array variable
            ld de, 0                    ; 2nd dimension factor
            inc bc                      ; bc = elements count
            jp ARRAY.DIM
        )"
    },

    {
        "DIM2",
        R"(
            pop.parm                    ; get parameter (array variable)
            ld (BIOS_TEMP), hl
            pop.parm                    ; get parameter (dimension x size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            inc bc
            ld (BIOS_TEMP2), bc         ; temp2 = x+1
            pop.parm                    ; get parameter (dimension y size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            inc bc
            ld (BIOS_TEMP3), bc         ; temp3 = y+1

            ; bc = (x+1) * (y+1) = elements count
            ld hl, (BIOS_TEMP2)
            ;ld bc, (BIOS_TEMP3)
            call MATH.MULT.16     ; hl = hl * bc

            ld de, (BIOS_TEMP2)         ; 2nd dimension factor = x+1
            ld c, l
            ld b, h                     ; elements count
            ld hl, (BIOS_TEMP)          ; array variable
            jp ARRAY.DIM
        )"
    },

    {
        "REDIM1",
        R"(
            pop.parm                    ; get parameter (array variable)
            ld (BIOS_TEMP), hl
            pop.parm                    ; get parameter (dimension size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            ld hl, (BIOS_TEMP)          ; array variable
            ld de, 0                    ; 2nd dimension factor
            inc bc                      ; bc = elements count
            jp ARRAY.REDIM
        )"
    },

    {
        "REDIM2",
        R"(
            pop.parm                    ; get parameter (array variable)
            ld (BIOS_TEMP), hl
            pop.parm                    ; get parameter (dimension x size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            inc bc
            ld (BIOS_TEMP2), bc         ; temp2 = x+1
            pop.parm                    ; get parameter (dimension y size)
            call CAST_TO.INT            ;
            call GET_INT.VALUE          ; output BC with integer value
            inc bc
            ld (BIOS_TEMP3), bc         ; temp3 = y+1

            ; bc = (x+1) * (y+1) = elements count
            ld hl, (BIOS_TEMP2)
            ;ld bc, (BIOS_TEMP3)
            call MATH.MULT.16     ; hl = hl * bc

            ld de, (BIOS_TEMP2)         ; 2nd dimension factor = x+1
            ld c, l
            ld b, h                     ; elements count
            ld hl, (BIOS_TEMP)          ; array variable
            jp ARRAY.REDIM
        )"
    },

    {
        "READ",
        R"(
                    pop.parm                    ; get first parameter (variable to be read)
                    push hl                     ; save input variable...
					  xor a                       ; clear carry
					  ld hl, (BASIC_DATLIN)       ; index of data items
					  ld de, DATA_ITEMS_COUNT     ; count of data items
					  sbc hl, de                  ; compare hl >= de
					  jr z, READ.END              ; jump if equal
					  jr nc, READ.END             ; jump if above
                        pop de                      ; restore saved input variable to DE
					    ld ix, (BASIC_DATPTR)       ; data items pointer
					    ld l, (ix)
					    ld h, (ix+1)
                        push.parm                   ; LET parameter 2 - data as right operand
					    ex de, hl                   ; put DE into HL
						push.parm                   ; LET parameter 1 - input variable as left operand
                        call LET                    ; put data into variable
						ld hl, (BASIC_DATLIN)       ; old index of data items
						inc hl                      ; next item
						ld (BASIC_DATLIN), hl       ; save new index of data items
						ld hl, (BASIC_DATPTR)       ; old data items pointer
						push hl
						inc hl                      ; next item
						inc hl                      ; next item
						ld (BASIC_DATPTR), hl       ; save new data items pointer
		READ.END:   pop hl
					ret
        )"
    },
    {
        "RESTORE",
        R"(
					ld hl, DATA_ITEMS
					ld (BASIC_DATPTR), hl       ; data items pointer
					ld hl, 0
					ld (BASIC_DATLIN), hl       ; data items index
					ret
        )"
    },

    {
        "PRINT",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.STR        ;
                    or 0
                    ret z                   ; return if string size zero
                    if defined EXIST_DATA_SET
                       ld (BIOS_TEMP), a    ; size of string
                       call gfxIsTileMode
                       ld a, (BIOS_TEMP)
                       jp nz, STRING.PRINT
                         ; discard if first char < 32 or > 126
                         ld a, (hl)
                         cp 126
                         ret nc
                         cp 31
                         ret c
                         push hl
                           ; adjust default color
                           ld a, (BIOS_GRPACY)
                           sra a
                           sra a
                           sra a   ; Y / 8 = bank
                           ld (BIOS_TEMP+1), a
                           ld a, (BIOS_TEMP)
                           PRINT.1:
                             push af
                               ld a, (BIOS_TEMP+1)
                               ld d, a
                               ld e, (hl)
                               push hl
                                 call gfxSetTileDefaultColor
                               pop hl
                               inc hl
                             pop af
                             dec a
                             jr nz, PRINT.1
                           ; print string
                           ld hl, (BIOS_GRPACY)
                           ;ld bc, 32
                           ;call MATH.MULT.16     ; slow y * 32
                           ld h, l
                           sra h
                           sra h
                           sra h
                           sla l
                           sla l
                           sla l
                           sla l
                           sla l            ; fast y * 32
                           ld de, (BIOS_GRPACX)
                           add hl, de
                           ld de, (BIOS_GRPNAM)
                           add hl, de
                           ex de, hl
                         pop hl
                         ld b, 0
                         ld a, (BIOS_TEMP)
                         ld c, a
                         jp gfxLDIRVM
                    else
		               jp STRING.PRINT
                    endif
        )"
    },

    {
        "INPUT",
        R"(
                    pop.parm                    ; get first parameter
                    ld (BIOS_TEMP), hl          ; input variable
                    ld a, 1
                    ld (BIOS_CLIKSW), a         ; enable keyboard click
                    __call_bios BIOS_QINLIN     ; get user input (hl = text start, carry if STOP)
                    jp INPUT.CONTINUE
        )"
    },

    {
        "LINE_INPUT",
        R"(
                    pop.parm                    ; get first parameter
                    ld (BIOS_TEMP), hl          ; input variable
                    ld a, 1
                    ld (BIOS_CLIKSW), a         ; enable keyboard click
                      __call_bios BIOS_INLIN    ; get user input (hl = text start, carry if STOP)
                    jp INPUT.CONTINUE
        )"
    },

    {
        "PLAY",
        R"(
                    pop.parm                ;
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    xor a
                    or c
                    jr nz,PLAY.1
                       ld c, 7              ; 3 bits on
                       jr PLAY.2
            PLAY.1:
                    ld a, c
                    cp 3
                    jr nz,PLAY.2
                       ld c, 4
            PLAY.2:
                    di
                      ld a, (BIOS_MUSICF)
                    ei
                    and c
                    jp z, BOOLEAN.RET.FALSE ;
                    jp BOOLEAN.RET.TRUE
        )"
    },

    {
        "MUSIC_STOP",
        R"(
            ; clear all voices
            ld hl, (GFX_MUSIC_START)
            inc hl
            inc hl
            ld (GFX_MUSIC_NEXT), hl
            ld (GFX_MUSIC_PREV), hl
            jp BIOS_GICINI
        )"
    },

    {
        "MUSIC_PLAY",
        R"(
            ; check audio stream
            ld hl, (GFX_MUSIC_START)
            ld a, h
            cp l
            ret z

            ; test voice
            di
                ld a, (BIOS_MUSICF)
            ei
            and 1   ; voice 1 is on?
            ret nz

            ; play next audio stream
            xor a ;ld a, 0
            ld (BIOS_TEMP), a       ; save voice number

            ld hl, (GFX_MUSIC_NEXT)
            ld a, (hl)
            cp 255                  ; test if repeat
            jr z, MUSIC_PLAY.REPEAT
              push hl                 ; current stream
                xor a
                ld bc, 0
                cpir
                ld a, c               ; get string size
                cpl
                ld d, h
                ld e, l
                ;inc de                ; next stream
              pop hl
              jr MUSIC_PLAY.CONTINUE

        MUSIC_PLAY.REPEAT:
              ld d, h
              ld e, l
              inc de     ; save next
              ld hl, (GFX_MUSIC_PREV)
              push hl                 ; current stream
                xor a
                ld bc, 0
                cpir
                ld a, c               ; get string size
                cpl
              pop hl

        MUSIC_PLAY.CONTINUE:
            push de
              ld (GFX_MUSIC_PREV), hl
              call SET_PLAY_VOICE.1
              call PLAY_HOOK
            pop de

            ; skip to next audio stream
            ; verifing if end of audio stream
            ld hl, (GFX_MUSIC_START)
            ld c, (hl)
            inc hl
            ld b, (hl)
            inc hl
            add hl, bc
            xor a
            sbc hl, de   ; verify if DE is end of audio stream
            jr nz, MUSIC_PLAY.NEXT
              ld de, (GFX_MUSIC_START)
              inc de
              inc de

        MUSIC_PLAY.NEXT:
            ld (GFX_MUSIC_NEXT), de
            ret

        MUSIC_PLAY.VERIFY_END_OF_STREAM:
            ret

        )"
    },

    {
        "DRAW",
        R"(
                    pop.parm                ;
                    call CAST_TO.STR        ;
                    call COPY_TO.BASIC_BUF  ;
                    __call_basic BASIC_DRAW ; draw a string stream in HL
                    ret                     ;
        )"
    },

    {
        "SET_PLAY_VOICE_1",
        R"(
					ld a, 0
					jp SET_PLAY_VOICE
        )"
    },

    {
        "SET_PLAY_VOICE_2",
        R"(
					ld a, 1
					jp SET_PLAY_VOICE
        )"
    },

    {
        "SET_PLAY_VOICE_3",
        R"(
					ld a, 2
					jp SET_PLAY_VOICE
        )"
    },

    {
        "DO_PLAY",
        R"(
             jp PLAY_HOOK
        )"
    },

    {
        "DO_DRAW",
        R"(
                    pop.parm                ;
					ld a, (hl)
					cp 3
					ret nz                  ; return if not string
					ld a, (BIOS_SCRMOD)
					cp 2
					ret c                   ; return if not graphical screen mode
                    call GET_STR.ADDR
                    jp DRAW_HOOK
        )"
    },

    {
        "SOUND",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (BIOS_TEMP), bc
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld e, c                 ; E = register content
                    ld a, (BIOS_TEMP)       ; A = PSG register number
                    __call_bios BIOS_WRTPSG ; write to sound system
        ;            __call_bios BIOS_STRTMS      ; start to play the sound
                    ret                     ;
        )"
    },

    {
        "SCREEN",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = screen number (0 to 3)
                    cp 9
                    jr c, SCREEN.1          ; if mode < 9, jump
                    ld a, 8                 ; else, fix to 8
         SCREEN.1:
             if defined EXIST_DATA_SET
                    call gfxSetScreenMode
                    call gfxIsTileMode
                    ret nz

                    jp gfxClearTileScreen
             else
                    jp gfxSetScreenMode
             endif
        )"
    },

    {
        "SPRITEMODE",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = sprite mode (0 to 3)
                    jp gfxSetSpriteMode
        )"
    },

    {
        "WIDTH",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, (BIOS_OLDSCR)     ; last text screen mode
                    and a                   ; test if zero
                    ld a, c                 ; copy parameter to A
                    jr z, WIDTH.40          ; if zero = 40 columns
         WIDTH.32:  ld (BIOS_LINL32), a
                    jr WIDTH.SET
         WIDTH.40:  ld (BIOS_LINL40), a
         WIDTH.SET: ld (BIOS_LINLEN), a
                    sub 0x0E
                    add a, 0x1C
                    cpl
                    inc a
                    add a, c
                    ld (BIOS_CLMLST), a
                    ld a, 0x0C              ; new page (clear the screen)
                    rst BIOS_OUTDO          ; output to screen
                    ret                     ;
        )"
    },

    {
        "COLOR_FOREGROUND",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = pixel color
                    ;call gfxSetForeColor
                    ld (BIOS_FORCLR), a     ; foreground color
                    ld (BIOS_ATRBYT), a
                    ret                     ;
        )"
    },

    {
        "COLOR_BACKGROUND",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = pixel color
                    ;call gfxSetBackColor
                    ld (BIOS_BAKCLR), a     ; foreground color
                    ret                     ;
        )"
    },

    {
        "COLOR_BORDER",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = pixel color
                    ;call gfxSetBorderColor
                    ld (BIOS_BDRCLR), a     ; border color
                    ret                     ;
        )"
    },

    {
        "COLOR",
        R"(
                    ;call gfxSetColor
                    __call_bios BIOS_CHGCLR ; change VDP colors
                    ret                     ;
        )"
    },

    {
        "BEEP",
        R"(
                    __call_bios BIOS_BEEP   ; play a beep in the system speaker
                    ret                     ;
        )"
    },

    {
        "LOCATE",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (BIOS_TEMP), bc
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld l, c                 ; set second parameter (L = Y position)
                    inc l                   ; BASIC has start position on 0,0 instead of 1,1 of BIOS
                    ld a, (BIOS_TEMP)       ; set first parameter (H = X position)
                    ld h, a
                    inc h                   ; BASIC has start position on 0,0 instead of 1,1 of BIOS
		            ld a, (BIOS_SCRMOD)     ; 0=40x24 Text Mode, 1=32x24 Text Mode, 2=Graphics Mode, 3=Multicolour Mode
		            cp 2                    ;
		            jr nc, LOCATE.G         ; jump if graphic screen mode (not a < 2)
        LOCATE.T:   __call_bios BIOS_POSIT  ; H = X, L = Y
                    ret                     ;
        LOCATE.G:   ld b, 0                 ;
                    ld c, h                 ;
                    ld d, 0                 ;
                    ld e, l                 ;
                    ld (BIOS_GRPACX), bc    ;
                    ld (BIOS_GRPACY), de    ;
                    __call_bios BIOS_SCALXY ; BC = X, DE = Y
                    __call_bios BIOS_MAPXYC ; BC = X, DE = Y
                    ret                     ;
        )"
    },

    {
        "SLEEP",
        R"(
                     pop.parm               ; get first parameter
                     call CAST_TO.INT       ;
                     call GET_INT.VALUE     ; put parameter into BC
                     ld hl, BIOS_JIFFY      ; time counter address
                     ei                     ; enable interrupts
        SLEEP.START: ld a, (hl)             ; get time counter
        SLEEP.LOOP:  halt                   ; execution continues after some interrupt occurred
                     cp (hl)                ; time counter changed from before?
                     jr z, SLEEP.LOOP       ; no change -> interrupted but from elsewhere than VDP
                     djnz SLEEP.START       ; loop until BC = zero
                     ret                    ;
        )"
    },

    {
        "CLEAR",
        R"(
                     ld hl, 0xc5b7            ; himem location
                     ld (BIOS_HIMEM), hl      ; set the max memory location
                     ld hl, 0xc2d7            ; equals MEMSIZ value - 2 - 267*2 [i/o buffer size] - argument for CLEAR
                     ld (BIOS_STKTOP), hl     ; into the location for the top of the stack
                     ld hl, 0xc39f            ; equals STKTOP + (size of CLEAR argument, in this case 200 (the default))
                     ld (BIOS_MEMSIZ), hl     ; top of string storage
                     __call_bios BIOS_IOALLOC ; runs the setup routine necessary, and then returns
                     ret                      ;
        )"
    },

    {
        "BASE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld hl, BIOS_TXTNAM
                     add hl, bc
                     add hl, bc
                     ld a, (hl)
                     ld c, a
                     inc hl
                     ld a, (hl)
                     ld b, a
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "SET_BASE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (base number)
                     ld hl, BIOS_TXTNAM
                     add hl, bc
                     add hl, bc
                     ld (BIOS_TEMP), hl
                     pop.parm                    ; get second parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (data)
                     ld ix, (BIOS_TEMP)
                     ld (ix), c
                     ld (ix+1), b
                     ret
        )"
    },

    {
        "PEEK",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
					 ld h, b
					 ld l, c
                     ld a, (hl)
                     ld c, a
                     ld b, 0
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "VPEEK",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
					 ld h, b
					 ld l, c
                     call gfxRDVRM
                     ld c, a
                     ld b, 0
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "POKE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld (BIOS_TEMP), bc          ; address
                     pop.parm                    ; get second parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld ix, (BIOS_TEMP)          ; address
                     ld (ix), c                  ; data (low)
                     ret
        )"
    },

    {
        "VPOKE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld (BIOS_TEMP), bc          ; address
                     pop.parm                    ; get second parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld hl, (BIOS_TEMP)          ; address
                     ld a, c                     ; data (low)
                     call gfxWRTVRM
                     ret
        )"
    },

    {
        "VDP",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld a, c                     ; register number
                     call gfxRDVDP
                     ld c, a                     ; register data
                     ld b, 0
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "SET_VDP",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld (BIOS_TEMP), bc          ; register number
                     pop.parm                    ; get second parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld b, c                     ; data to write
                     ld a, (BIOS_TEMP)
                     ld c, a                     ; register number
                     call gfxWRTVDP
                     ret
        )"
    },

    {
        "VARPTR",
        R"(
                     pop.parm                    ; get first parameter in HL
                     inc hl
                     inc hl
                     inc hl
					 ld b, h
					 ld c, l
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "HOTSPOT",
        R"(
                     ld a, (GFX_SPRITE_FLAGS)
                     and 0x20                      ; check hotspot flag
                     jp z, BOOLEAN.RET.FALSE
                     jp BOOLEAN.RET.TRUE
        )"
    },

    {
        "COLLISION",
        R"(
                     ld a, (GFX_SPRITE_FLAGS)
                     and 0x40                      ; check collision flag
                     jp z, BOOLEAN.RET.FALSE
                     jp BOOLEAN.RET.TRUE
        )"
    },

    {
        "TOUCHER",
        R"(
                     ld a, (GFX_SPRITE_HOTSPOT_TILE) ; last tile touched
                     ld b, 0
                     ld c, a
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "COLLIDER",
        R"(
                     ld a, (GFX_SPRITE_COLLIDER) ; last collider detected
                     ld b, 0
                     ld c, a
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "COLLIDED",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld (BIOS_TEMP), bc          ; sprite number
                     call gfxCheckSpriteCollisionTable
                     jp z, BOOLEAN.RET.FALSE
                     jp BOOLEAN.RET.TRUE
        )"
    },

    {
        "COLLISION_ALL",
        R"(
                     ld a, (BIOS_STATFL)
                     bit 5, a
                     jp z, BOOLEAN.RET.FALSE     ; return false

                     ; copy sprite attribute table to ram
                     ld hl, (BIOS_ATRBAS)        ; source: attribute table
                     ld de, BASIC_BUF            ; dest: ram
                     ld bc, 128                  ; 32*4 = size of attribute table
                     call gfxLDIRMV

                     ; pre-calculate each sprite width
                     ld ix, BASIC_BUF            ; start of sprites attributes
                     ld b, 32                    ; sprite count
			         ld a, (GFX_SPRITE_SIZE_DAT) ; sprite size
			         ld c, a
			         ld d, 208                   ; Y no-display flag

			         ; test screen mode
			         ld a, (BIOS_SCRMOD)
			         cp 3           ; above screen 3?
			         jr c, COLLISION.next
			           ld d, 216    ; Y no-display flag

                COLLISION.next:
                     ; test IC flag (no collision) in color table
                     ; test EC flag (early clock, shift 32 dots to the left) in color table
                     ; set x1 = x + size, y1 = y + size
                     ld a, (ix)   ; y
                     cp d         ; test if sprite will not be displayed
                     jp z, BOOLEAN.RET.TRUE
                     add a, c
                     ld (ix+2), a ; y1
                     ld a, (ix+1) ; x
                     add a, c
                     ld (ix+3), a ; x1

                     inc ix
                     inc ix
                     inc ix
                     inc ix
                     djnz COLLISION.next

                     jp nz, BOOLEAN.RET.TRUE     ; return true
        )"
    },

    {
        "COLLIDED_ALL",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     ld (BIOS_TEMP), bc          ; sprite number

                     ; get target sprite address (iy)
                     ld hl, (BIOS_TEMP)
                     ld de, BASIC_BUF
                     add hl, hl            ; x4
                     add hl, hl
                     add hl, de
                     push hl
                     pop iy

                     ; start test against others sprites
                     ld ix, BASIC_BUF
                     ld bc, 0
			         ld d, 208      ; Y no-display flag

			         ; test screen mode
			         ld a, (BIOS_SCRMOD)
			         cp 3           ; above screen 3?
			         jr c, COLLIDED.test
			           ld d, 216    ; Y no-display flag

                COLLIDED.test:
                     ; skip target sprite
                     ld a, (BIOS_TEMP)
                     cp c
                     jr z, COLLIDED.next

                     ; test if x1 > nx and x < nx1 and y1 > ny and y < ny1
                     ld a, (ix)   ; ny
                     cp d         ; test if sprite will not be displayed
                     jp z, BOOLEAN.RET.FALSE
                     cp (iy+2)    ; y1
                     jr nc, COLLIDED.next

                     ld a, (ix+1) ; nx
                     cp (iy+3)    ; x1
                     jr nc, COLLIDED.next

                     ld a, (iy+1) ; x
                     cp (ix+3)    ; nx1
                     jr nc, COLLIDED.next

                     ld a, (iy)   ; y
                     cp (ix+2)    ; ny1
                     jr nc, COLLIDED.next

                     ; if so, save collider sprite
                     ld a, c
                     ld (GFX_SPRITE_COLLIDER), a
                     jp BOOLEAN.RET.TRUE      ; return true

                     ; else, next
                COLLIDED.next:
                     inc c
                     inc ix
                     inc ix
                     inc ix
                     inc ix
                     ld a, 32
                     cp c
                     jr nz, COLLIDED.test

                     jp BOOLEAN.RET.FALSE     ; return false
        )"
    },

    {
        "INT",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "TIME",
        R"(
                     ld ix, BIOS_JIFFY           ; time counter address
                     di                          ;
                       ld c, (ix)                ; get time counter (low)
                       ld b, (ix+1)              ; get time counter (high)
                     ei                          ;
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "SET_TIME",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
                     di                          ;
                       ld (BIOS_JIFFY), bc
                     ei                          ;
                     ret
        )"
    },

    {
        "ON_GOTO_START",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
					 ld (BIOS_TEMP), bc          ; save destination index
					 ld hl, 0
					 ld (BIOS_TEMP2), hl         ; save current index
					 ret
        )"
    },

    {
        "ON_GOTO_NEXT",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (current destination)
					 ld hl, (BIOS_TEMP2)         ; restore current index
					 inc hl                      ; skip to next index
					 ld (BIOS_TEMP2), hl         ; save current index
					 ld de, (BIOS_TEMP)          ; restore destination index
					 xor a
					 sbc hl, de
					 ret nz                      ; return if current index <> destination index
					   pop hl                    ; delete old return destination
					   push bc                   ; insert new return destination
					   ret
        )"
    },

    {
        "ON_GOSUB_START",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC
					 ld (BIOS_TEMP), bc          ; save destination index
					 ld hl, 0
					 ld (BIOS_TEMP2), hl         ; save current index
					 ret
        )"
    },

    {
        "ON_GOSUB_NEXT",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (current destination)
					 ld hl, (BIOS_TEMP2)         ; restore current index
					 inc hl                      ; skip to next index
					 ld (BIOS_TEMP2), hl         ; save current index
					 ld de, (BIOS_TEMP)          ; restore destination index
					 xor a
					 sbc hl, de
					 ret nz                      ; return if current index <> destination index
					   push bc                   ; insert new return destination
					   ret
        )"
    },

    {
        "ON_ERROR",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_OTHER
					 jp SET_TRAP                 ; hl = trap block, de = trap handler pointer
        )"
    },

    {
        "ON_INTERVAL",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (interval value)
                     di
					   ld (BASIC_INTVAL), bc
					   ld (BASIC_INTCNT), bc
                     ei
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_INTERVAL
					 jp SET_TRAP                 ; hl = trap block, de = trap handler pointer
        )"
    },

    {
        "ON_KEY_START",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_KEY
					 call SET_TRAP               ; hl = trap block, de = trap handler pointer
					 ld (BIOS_TEMP), hl
					 ret
        )"
    },

    {
        "ON_KEY_NEXT",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, (BIOS_TEMP)
					 call SET_TRAP               ; hl = trap block, de = trap handler pointer
					 ld (BIOS_TEMP), hl
					 ret
        )"
    },

    {
        "ON_SPRITE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_SPRITE
					 jp SET_TRAP                 ; hl = trap block, de = trap handler pointer
        )"
    },

    {
        "ON_STOP",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_STOP
					 jp SET_TRAP                 ; hl = trap block, de = trap handler pointer
        )"
    },

    {
        "ON_STRIG_START",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, BASIC_TRPTBL_STRIG
					 call SET_TRAP               ; hl = trap block, de = trap handler pointer
					 ld (BIOS_TEMP), hl
					 ret
        )"
    },

    {
        "ON_STRIG_NEXT",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap handler)
					 ld d, b
					 ld e, c
					 ld hl, (BIOS_TEMP)
					 call SET_TRAP               ; hl = trap block, de = trap handler pointer
					 ld (BIOS_TEMP), hl
					 ret
        )"
    },

	{
        "INTERVAL_ON",
        R"(
					 ld hl, BASIC_TRPTBL_INTERVAL
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

	{
        "INTERVAL_OFF",
        R"(
					 ld hl, BASIC_TRPTBL_INTERVAL
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

	{
        "INTERVAL_STOP",
        R"(
					 ld hl, BASIC_TRPTBL_INTERVAL
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

	{
        "SPRITE_ON",
        R"(
					 ld hl, BASIC_TRPTBL_SPRITE
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

	{
        "SPRITE_OFF",
        R"(
					 ld hl, BASIC_TRPTBL_SPRITE
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

	{
        "SPRITE_STOP",
        R"(
					 ld hl, BASIC_TRPTBL_SPRITE
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

	{
        "STOP_ON",
        R"(
					 ld hl, BASIC_TRPTBL_STOP
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

	{
        "STOP_OFF",
        R"(
					 ld hl, BASIC_TRPTBL_STOP
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

	{
        "STOP_STOP",
        R"(
					 ld hl, BASIC_TRPTBL_STOP
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

	{
        "STOP",
        R"(
		             xor a
					 __call_basic BASIC_STOP
					 ret
        )"
    },

	{
        "KEY_ON",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (key number)
					 ld hl, BASIC_TRPTBL_KEY
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

	{
        "KEY_OFF",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (key number)
					 ld hl, BASIC_TRPTBL_KEY
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

	{
        "KEY_STOP",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (key number)
					 ld hl, BASIC_TRPTBL_KEY
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

	{
        "STRIG_ON",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (strig number)
					 ld hl, BASIC_TRPTBL_STRIG
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

	{
        "STRIG_OFF",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (strig number)
					 ld hl, BASIC_TRPTBL_STRIG
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

	{
        "STRIG_STOP",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (strig number)
					 ld hl, BASIC_TRPTBL_STRIG
					 add hl, bc
					 add hl, bc
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

    {
        "TRAP_ENABLE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap block)
					 ld h, b
					 ld l, c
                     __call_basic BASIC_TRAP_ENABLE
                     ret
        )"
    },

    {
        "TRAP_DISABLE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap block)
					 ld h, b
					 ld l, c
                     __call_basic BASIC_TRAP_DISABLE
                     ret
        )"
    },

    {
        "TRAP_PAUSE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap block)
					 ld h, b
					 ld l, c
                     __call_basic BASIC_TRAP_PAUSE
                     ret
        )"
    },

    {
        "TRAP_UNPAUSE",
        R"(
                     pop.parm                    ; get first parameter in HL
                     call CAST_TO.INT            ;
                     call GET_INT.VALUE          ; put parameter into BC (trap block)
					 ld h, b
					 ld l, c
                     __call_basic BASIC_TRAP_UNPAUSE
                     ret
        )"
    },

    {
        "RANDOMIZE",
        R"(
                     jp MATH.RANDOMIZE        ; randomize the seed
        )"
    },

    {
        "SEED",
        R"(
                     pop.parm                 ; get first parameter in HL
                     call CAST_TO.INT         ;
                     call GET_INT.VALUE       ; put parameter into BC
                     jp MATH.SEED             ; put parameter as random seed
        )"
    },

    {
        "RND",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, RND.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        RND.1:       __call_math MATH_RND     ; put in DAC a new random number from previous DAC parameter
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "IRND",
        R"(
		             ld hl, (BASIC_RNDX)         ; Linear-feedback shift register (LFSR, adapted from Ricardo Bittencourt)
		             add hl, hl                  ;
		             sbc a, a                    ;
		             and 83h                     ;
		             xor l                       ;
		             ld l, a                     ;
		             ld (BASIC_RNDX), hl         ;
                     ld c, l                     ;
                     ld b, h                     ;
                     call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                     ret.parm                    ;
        )"
    },

    {
        "COS",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, COS.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        COS.1:       __call_math MATH_COS     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "SIN",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, SIN.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        SIN.1:       __call_math MATH_SIN     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "TAN",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, TAN.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        TAN.1:       __call_math MATH_TAN     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "ATN",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, ATN.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        ATN.1:       __call_math MATH_ATN     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "LOG",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, LOG.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        LOG.1:       __call_math MATH_LOG     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "EXP",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, EXP.1             ; if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        EXP.1:       __call_math MATH_EXP     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "SQR",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 12                   ; test if single/double
		             jr nz, SQR.1             ; jump if already double
		             __call_math MATH_FRCDBL  ; convert DAC to double
        SQR.1:       __call_math MATH_SQR     ; get parameter from DAC and put result in DAC
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "ISQR",
        R"(
                     pop.parm                 ; get parameter
		             call COPY_TO.DAC         ; put in DAC
		             and 2                    ; test if integer
		             jr nz, ISQR.1            ; jump if already integer
		             __call_math MATH_FRCINT  ; convert DAC to integer
        ISQR.1:      ld hl, (BASIC_DAC+2)     ; get integer from DAC
                     call MATH.INT.SQR        ;
                     ld (BASIC_DAC+2), a      ;
                     xor a                    ;
                     ld (BASIC_DAC+3), a      ;
                     jp MATH.PARM.PUSH        ; return a dummy double variable from DAC
        )"
    },

    {
        "PSET.COLOR",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c
                    call gfxSetForeColor
                    jp gfxSetColor
        )"
    },

    {
        "PSET.XY",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (BIOS_GRPACX), bc    ; X
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (BIOS_GRPACY), bc    ; Y
                    jp gfxRefreshXY
        )"
    },

    {
        "LINE",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (GFX_TEMP1), bc      ; dX
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
					ld d, b                 ; dY
					ld e, c
                    ld bc, (GFX_TEMP1)      ; dX
                    jp gfxDrawLine
        )"
    },

    {
        "BOX",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (GFX_TEMP1), bc      ; dX
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
					ld d, b                 ; dY
					ld e, c
                    ld bc, (GFX_TEMP1)      ; dX
                    xor a                   ; a = 0 (framed box)
                    jp gfxDrawBox
        )"
    },

    {
        "BOX_STEP",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACX)
                    xor a
                    adc hl, bc
                    ld (GFX_TEMP1), hl      ; dX
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACY)
                    xor a
                    adc hl, bc
					ld d, h                 ; dY
					ld e, l
                    ld bc, (GFX_TEMP1)      ; dX
                    xor a                   ; a = 0 (framed box)
                    jp gfxDrawBox
        )"
    },

    {
        "FBOX",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld (GFX_TEMP1), bc      ; dX
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
					ld d, b                 ; dY
					ld e, c
                    ld bc, (GFX_TEMP1)      ; dX
                    or 1                    ; a = 1 (filled box)
                    jp gfxDrawBox
        )"
    },

    {
        "FBOX_STEP",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACX)
                    xor a
                    adc hl, bc
                    ld (GFX_TEMP1), hl      ; dX
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACY)
                    xor a
                    adc hl, bc
					ld d, h                 ; dY
					ld e, l
                    ld bc, (GFX_TEMP1)      ; dX
                    or 1                    ; a = 1 (filled box)
                    jp gfxDrawBox
        )"
    },

    {
        "PSET.XY.STEP",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACX)
                    adc hl, bc              ; +X
                    ld (BIOS_GRPACX), hl
                    pop.parm                ; get second parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld hl, (BIOS_GRPACY)
                    adc hl, bc              ; +Y
                    ld (BIOS_GRPACY), hl
                    jp gfxRefreshXY
        )"
    },

    {
        "PSET",
        R"(
                    jp gfxSetPixel
        )"
    },

    {
        "CIRCLE",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
					ld h, b                 ; radius
					ld l, c
					xor a                   ; not filled
                    jp gfxDrawCircle
        )"
    },

    {
        "PAINT.BORDER",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c
                    jp gfxSetBorderFill
        )"
    },

    {
        "PAINT",
        R"(
					xor a                   ; not symmetric
                    jp gfxBorderFill
        )"
    },

    {
        "INKEY",
        R"(
                    __call_bios BIOS_CHSNS      ; is there some character in keyboard buffer?
                    jr nz, INKEY.GET            ;
                      ld hl, LIT_FALSE          ;
                      ret.parm                  ;
        INKEY.GET:  __call_bios BIOS_CHGET      ; get a character from keyboard buffer
                    ld b, 0                     ;
                    ld c, a                     ;
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "INKEY.STR",
        R"(
                    xor a
                    __call_bios BIOS_CHSNS      ; is there some character in keyboard buffer?
                    ld hl, LIT_NULL_STR         ;
                    jr z, INKEY.STR.EXIT
            INKEY.STR.GET:
                      __call_bios BIOS_CHGET
                      call COPY_CHAR_TO_STR
            INKEY.STR.EXIT:
                    call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                    ret.parm                    ;
        )"
    },

    {
        "ASC",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.STR        ; cast to string
                    ld b, 0                 ;
                    ld c, (hl)              ; get first character
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "CHR",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c                 ; A = char
                    call COPY_CHAR_TO_STR
                    call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                    ret.parm                    ;
        )"
    },

    {
        "SPC",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    call GET_STRING_SPACE
                    call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                    ret.parm                    ;
        )"
    },

    {
        "SPACE",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    call GET_STRING_SPACE
                    call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                    ret.parm                    ;
        )"
    },

    {
        "INPUT.FUNCTION.STR",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    call GET_STRING_SPACE   ; in bc = size, out hl = address, out a = string size
                    or a
                    jr z, INPUT.FUNCTION.STR.2
                    push af
                    push hl
                      ld b, a
            INPUT.FUNCTION.STR.1:
                      __call_bios BIOS_CHGET
                      ld (hl), a
                      inc hl
                      xor a
                      ld (hl), a
                      djnz INPUT.FUNCTION.STR.1
					pop hl
					pop af
            INPUT.FUNCTION.STR.2:
                    call COPY_TO.VAR_DUMMY.STR  ; create a fake string variable from HL in HL
                    ret.parm                    ;
        )"
    },

    {
        "INPUT.FUNCTION.INT",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    __call_bios BIOS_CHGET
                    ld b, 0                     ;
                    ld c, a                     ;
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "STRIG",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c
                    __call_bios BIOS_GTTRIG
                    or a                        ; cp 0
                    jp z, BOOLEAN.RET.FALSE      ;
                    jp BOOLEAN.RET.TRUE
        )"
    },

    {
        "STICK",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    ld a, c
                    __call_bios BIOS_GTSTCK
                    ld b, 0                     ;
                    ld c, a                     ;
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "PAD",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    __call_bios BIOS_GTPAD
                    ld b, 0                     ;
                    ld c, a                     ;
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "PDL",
        R"(
                    pop.parm                ; get first parameter
                    call CAST_TO.INT        ;
                    call GET_INT.VALUE      ; output BC with integer value
                    __call_bios BIOS_GTPDL
                    ld b, 0                     ;
                    ld c, a                     ;
                    call COPY_TO.VAR_DUMMY.INT  ; create a fake integer variable from BC in HL
                    ret.parm                    ;
        )"
    },

    {
        "PAUSE",
        R"(
            push af
            push bc
            push de
            push hl
            push ix
            push iy
              __call_bios BIOS_BEEP
              __call_bios BIOS_CHGET
            pop iy
            pop ix
            pop hl
            pop de
            pop bc
            pop af
            ret
        )"
    },

    {
        "COLOR_SPRITE",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get sprite color
			ld a, (hl)
			cp 3                    ; is string?
			jr z, COLOR_SPRITE_STR
              call CAST_TO.INT
              call GET_INT.VALUE      ; output BC with integer value (color)
			  ld a, (GFX_TEMP)        ; sprite number
			  call gfxSetSpriteColorInt
              ret

		COLOR_SPRITE_STR:
			call GET_STR.ADDR       ; put string address into hl
			;push hl
			;pop de                  ; color address array
			ld d, h
			ld e, l
			ld a, (GFX_TEMP)        ; sprite number
			ld c, (ix+3)            ; color array size
			call gfxSetSpriteColorStr
		    ret
        )"
    },

    {
        "SPRITE",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get sprite data
			ld a, (hl)
			cp 3                    ; is string?
			ret nz

			call GET_STR.ADDR       ; put string address into hl
			or a
			ret z                   ; is string empty?

			ld c, a                 ; string size
			ld a, (GFX_TEMP)        ; sprite number
			call gfxSetSpriteData

			ld a, (GFX_TEMP)
			ld b, 0
			ld c, a                ; sprite number
			call gfxSetSpritePattern

			ld b, 0
			ld a, (BIOS_FORCLR)
			ld c, a
			ld a, (GFX_TEMP)
			call gfxSetSpriteColorInt

            ret
        )"
    },

    {
        "PUT_SPRITE",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

			;ld bc, (GFX_TEMP2)
			ld a, (GFX_TEMP1)
			ld b, a
			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld bc, (GFX_TEMP)       ; sprite number
			jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_COLOR",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get color
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP3), bc      ; color

			ld bc, (GFX_TEMP2)
			ld a, (GFX_TEMP1)
			ld b, a
			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP3)
			call gfxSetSpriteColorInt

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_PATNUM",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get pattern number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP4), bc      ; pattern number

			ld bc, (GFX_TEMP2)
			ld a, (GFX_TEMP1)
			ld b, a
			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP4)
			call gfxSetSpritePattern

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_COLOR_PATNUM",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get color
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP3), bc      ; color

            pop.parm                ; get pattern number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP4), bc      ; pattern number

			ld bc, (GFX_TEMP2)
			ld a, (GFX_TEMP1)
			ld b, a
			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP3)
			call gfxSetSpriteColorInt

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP4)
			call gfxSetSpritePattern

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_STEP",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

			ld a, (GFX_TEMP)
			call gfxCALATR

			call gfxRDVRM
            ld bc, (GFX_TEMP2)
			add a, c
			ld c, a

			inc hl
			call gfxRDVRM
            ld de, (GFX_TEMP1)
			add a, e
			ld b, a

			call gfxSpriteStepCheck
            ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
            and 8+16                   ; if limit or wall touched
            ret nz

			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_STEP_COLOR",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get color
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP3), bc      ; color

			ld a, (GFX_TEMP)
			call gfxCALATR

			call gfxRDVRM
            ld bc, (GFX_TEMP2)
			add a, c
			ld c, a

			inc hl
			call gfxRDVRM
            ld de, (GFX_TEMP1)
			add a, e
			ld b, a

			call gfxSpriteStepCheck
            ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
            and 8+16                   ; if limit or wall touched
            ret nz

			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP3)
			call gfxSetSpriteColorInt

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_STEP_PATNUM",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get pattern number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP4), bc      ; pattern number

			ld a, (GFX_TEMP)
			call gfxCALATR

			call gfxRDVRM
            ld bc, (GFX_TEMP2)
			add a, c
			ld c, a

			inc hl
			call gfxRDVRM
            ld de, (GFX_TEMP1)
			add a, e
			ld b, a

			call gfxSpriteStepCheck
            ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
            and 8+16                   ; if limit or wall touched
            ret nz

			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP4)
			call gfxSetSpritePattern

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "PUT_SPRITE_STEP_COLOR_PATNUM",
        R"(
            pop.parm                ; get sprite number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; sprite number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; Y

            pop.parm                ; get color
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP3), bc      ; color

            pop.parm                ; get pattern number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP4), bc      ; pattern number

			ld a, (GFX_TEMP)
			call gfxCALATR

			call gfxRDVRM
            ld bc, (GFX_TEMP2)
			add a, c
			ld c, a

			inc hl
			call gfxRDVRM
            ld de, (GFX_TEMP1)
			add a, e
			ld b, a

			call gfxSpriteStepCheck
            ld a, (GFX_SPRITE_FLAGS)   ; bits 0=chk limits, 1=chk walls, 2=chk hotspots, 3=limit touched, 4=wall touched, 5=hotspot touched, 6=collided
            and 8+16                   ; if limit or wall touched
            ret nz

			ld a, (GFX_TEMP)
			call gfxSetSpriteXY

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP3)
			call gfxSetSpriteColorInt

			ld a,  (GFX_TEMP)
			ld bc, (GFX_TEMP4)
			call gfxSetSpritePattern

			ld bc, (GFX_TEMP)       ; sprite number
            jp gfxUpdateSpriteCollisionTable
        )"
    },

    {
        "LOAD_SET",
        R"(
        DATASET.TILE.BANK:    equ BIOS_PARM1+50
        DATASET.TILE.FONT:    equ DATASET.TILE.BANK+1
        DATASET.DECODER:      equ DATASET.TILE.FONT+1
        DATASET.MEMORY:       equ DATASET.DECODER+1
        DATASET.DATA.ADDR:    equ DATASET.MEMORY+2
        DATASET.DATA.SIZE:    equ DATASET.DATA.ADDR+2
		DATASET.SPRITE.COLOR: equ DATASET.DATA.SIZE+2

        LOAD_SET.INIT:
            call gfxIsTileMode
            ret nz
            ld a, 7                      ; copy to bank 0, 1 and 2
            ld (DATASET.TILE.BANK), a    ; bit 0 = bank 0, bit 1 = bank 1, bit 2 = bank 2
            xor a
            ld (DATASET.TILE.FONT), a    ; 0=bios, 1=zanac, 2=gradius1, 3=gradius2
            ld (DATASET.DECODER), a      ; 0=raw, 1=rle, 2=pletter
			ld (DATASET.SPRITE.COLOR), a ; 0=no set, 1=set
            ld (DATASET.MEMORY), a       ; memory pointer (256 * 3, tile pattern table size)
            ld (DATASET.MEMORY+1), a
            call gfxInitScreenWorkspace
            dec hl
        LOAD_SET.NEXT:
            inc hl
            ld a, (hl)
            cp 0xFF    ; end of dataset
            jp z, LOAD_SET.END
            cp 0x01
            jp z, LOAD_SET.DATA_DECODER
            cp 0x02
            jp z, LOAD_SET.DATA_TILE_BANK
            cp 0x03
            jp z, LOAD_SET.DATA_TILE_FONT
            cp 0x05
            jp z, LOAD_SET.DATA_TILE_LOAD_SET
            cp 0x06
            jp z, LOAD_SET.DATA_TILE_LOAD_COLOR
            cp 0x07
            jp z, LOAD_SET.DATA_SPRITE_LOAD_SET
            cp 0x08
            jp z, LOAD_SET.DATA_SPRITE_LOAD_COLOR
            cp 0x09
            jp z, LOAD_SET.DATA_SPRITE_LOAD_MULTI
            cp 0x0A
            jp z, LOAD_SET.DATA_SCREEN_LOAD_SET
            cp 0x0B
            jp z, LOAD_SET.DATA_TILE_LIMIT
            cp 0x10
            jp z, LOAD_SET.DATA_TILE_SET
            cp 0x11
            jp z, LOAD_SET.DATA_TILE_COLOR
            cp 0x12
            jp z, LOAD_SET.DATA_SPRITE_SET
            cp 0x13
            jp z, LOAD_SET.DATA_SPRITE_MONO
            cp 0x14
            jp z, LOAD_SET.DATA_SPRITE_MULTI
            cp 0x15
            jp z, LOAD_SET.DATA_SCREEN_SET
            cp 0x16
            jp z, LOAD_SET.DATA_TILE_WALL
            cp 0x17
            jp z, LOAD_SET.DATA_TILE_HOTSPOT
            cp 0x18
            jp z, LOAD_SET.DATA_MUSIC

        LOAD_SET.END:
            jp LOAD_SET.FREE_MEM

        LOAD_SET.DATA_DECODER:
            inc hl
            ld a, (hl)
            ld (DATASET.DECODER), a
            or 0
            call nz, LOAD_SET.GET_MEM
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_BANK:
            inc hl
            ld a, (hl)
            ld (DATASET.TILE.BANK), a
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_LIMIT:
            inc hl
            ld a, (GFX_SPRITE_FLAGS)
            and 0xFE  ; clear check limit flag
            or (hl)   ; set limit flag
            ld (GFX_SPRITE_FLAGS), a
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_FONT:
            inc hl
            ld a, (hl)
            ld (DATASET.TILE.FONT), a
            or 0
            jp z, LOAD_SET.DATA_TILE_FONT.0
            cp 1
            jp z, LOAD_SET.DATA_TILE_FONT.1
            cp 2
            jp z, LOAD_SET.DATA_TILE_FONT.2
            cp 3
            jp z, LOAD_SET.DATA_TILE_FONT.3
            ret
        LOAD_SET.DATA_TILE_FONT.0:
            if defined TILE_FONT_SUPPORT
              push bc
                ld bc, 95 * 8    ; bytes to be copied (chars * 8)
                ld (BIOS_TEMP2), bc
                ld bc, 32 * 8    ; pattern generator start position * 8
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_BIOS
            endif
            jp LOAD_SET.NEXT
        LOAD_SET.DATA_TILE_FONT.1:
            if defined FONT_STYLE_1
              push bc
                ld bc, FONT_STYLE_1_DATA
                ld (BIOS_TEMP2), bc
                ld bc, FONT_STYLE_1_SIZE
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_MEMORY
            endif
            jp LOAD_SET.NEXT
        LOAD_SET.DATA_TILE_FONT.2:
            if defined FONT_STYLE_2
              push bc
                ld bc, FONT_STYLE_2_DATA
                ld (BIOS_TEMP2), bc
                ld bc, FONT_STYLE_2_SIZE
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_MEMORY
              push bc
                ld bc, 29 * 8    ; bytes to be copied (chars * 8)
                ld (BIOS_TEMP2), bc
                ld bc, 97 * 8    ; pattern generator start position * 8
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_BIOS
            endif
            jp LOAD_SET.NEXT
        LOAD_SET.DATA_TILE_FONT.3:
            if defined FONT_STYLE_3
              push bc
                ld bc, FONT_STYLE_3_DATA
                ld (BIOS_TEMP2), bc
                ld bc, FONT_STYLE_3_SIZE
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_MEMORY
              push bc
                ld bc, 29 * 8    ; bytes to be copied (chars * 8)
                ld (BIOS_TEMP2), bc
                ld bc, 97 * 8    ; pattern generator start position * 8
                ld (BIOS_TEMP3), bc
              pop bc
              call LOAD_SET.DATA_TILE_FONT.LOAD_FROM_BIOS
            endif
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_FONT.LOAD_FROM_MEMORY:
            if defined FONT_STYLE_1 or defined FONT_STYLE_2 or defined FONT_STYLE_3
                push hl
                    call LOAD_SET.GET_MEM      ; de = temporary memory
                    ld hl, (BIOS_TEMP2)
                    ld bc, (BIOS_TEMP3)
                    call pletter.unpack
                    ld hl, 32 * 8    ; pattern generator start position * 8
                    ld de, (BIOS_GRPCGP)
                    add hl, de
                    ld (BIOS_TEMP), hl
                    ld hl, (DATASET.MEMORY)
                    ld bc, 95 * 8    ; bytes to be copied (chars * 8)
                    call LOAD_SET.COPY_TO_BANK
                pop hl
            endif
            ret

        LOAD_SET.DATA_TILE_FONT.LOAD_FROM_BIOS:
            if defined TILE_FONT_SUPPORT
                push hl
                    ld bc, (BIOS_TEMP2)    ; bytes to be copied (chars * 8)
                    ld hl, (BIOS_TEMP3)    ; pattern generator start position * 8
                    push hl
                        ld de, (BIOS_GRPCGP)
                        add hl, de
                        ld (BIOS_TEMP), hl
                    pop hl
                    ld de, (BIOS_CGPNT)
                    add hl, de
                    call LOAD_SET.COPY_TO_BANK
                pop hl
            endif
            ret

        LOAD_SET.DATA_TILE_LOAD_SET:
            ret

        LOAD_SET.DATA_TILE_LOAD_COLOR:
            ret

        LOAD_SET.DATA_SPRITE_LOAD_SET:
            ret

        LOAD_SET.DATA_SPRITE_LOAD_COLOR:
            ret

        LOAD_SET.DATA_SPRITE_LOAD_MULTI:
            ret

        LOAD_SET.DATA_SCREEN_LOAD_SET:
            ret

        LOAD_SET.DATA_TILE_WALL:
            inc hl
            ld (GFX_SPRITE_WALLS), hl
            ld a, (GFX_SPRITE_FLAGS)
            or 2   ; set walls flag
            ld (GFX_SPRITE_FLAGS), a
            ld c, (hl)
            inc hl
            ld b, (hl)
            add hl, bc
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_HOTSPOT:
            inc hl
            ld (GFX_SPRITE_HOTSPOTS), hl
            ld a, (GFX_SPRITE_FLAGS)
            or 4   ; set walls flag
            ld (GFX_SPRITE_FLAGS), a
            ld c, (hl)
            inc hl
            ld b, (hl)
            add hl, bc
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_MUSIC:
            inc hl
            ld (GFX_MUSIC_START), hl
            ld c, (hl)
            inc hl
            ld b, (hl)
            inc hl
            ld (GFX_MUSIC_NEXT), hl
            ld (GFX_MUSIC_PREV), hl
            dec hl
            add hl, bc
            jp LOAD_SET.NEXT

        LOAD_SET.DATA_TILE_SET:
            call LOAD_SET.GET_ADDR
            ret z
            ld de, (BIOS_GRPCGP)
            ld (BIOS_TEMP), de
            call LOAD_SET.COPY_TO_BANK
            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.DATA_TILE_COLOR:
            call LOAD_SET.GET_ADDR
            ret z
            ld de, (BIOS_GRPCOL)
            ld (BIOS_TEMP), de
            call LOAD_SET.COPY_TO_BANK
            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.COPY_TO_BANK:
            ld a, (DATASET.TILE.BANK)
            and 1
            jr z, LOAD_SET.COPY_TO_BANK.1
              ld de, (BIOS_TEMP)
              call LOAD_SET.COPY_TO_VRAM
        LOAD_SET.COPY_TO_BANK.1:
            ld a, (DATASET.TILE.BANK)
            and 2
            jr z, LOAD_SET.COPY_TO_BANK.2
              ld de, (BIOS_TEMP)
              ld a, 8
              add a, d
              ld d, a
              call LOAD_SET.COPY_TO_VRAM
        LOAD_SET.COPY_TO_BANK.2:
            ld a, (DATASET.TILE.BANK)
            and 4
            ret z

            ld de, (BIOS_TEMP)
            ld a, 16
            add a, d
            ld d, a
            jp LOAD_SET.COPY_TO_VRAM

        LOAD_SET.DATA_SPRITE_SET:
            call LOAD_SET.GET_ADDR
            ret z
            ld de, (BIOS_GRPPAT)
            call LOAD_SET.COPY_TO_VRAM
            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.DATA_SPRITE_MONO:		; mono color
            call LOAD_SET.GET_ADDR
            ret z

			ld a, (DATASET.SPRITE.COLOR)
			or 0       ; color settled?
			jp nz, LOAD_SET.SKIP_TO_NEXT

			ld a, (BIOS_SCRMOD)
			cp 4
			jp z, LOAD_SET.DATA_SPRITE_MONO_TO_MULTI

			push bc
			push hl

			  ld de, (BIOS_ATRBAS)     	; sprite attribute table
			  inc de
			  inc de
			  inc de					; mono color attribute

			  ex de, hl

		LOAD_SET.DATA_SPRITE_MONO.NEXT:
			  ld a, (de)        ; color
			  call gfxWRTVRM
			  inc hl
			  inc hl
			  inc hl
			  inc hl

			  inc de
			  dec bc
              jr nz, LOAD_SET.DATA_SPRITE_MONO.NEXT

        LOAD_SET.DATA_SPRITE_MONO.END:
            pop hl
            pop bc

			ld a, (DATASET.SPRITE.COLOR)
			or 1       ; color settled
			ld (DATASET.SPRITE.COLOR), a

            jp LOAD_SET.SKIP_TO_NEXT

		LOAD_SET.DATA_SPRITE_MONO_TO_MULTI:
			push bc
			push hl

			  ex de, hl

			  push de
			    ld hl, (BIOS_ATRBAS)    ; sprite attribute table
			    ld de, 512
			    xor a
			    sbc hl, de 				; multi color table
			  pop de

		LOAD_SET.DATA_SPRITE_MONO_TO_MULTI.NEXT:
		      ld a, (de)
			  push bc
			    ld bc, 16
  		        call gfxFILVRM
			  pop bc
			  inc de
			  dec bc
			  jr nz, LOAD_SET.DATA_SPRITE_MONO_TO_MULTI.NEXT

			pop hl
			pop bc
            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.DATA_SPRITE_MULTI:		; multi color
            call LOAD_SET.GET_ADDR
            ret z

			ld a, (DATASET.SPRITE.COLOR)
			or 0       ; color settled?
			jp nz, LOAD_SET.SKIP_TO_NEXT

			ld a, (BIOS_SCRMOD)
			cp 2
			jp z, LOAD_SET.DATA_SPRITE_MULTI_TO_MONO

			push bc
			push hl

			  ld de, (BIOS_ATRBAS)     	; sprite attribute table

			  ex de, hl

			  push de
			    xor a
			    ld de, 512
				sbc hl, de      		; address of color table from sprite multicolor
			  pop de

			  ex de, hl

			  ld a, (BIOS_RG1SAV)  		; bit 0 = double size, bit 1 = sprite size (0=8 pixels, 1=16 pixels)
			  bit 1, a
			  jr nz, LOAD_SET.DATA_SPRITE_MULTI.16.NEXT

		LOAD_SET.DATA_SPRITE_MULTI.8.NEXT:
		      push bc
			    ld bc, 8
			    call gfxLDIRVM		; hl + 8, de + 8
				ex de, hl
  			    ld bc, 8
			    xor a
  		        call gfxFILVRM
				ex de, hl
			  pop bc

			  push hl
			    ld h, b
				ld l, c
				ld bc, 8
				xor a
				sbc hl, bc
			    ld b, h
				ld c, l
			  pop hl
              jr c, LOAD_SET.DATA_SPRITE_MULTI.END
              jr z, LOAD_SET.DATA_SPRITE_MULTI.END

			  jr LOAD_SET.DATA_SPRITE_MULTI.8.NEXT

		LOAD_SET.DATA_SPRITE_MULTI.16.NEXT:
			  call gfxLDIRVM		; hl + bc, de + bc

        LOAD_SET.DATA_SPRITE_MULTI.END:
            pop hl
            pop bc

			ld a, (DATASET.SPRITE.COLOR)
			or 1       ; color settled
			ld (DATASET.SPRITE.COLOR), a

            jp LOAD_SET.SKIP_TO_NEXT

		LOAD_SET.DATA_SPRITE_MULTI_TO_MONO:
			push bc
			push hl

			  ld de, (BIOS_ATRBAS)     	; sprite attribute table
			  inc de
			  inc de
			  inc de 					; mono color attribute

			  ex de, hl

			  ld a, 8
			  ld (BIOS_TEMP2), a

			  ld a, (BIOS_RG1SAV)  		; bit 0 = double size, bit 1 = sprite size (0=8 pixels, 1=16 pixels)
			  bit 1, a
			  jr z, LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.NEXT

			     ld a, 16
				 ld (BIOS_TEMP2), a

		LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.NEXT:
			  ld a, (de)        ; color
			  call gfxWRTVRM
			  inc hl
			  inc hl
			  inc hl
			  inc hl

			  push hl
			    ld h, b
				ld l, c
				ld a, (BIOS_TEMP2)
				ld b, 0
				ld c, a
				ex de, hl
				add hl, bc
				ex de, hl
				xor a
				sbc hl, bc
			    ld b, h
				ld c, l
			  pop hl
              jr c, LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.END
              jr z, LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.END

			  jr LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.NEXT

        LOAD_SET.DATA_SPRITE_MULTI_TO_MONO.END:
            pop hl
            pop bc

            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.DATA_SCREEN_SET:
            call LOAD_SET.GET_ADDR
            ret z
            ld de, (BIOS_GRPNAM)
            call LOAD_SET.COPY_TO_VRAM
            jp LOAD_SET.SKIP_TO_NEXT

        LOAD_SET.COPY_TO_VRAM:
            push hl
            push bc
              call gfxLDIRVM
            pop bc
            pop hl
            ret

        LOAD_SET.DECODER.INIT:
            ld de, (DATASET.MEMORY)
            ld (DATASET.DATA.ADDR), hl
            ld (DATASET.DATA.SIZE), bc
            ld a, c
            or b
            ret

        LOAD_SET.DECODER.RAW:
            ; test data size
            ld a, c
            or b
            ret

        LOAD_SET.DECODER.RLE:
            call LOAD_SET.DECODER.INIT
            if defined RLE_DECODER
              jp nz, rle.unpack
            endif
            ret

        LOAD_SET.DECODER.PLETTER:
            call LOAD_SET.DECODER.INIT
            ret z
            if defined PLETTER_DECODER
              call pletter.unpack
			  exx 						; calculate decoded length (de' point to end of decoded data)
			  ex de, hl
              ld de, (DATASET.MEMORY)
			  xor a
			  sbc hl, de
			  ld b, h
			  ld c, l 					; bc = decoded length
              ld hl, (DATASET.MEMORY)	; hl = decoded data
              ret
            endif

        LOAD_SET.GET_ADDR:
            inc hl
            ld c, (hl)
            inc hl
            ld b, (hl)

            ; data pointer
            inc hl

            ; test decoder type
            ld a, (DATASET.DECODER)
            cp 1
            jp z, LOAD_SET.DECODER.RLE
            cp 2
            jp z, LOAD_SET.DECODER.PLETTER
            jp LOAD_SET.DECODER.RAW

        LOAD_SET.SKIP_TO_NEXT:
            ; test decoder type
            ld a, (DATASET.DECODER)
            or 0
            call nz, LOAD_SET.SKIP_TO_NEXT.1

            add hl, bc
            dec hl
            jp LOAD_SET.NEXT

        LOAD_SET.SKIP_TO_NEXT.1:
            ld hl, (DATASET.DATA.ADDR)
            ld bc, (DATASET.DATA.SIZE)
            ret

        LOAD_SET.GET_MEM:
            ld de, (DATASET.MEMORY)
            ld a, d
            or e
            ret nz
            push hl
              ld bc, 0x300               ; 256 * 3 (tile pattern table size)
              call memory.alloc          ; in bc size, out ix new memory address, nz=OK
              jp z, memory.error
              ld (DATASET.MEMORY), ix
              ld de, (DATASET.MEMORY)
            pop hl
            ret

        LOAD_SET.FREE_MEM:
            ld de, (DATASET.MEMORY)
            ld a, d
            or e
            ret z
            ld ix, (DATASET.MEMORY)
            jp memory.free          ; in bc size, out ix new memory address, nz=OK
        )"
    },

    {
        "PUT_TILE",
        R"(
            pop.parm                ; get tile number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; tile number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1+1), bc    ; Y

            call gfxIsTileMode
            ret nz

            ld bc, (GFX_TEMP1)
            ld a, (GFX_TEMP)
            jp gfxSetScreenTile
        )"
    },

    {
        "PUT_TILE_COLOR",
        R"(
            pop.parm                ; get tile number
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP), bc       ; tile number

            pop.parm                ; get X
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1), bc      ; X

            pop.parm                ; get Y
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP1+1), bc    ; Y

            pop.parm                ; get color
            call CAST_TO.INT        ;
            call GET_INT.VALUE      ; output BC with integer value
			ld (GFX_TEMP2), bc      ; color

            call gfxIsTileMode
            ret nz

            ld bc, (GFX_TEMP1)
            ld a, (GFX_TEMP)
            call gfxSetScreenTile

            ld hl, GFX_TEMP2
            ld bc, 1
            ld a, (GFX_TEMP)        ; tile number
            ld e, a
            ld a, (GFX_TEMP1+1)     ; y
            srl a
            srl a
            srl a                   ; y / 8 = bank number
            ld d, a
            jp gfxSetTileColor
        )"
    },

    {
        "MATH.ADD",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.ADD.INT     ;
		             cp 3                   ; test if string
		             jp z, STRING.CONCAT    ;
		             cp 4                   ; test if single
		             jp z, MATH.ADD.SGL     ;
		             jp MATH.ADD.DBL        ; it is a double
        )"
    },

	{
        "MATH.SUB",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.SUB.INT     ;
		             cp 3                   ; test if string
		             jp z, MATH.ERROR       ;
		             cp 4                   ; test if single
		             jp z, MATH.SUB.SGL     ;
		             jp MATH.SUB.DBL        ; it is a double
        )"
    },

    {
        "MATH.MULT",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.MULT.INT    ;
		             cp 3                   ; test if string
		             jp z, MATH.ERROR       ;
		             cp 4                   ; test if single
		             jp z, MATH.MULT.SGL    ;
		             jp MATH.MULT.DBL       ; it is a double
        )"
    },

    {
        "MATH.DIV",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.DIV.INT     ;
		             cp 3                   ; test if string
		             jp z, MATH.ERROR       ;
		             cp 4                   ; test if single
		             jp z, MATH.DIV.SGL     ;
		             jp MATH.DIV.DBL        ; it is a double
        )"
    },

    {
        "MATH.IDIV",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.IDIV.INT    ;
		             cp 3                   ; test if string
		             jp z, MATH.ERROR       ;
		             cp 4                   ; test if single
		             jp z, MATH.IDIV.SGL    ;
		             jp MATH.IDIV.DBL       ; it is a double
        )"
    },

    {
        "MATH.POW",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, MATH.POW.INT     ;
		             cp 3                   ; test if string
		             jp z, MATH.ERROR       ;
		             cp 4                   ; test if single
		             jp z, MATH.POW.SGL     ;
		             jp MATH.POW.DBL        ; it is a double
        )"
    },

    {
        "MATH.NEG",
        R"(
		               call CLEAR.DAC       ; put zero in DAC
                       pop.parm             ; get parameter
		               call COPY_TO.ARG     ; put in ARG
		               cp 2                 ; test if integer
		               jp z, MATH.SUB.INT   ;
		               cp 4                 ; test if single
		               jp z, MATH.SUB.SGL   ;
		               cp 8                 ; test if double
		               jp z, MATH.SUB.DBL   ;
                       jp MATH.ERROR        ;
        )"
    },

    {
        "MATH.MOD",
        R"(
                       call MATH.PARM.POP.INT   ; get parameters as integer into DAC/ARG
		               jp MATH.MOD.INT          ;
        )"
    },

    {
        "BOOLEAN.GT",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.GT.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.GT.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.GT.SGL   ;
		             jp BOOLEAN.GT.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.LT",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.LT.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.LT.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.LT.SGL   ;
		             jp BOOLEAN.LT.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.GE",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.GE.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.GE.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.GE.SGL   ;
		             jp BOOLEAN.GE.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.LE",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.LE.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.LE.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.LE.SGL   ;
		             jp BOOLEAN.LE.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.NE",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.NE.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.NE.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.NE.SGL   ;
		             jp BOOLEAN.NE.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.EQ",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.EQ.INT   ;
		             cp 3                   ; test if string
		             jp z, BOOLEAN.EQ.STR   ;
		             cp 4                   ; test if single
		             jp z, BOOLEAN.EQ.SGL   ;
		             jp BOOLEAN.EQ.DBL      ; it is a double
        )"
    },

    {
        "BOOLEAN.AND",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.AND.INT  ;
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.OR",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.OR.INT   ;
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.XOR",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.XOR.INT  ;
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.EQV",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.EQV.INT  ;
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.IMP",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.IMP.INT  ;
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.SHR",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.SHR.INT  ; shift to right
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.SHL",
        R"(
                     call MATH.PARM.POP     ; get parameters into DAC/ARG
		             ld a, (BASIC_VALTYP)   ;
		             cp 2                   ; test if integer
		             jp z, BOOLEAN.SHL.INT  ; shift to left
		             jp MATH.ERROR          ; it is a double
        )"
    },

    {
        "BOOLEAN.NOT",
        R"(
                       pop.parm              ; get parameter
		               call COPY_TO.DAC      ; put in DAC
		               cp 2                  ; test if integer
		               jp z, BOOLEAN.NOT.INT ;
                       jp MATH.ERROR         ;
        )"
    },

    {
        "BOOLEAN.IF",
        R"(
                       pop.parm               ; get parameter boolean result in hl
                       push hl                ; ix = hl
                       pop ix                 ;
                       ld a, (ix+5)           ; put boolean integer result in a
                       ret                    ;
        )"
    },

    { 0, 0 }
};


#endif // SYMBOLS_H_INCLUDED

